<?php

/**
 * @package Festi
 *
 * @property DefaultUser $user
 * @property array $seo
 */
class Core extends Entity
{
    public const EVENT_GET_URL     = "get_url";
    public const EVENT_ON_REDIRECT = "redirect";
    public const EVENT_ON_REQUEST  = "request";
    public const EVENT_ON_RESPONSE = "response";
    public const EVENT_PLUGIN_CALL = "plugin_call";
    public const EVENT_PLUGIN_INIT = "plugin_init";

    /**
     * @deprecated
     */
    public const EVENT_THEME_TITLE = "theme_title";
    
    public const EVENT_ON_CREATE_STORE   = "create_store";
    public const EVENT_ON_AFTER_INIT     = "core_after_init";
    public const EVENT_ON_OPTIONS_UPDATE = "core_options_update";

    public const OPTION_CHARSET                     = 'charset';
    public const OPTION_CLASS_POSTFIX               = 'classPostfix';
    public const OPTION_CONNECTION                  = 'connection';
    public const OPTION_DEFAULT                     = 'default';
    public const OPTION_DEFS_PATH                   = 'defs_path';
    public const OPTION_ENGINE_FOLDER               = 'engine_folder';
    public const OPTION_ENGINE_PATH                 = 'engine_path';
    public const OPTION_ENGINE_URL                  = 'engine_url';
    public const OPTION_ENGINE_TEMPLATE_PATH        = 'engine_template_path';
    public const OPTION_GLOBAL_STORE_EXCEPTION_MODE = 'storeExceptionMode';
    public const OPTION_HTTP_BASE                   = 'http_base';
    public const OPTION_IMAGEMAGIC_PATH             = 'imagemagic_path';
    public const OPTION_LANG                        = 'lang';
    public const OPTION_LOCALE_PATH                 = 'locale_path';
    public const OPTION_OBJECTS_PATH                = 'objects_path';
    public const OPTION_PLUGIN_TEMPLATE_POSTFIX     = 'plugin_template_postfix';
    public const OPTION_PLUGINS_FOLDER              = 'plugins_folder';
    public const OPTION_PLUGINS_HTTP                = 'plugins_http';
    public const OPTION_PLUGINS_PATH                = 'plugins_path';
    public const OPTION_SESSION_DATA                = 'session_data';
    public const OPTION_SESSION_TOKEN_KEY           = 'session_token_key';
    public const OPTION_SESSION_TOKENS              = 'tokens';
    public const OPTION_SSL_PRIVATE_KEY             = 'ssl_private_key';
    public const OPTION_SSL_PUBLIC_KEY              = 'ssl_public_key';
    public const OPTION_SYSTEM_PATH                 = 'system_path';
    public const OPTION_TEMPLATE_PATH               = 'template_path';
    public const OPTION_THEME                       = 'theme';
    public const OPTION_THEME_NAME                  = 'theme_name';
    public const OPTION_THEME_PATH                  = 'theme_path';
    public const OPTION_THEME_TEMPLATE_PATH         = 'theme_template_path';
    public const OPTION_THEME_URL                   = 'theme_url';
    public const OPTION_PATH                        = 'path';

    public const SERVER_REQUEST_URI = "REQUEST_URI";
    
    public const FOLDER_LOCALE = 'locale'.DIRECTORY_SEPARATOR;
    public const FOLDER_THEMES = 'themes'.DIRECTORY_SEPARATOR;

    /**
     * @var Core
     */
    static private $_instance = null;
    static private $_plugins = null;
    static private $_template;

    private $_options;
    private $_store = array();
    private $_hooks = array();
    private $_systemPlugin;

    protected $properties = array();

    /**
     * @var LocaleModel
     */
    protected $locale;

    /**
     * @var ObjectAdapter
     */
    public $db;
    
    /**
     * @param array $options
     * @return Core
     * @throws SystemException
     */
    public static function &getInstance(array $options = array()): Core
    {
        if (self::$_instance == null) {
            self::$_instance = new self($options);
            self::$_instance->_onInit();
        } else if ($options) {
            self::$_instance->_setOptions($options);
        }

        return self::$_instance;
    } // end getInstance
    
    /**
     * Core constructor.
     * @param array $options
     * @throws SystemException
     */
    public function __construct(array $options = array())
    {
        if (isset(self::$_instance)) {
            throw new SystemException(
                'Core already defined use Core::getInstance'
            );
        }

        $this->_verifySystem();

        parent::__construct();

        $this->_setOptions($options);

        $this->onInitLocale();
    } // end __construct

    private function _setOptions(array $options): void
    {
        if (array_key_exists(static::OPTION_CONNECTION, $options)) {
            $this->db = $options[static::OPTION_CONNECTION];
            unset($options[static::OPTION_CONNECTION]);
        }

        $this->_options = $options;

        $this->_setDefaultOptions();

        $this->fireEvent(static::EVENT_ON_OPTIONS_UPDATE);

    } // end _setOptions
    
    private function _verifySystem()
    {
        $originalVersion = phpversion();
        $version = array_map('intval', explode('.', $originalVersion));

        $currentVersion = $version[0] * 10000 + $version[1] * 100 + $version[2];
        $minimalPhpVersion = 70100;

        if ($currentVersion < $minimalPhpVersion) {
            $message = __(
                'The PHP %s version is unsupportable.',
                $originalVersion
            );
            throw new SystemException($message);
        }
    } // end _verifySystem
    
    private function _onInit()
    {
        $this->_onInitPlugins();

        $this->fireEvent(static::EVENT_ON_AFTER_INIT);
    } // end _onInit
    
    /**
     * @return bool
     */
    private function _onInitPlugins(): bool
    {
        $pluginsPath = $this->getOption(static::OPTION_PLUGINS_PATH);
        
        $globPattern = $pluginsPath.'*/init.php';
        
        $initFiles = glob($globPattern);
        if (!$initFiles) {
            return true;
        }
        
        foreach ($initFiles as $path) {
            require_once $path;
        }
        
        return true;
    } // end _onInitPlugins
    
    public function onInitLocale(): void
    {
        $localePaths = array(
            $this->getOption(static::OPTION_LOCALE_PATH),
            $this->getOption(static::OPTION_ENGINE_PATH).static::FOLDER_LOCALE,
        );

        foreach ($localePaths as $localePath) {
            $moFile = $localePath.$this->getOption(static::OPTION_LANG).'.mo';
            if (!file_exists($moFile)) {
                continue;
            }

            $dictionary = new MoDictionaryLocale($moFile);

            $localeModel = new LocaleModel($dictionary);
            $this->setLocaleModel($localeModel);

            return;
        }

        $message = 'Not found locale file or incorrect locale settings';
        throw new SystemException($message);
    } // end onInitLocale
    
    private function _setDefaultOptions(): void
    {
        if (!isset($this->_options[static::OPTION_HTTP_BASE])) {
            $this->_options[static::OPTION_HTTP_BASE] = '/';
        }
        
        if (!isset($this->_options[static::OPTION_THEME_NAME])) {
            $this->_options[static::OPTION_THEME_NAME] = static::OPTION_DEFAULT;
        }
        $themeName = $this->_options[static::OPTION_THEME_NAME];
        
        if (!isset($this->_options[static::OPTION_ENGINE_FOLDER])) {
            $this->_options[static::OPTION_ENGINE_FOLDER] = 'core';
        }
        
        if (!isset($this->_options['core_path'])) {
            $this->_options['core_path'] = __DIR__.DIRECTORY_SEPARATOR;
        }

        if (!isset($this->_options[static::OPTION_PLUGINS_FOLDER])) {
            $this->_options[static::OPTION_PLUGINS_FOLDER] = 'plugins';
        }
        $engineFolder = $this->_options[static::OPTION_ENGINE_FOLDER];
        
        if (!isset($this->_options[static::OPTION_SESSION_DATA]) && session_id()) {
            $this->_options[static::OPTION_SESSION_DATA] = &$_SESSION['festi'];
        }

        $path           = $this->_getRootPath();
        $engineHttpPath = $this->_options[static::OPTION_HTTP_BASE].$engineFolder;
        $defaultThemePath = $path.static::FOLDER_THEMES.$themeName.DIRECTORY_SEPARATOR;

        $enginePath = realpath(__DIR__.DIRECTORY_SEPARATOR.'..'.DIRECTORY_SEPARATOR).DIRECTORY_SEPARATOR;

        //
        $engineTemplatePath = $enginePath.DIRECTORY_SEPARATOR.'templates'.DIRECTORY_SEPARATOR;
        if (defined('ENGINE_TEMPLATE_PATH')) {
            $engineTemplatePath = ENGINE_TEMPLATE_PATH;
        }

        $optionsFields = array(
            static::OPTION_SESSION_TOKEN_KEY => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => 'stoken',
            ),

            static::OPTION_ENGINE_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => $enginePath,
            ),
            
            static::OPTION_THEME_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => $defaultThemePath,
            ),
            
            static::OPTION_THEME_TEMPLATE_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => $defaultThemePath.'templates'.DIRECTORY_SEPARATOR,
            ),
            
            static::OPTION_ENGINE_TEMPLATE_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => $engineTemplatePath,
            ),

            static::OPTION_THEME_URL => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT =>
                    $this->_options[static::OPTION_HTTP_BASE].
                    static::FOLDER_THEMES.$themeName.'/',
            ),
            
            static::OPTION_SYSTEM_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT =>
                    realpath(dirname(__FILE__).'/../../../../').
                    DIRECTORY_SEPARATOR,
            ),
            
            static::OPTION_ENGINE_URL => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => $engineHttpPath.'/public/',
            ),
            
            static::OPTION_IMAGEMAGIC_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => '/usr/bin/convert',
            ),
            
            static::OPTION_CHARSET => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => 'UTF-8',
            ),

            static::OPTION_LANG => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => 'en',
            ),
            
            static::OPTION_PLUGINS_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT =>
                    $path.$this->_options[static::OPTION_PLUGINS_FOLDER].
                    DIRECTORY_SEPARATOR,
            ),
            
            static::OPTION_PLUGINS_HTTP => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT =>
                    '/'.$this->_options[static::OPTION_PLUGINS_FOLDER].'/',
            ),

            static::OPTION_TEMPLATE_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => $path.'templates'.
                                              DIRECTORY_SEPARATOR,
            ),
            
            static::OPTION_DEFS_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => $path.'tblDefs'.
                                              DIRECTORY_SEPARATOR,
            ),
            
            static::OPTION_OBJECTS_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => $path.'objects'.
                                              DIRECTORY_SEPARATOR,
            ),
            
            static::OPTION_LOCALE_PATH => array(
                static::OPTION_KEY_TYPE    => static::FIELD_TYPE_STRING,
                static::OPTION_KEY_DEFAULT => $path.static::FOLDER_LOCALE,
            ),
            
            static::OPTION_PLUGIN_TEMPLATE_POSTFIX => array(
                static::OPTION_KEY_TYPE => static::FIELD_TYPE_STRING,
            ),
        );

        $this->_options = $this->getExtendData($this->_options, $optionsFields);

        if (!defined('FESTI_PLUGINS_PATH')) {
            define(
                'FESTI_PLUGINS_PATH',
                $this->_options[static::OPTION_PLUGINS_PATH]
            );
        }
    } // end _setDefaultOptions

    /**
     * Returns path to root directory of project. You can use FS_ROOT, ENGINE_BASE_PATH constants to override path
     * and core option Core::OPTION_PATH as well.
     *
     * @return string
     */
    private function _getRootPath(): string
    {
        if (empty($this->_options[static::OPTION_PATH])) {
            if (defined('ENGINE_BASE_PATH')) {
                $path = ENGINE_BASE_PATH;
            } else if (isset($_SERVER) && !empty($_SERVER['DOCUMENT_ROOT'])) {
                $path = realpath($_SERVER['DOCUMENT_ROOT']).DIRECTORY_SEPARATOR;
            } else if (defined('FS_ROOT')) {
                $path = FS_ROOT;
            } else {
                $path = realpath(__DIR__.'/../../').DIRECTORY_SEPARATOR;
            }

            $this->_options[static::OPTION_PATH] = $path;
        } else {
            $path = $this->_options[static::OPTION_PATH];
        }

        return $path;
    } // end _getRootPath

    public function __set($label, $object)
    {
        if (!isset($this->_store[$label])) {
            $this->_store[$label] = $object;
        }
    }

    public function __unset($label)
    {
        if (isset($this->_store[$label])) {
            unset($this->_store[$label]);
        }
    }

    public function __get($label)
    {
        if (isset($this->_store[$label])) {
            return $this->_store[$label];
        }
        return false;
    }

    public function __isset($label)
    {
        return isset($this->_store[$label]);
    }
    
    /**
     * @param bool $prefix
     * @return string|null
     */
    public function getCurrentURL(bool $prefix = false): ?string
    {
        if (!$prefix) {
            $prefix = $this->getOption(static::OPTION_HTTP_BASE);
        }
        
        $uri = '/';
        if (array_key_exists(static::SERVER_REQUEST_URI, $_SERVER)) {
            $uri = $_SERVER[static::SERVER_REQUEST_URI];
        }
        
        $path = parse_url($uri, PHP_URL_PATH);

        $url =  empty($path) ? '/' : $path;

        return preg_replace('#^'.$prefix.'#Umis', '/', $url);
    } // end getCurrentURL
    
    /**
     * @param string $plugin
     * @param string|null $classPostfix
     * @return bool
     */
    public function isPluginInstanceExists(
        string $plugin, ?string $classPostfix = null
    ): bool
    {
        if (!$classPostfix) {
            $classPostfix = 'Plugin';
        }
        
        return isset(self::$_plugins[$classPostfix]) && 
               isset(self::$_plugins[$classPostfix][$plugin]);    
    } // end isPluginInstanceExists
    
    /**
     * @param string $plugin
     * @param array $options
     * @return AbstractPlugin|ISystemPlugin
     * @throws SystemException
     */
    public function &getPluginInstance(string $plugin, array $options = array())
    {
        $classPostfix = 'Plugin';
        if (array_key_exists(static::OPTION_CLASS_POSTFIX, $options)) {
            $classPostfix = $options[static::OPTION_CLASS_POSTFIX];
        }
        
        if ($this->isPluginInstanceExists($plugin, $classPostfix)) {
            return self::$_plugins[$classPostfix][$plugin];
        }
    
        if (class_exists($plugin)) {
            $className = $plugin;
            $plugin    = strstr($className, $classPostfix, true);
        } else {
            $className = $plugin.$classPostfix;
        }

        if (isset($options['path'])) {
            $path = $options['path'];
        } else if (isset($options[static::OPTION_PLUGINS_PATH])) {
            $path = $options[static::OPTION_PLUGINS_PATH];
        } else {
            $path = $this->getOption(static::OPTION_PLUGINS_PATH);
        }
        
        $options[static::OPTION_PLUGINS_PATH]  = $path;

        if (isset($options['http_path'])) {
            $httpPath = $options['http_path'];
        } else if (!empty($options[static::OPTION_PLUGINS_HTTP])) {
            $httpPath = $options[static::OPTION_PLUGINS_HTTP];
        } else {
            $httpPath = $this->getOption(static::OPTION_PLUGINS_HTTP);
        }
        
        $options[static::OPTION_PLUGINS_PATH] = $path;

        if (is_dir($path.$plugin)) {
            $path .= $plugin.DIRECTORY_SEPARATOR;
        }

        $options['plugin_path']  = $path;
        $options[static::OPTION_CLASS_POSTFIX] = $classPostfix;
        $options['name']         = $plugin;
        
        $options['plugin_http'] = $httpPath.$plugin."/";

        if (!class_exists($className)) {

            $classNameWithNamespaces = 'plugin\\'.$plugin.'\\'.$className;

            if (!class_exists($classNameWithNamespaces)) {
                $this->_includePlugin($className, $path);
            } else {
                $className = $classNameWithNamespaces;
            }
        }

        $this->_setPluginLocale($plugin, $path);

        // FIXME:
        if (!isset($options['tpl'])) {
            $tpl = self::getTemplateInstance();
        } else {
            $tpl = $options['tpl'];
        }

        $pluginInstance = new $className($tpl);
        $pluginInstance->setOptions($options);
        
        //
        $target = array(
            'plugin' => &$pluginInstance
        );
        
        $event = new FestiEvent(self::EVENT_PLUGIN_INIT, $target);
        $this->dispatchEvent($event);

        $pluginInstance->onInit();
        
        self::$_plugins[$classPostfix][$plugin] = $pluginInstance;

        return self::$_plugins[$classPostfix][$plugin];
    } // end getPluginInstance

    private function _includePlugin(string $className, string $path): void
    {
        $classFile = $path.$className.'.php';

        if (!file_exists($classFile)) {
            $msg = __l(
                'File "%s" for plugin "%s" was not found.',
                $classFile,
                $className
            );

            throw new SystemException($msg);
        }

        require_once $classFile;
        if (!class_exists($className)) {
            $msg = __l(
                'Class "%s" was not found in file "%s".',
                $className,
                $classFile
            );

            throw new SystemException($msg);
        }
    } // end _includePlugin
    
    /**
     * @param string $pluginName
     * @param string $path
     */
    private function _setPluginLocale(string $pluginName, string $path)
    {
        $defaultLocalePath = $this->getOption(static::OPTION_LOCALE_PATH);
        $localePaths = array(
            $defaultLocalePath,
            $path.static::FOLDER_LOCALE
        );
    
        foreach ($localePaths as $localePath) {
            $moFilePath = $localePath.$pluginName."_".
                          $this->getOption(static::OPTION_LANG).".mo";
            if (file_exists($moFilePath)) {
                $pluginDictionary = new MoDictionaryLocale($moFilePath);
                $localeModel = new LocaleModel($pluginDictionary);
                $this->setLocaleModel($localeModel);
            }
        }
    }
    
    public static function &getTemplateInstance()
    {
        if (!is_null(self::$_template)) {
            return self::$_template;
        }

        $path = self::getInstance()->getOption(static::OPTION_TEMPLATE_PATH);
        self::$_template = new Template($path);
        
        return self::$_template;
    } // end getTemplateInstance
    

    public function call(
        $plugin, $method, $params = array(), $options = array()
    )
    {
        $obj = $this->getPluginInstance($plugin, $options);

        if (!is_callable(array($obj, $method))) {
            $msg = __l(
                'Method "%s" was not found in module "%s".', 
                $method, 
                $plugin
            );
            
            throw new SystemException($msg);
        }

        return call_user_func_array(array($obj, $method), $params);
    } // end call
    
    /**
     * Returns content from engine templates folder
     *
     * @param string $tplName
     * @param array $localVars
     * @return string
     * @throws SystemException
     */
    public function systemFetch(
        string $tplName, array $localVars = array()
    ): string
    {
        $path = $this->getOption(static::OPTION_THEME_TEMPLATE_PATH);
        
        $tpl = new Template($path);
        return $tpl->fetch($tplName, $localVars);
    } // end systemFetch

    /**
     * @deprecated
     */
    public function getView(
        &$connection, $table, $params = array(), Response &$response = null
    )
    {
        if (is_null($response)) {
            $response = new Response();
        }

        $store = $this->createStoreInstanceByConnection(
            $connection, 
            $table, 
            $params
        );

        $store->onRequest($response);

        return $response;
    } // end getView
    
    /**
     * Create and returns Store instance by system database connection. 
     * For override model need use $params['model'].
     *
     * @param string $storeName
     * @param mixed[] $params Array for store options.
     * @throws SystemException if undefined database connection in core.
     * @return Store
     */
    public function createStoreInstance(
        string $storeName, array $params = array()
    ): Store
    {
        if (!$this->db) {
            throw new SystemException("Undefined connection.");
        }
        
        return $this->createStoreInstanceByConnection(
            $this->db, 
            $storeName, 
            $params
        );
    } // end createStoreInstance
    
    /**
     * @param IDataAccessObject $connection
     * @param string $table
     * @param array $params
     * @return Store
     * @throws PermissionsException
     * @throws StoreException
     * @throws SystemException
     */
    public function createStoreInstanceByConnection(
        IDataAccessObject &$connection, string $table, $params = array()
    ): Store
    {
        $keyModel = 'model';
        
        if (!empty($params[$keyModel])) {
            $this->_options[$keyModel] = $params[$keyModel];
            unset($params[$keyModel]);
        }
        
        $this->_options['handler_options'] = $params;
        $this->_options['current_url'] = $this->getCurrentURL();

        $store = new Store($connection, $table, $this->_options);
        
        unset($this->_options[$keyModel]);
        
        $target = array(
            'store' => &$store
        );
        
        $this->fireEvent(static::EVENT_ON_CREATE_STORE, $target);
        
        return $store;
    } // end createStoreInstanceByConnection
    
    /**
     * @param string $url
     * @param bool $usePrefix
     */
    public function redirect(string $url = '', bool $usePrefix = true)
    {
        $url = $usePrefix
            ? $this->getOption(static::OPTION_HTTP_BASE).$url
            : $url;
        
        $target = array(
            'url' => &$url
        );
        
        $this->fireEvent(static::EVENT_ON_REDIRECT, $target);
        
        header('Location: '.$url);
        exit();
    } // end redirect
    

    /**
     * Add template property title
     *
     * @param string $title
     */
    public function setTitle(string $title): void
    {
        $this->addProperties('title', $title, true);
    } // end setTitle
    
    /**
     * @param string $path
     * @param bool $theme
     */
    public function includeJs(string $path, bool $theme = true): void
    {
        if ($theme) {
            $path = $this->getOption(static::OPTION_THEME_URL).$path;
        }
        
        $this->addProperties('js', $path);
    }
    
    /**
     * @param string $path
     * @param bool $theme
     */
    public function includeCss(string $path, bool $theme = true): void
    {
        if ($theme) {
            $path = $this->getOption(static::OPTION_THEME_URL).$path;
        }
        
        $this->addProperties('css', $path);
    }

    public function addProperties($name, $value, $is_scalar = false)
    {
        if ($is_scalar) {
            $this->properties[$name] = $value;
            return true;
        }

        if (!isset($this->properties[$name])) {
            $this->properties[$name] = array();
        }

        if (in_array($value, $this->properties[$name])) {
            return true;
        }

        $this->properties[$name][] = $value;

        return true;
    } // end addProperties

    public function getProperties($key = false)
    {
        if (!$key) {
            return $this->properties;
        }
        
        return $this->properties[$key] ?? null;
    }
    
    /**
     * @param string $url
     * @return string
     */
    public function getBaseUrl(string $url): string
    {
        $params = array();
        
        // XXX: Add session token to url, need for correct session work
        $key = $this->getOption(static::OPTION_SESSION_TOKEN_KEY);
        if ($this->_isTokenNeeded($url, $key)) {
            $params[$key] = $_REQUEST[$key];
        }

        if (func_num_args() > 1) {
            $args = func_get_args();
            $chunks = array_slice($args, 1);
            $lastChunk = array_pop($chunks);
            
            if (!is_array($lastChunk)) {
                $chunks[] = $lastChunk;
            } else {
                $params += $lastChunk;
            }
            
            if ($chunks) {
                $url = vsprintf($url, $chunks);
            }
        }
        
        if ($params) {
            $seperator = strpos($url, '?') === false ? '?' : '&';
            $url .= $seperator.http_build_query($params);
        }

        $eventTarget = array(
            'url' => &$url
        );
        
        $event = new FestiEvent(Core::EVENT_GET_URL, $eventTarget);
        $this->dispatchEvent($event);

        return $url;
    } // end getBaseUrl
    
    /**
     * @param string|null $url
     * @param mixed ...$args
     * @return string
     */
    public function getUrl(?string $url, mixed ...$args): string
    {
        if (is_null($url)) {
            $url = '';
        }

        if (!preg_match('#^(www|http)#Umis', $url)) {
            $url = preg_replace('#^/#Umis', $this->getOption(static::OPTION_HTTP_BASE), $url);
        }

        return $this->getBaseUrl($url, ...$args);
    } // end getUrl
    
    /**
     * @param string $url
     * @param string $key
     * @return bool
     */
    private function _isTokenNeeded(string $url, string $key): bool
    {
        return !preg_match('#^(www|http)#Umis', $url) &&
            !empty($_REQUEST[$key]) && strpos($url, $key) === false;
    } // end _isTokenNeeded
    
    
    /**
     * Returns valid fields values. All errors are written to the $errors.
     * 
     * @param mixed $needles
     * @param array $errors
     * @param array|bool $request
     * @return array
     */
    public function getRequestFields(
        $needles, &$errors = array(), $request = false
    )
    {
        if (!$request) {
            $request = &$_REQUEST;
        }
        
        return $this->getPreparedData($request, $needles, $errors);
    } // end getRequestFields

    /**
     * Returns a reference to data in the session used by the jimbo
     *
     * @return array|null
     */
    public function &getSessionData(): ?array
    {
        return $this->_options[static::OPTION_SESSION_DATA];
    } // end getSessionData

    public function setSessionData(&$sessionData)
    {
        $this->_options[static::OPTION_SESSION_DATA] = &$sessionData ;
    }
    
    public function setLocaleModel(LocaleModel $locale)
    {
        $this->locale = $locale;
    } // end setLocaleModel

    public function getLocaleModel(): LocaleModel
    {
        return $this->locale;
    } //end getLocaleModel

    public function setOption($name, $value)
    {
        $this->_options[$name] = $value;
    }
    
    /**
     * @param $name
     * @return mixed|null
     */
    public function getOption($name)
    {
        return $this->_options[$name] ?? null;
    }
    
    public function &getOptions()
    {
        return $this->_options;
    }
    
    /**
     * @param $objectName
     * @param string|bool $pluginName
     * @param string|null $path
     * @return IDataAccessObject
     * @phan-return mixed
     * @throws DatabaseException
     * @throws SystemException
     */
    public function &getObject(
        string $objectName, $pluginName = false, string $path = null
    )
    {
        if ($objectName === false) {
            throw new SystemException("Undefined Object Name");
        }

        if (!isset($this->db)) {
            $msg = __l("Undefined store connection in core.");
            throw new SystemException($msg);
        }

        if (!$path) {
            if ($pluginName) {
                if (is_bool($pluginName)) {
                    $pluginName = $objectName;
                }
                $path = $this->getOption(static::OPTION_PLUGINS_PATH).
                        $pluginName.DIRECTORY_SEPARATOR;
            } else {
                $path = $this->getOption(static::OPTION_OBJECTS_PATH);
            }
        }

        return DataAccessObject::getInstance(
            $objectName, 
            $this->db, 
            $path
        );
    } // end getObject

    public function exec($cmd, &$output = array())
    {
        $res = exec(escapeshellcmd($cmd), $output, $ret);
        if ($ret !== 0) {
            throw new SystemException("Can't exec: ".$cmd, $ret);
        }

        return $res;
    } // end exec
    
    public function getStoragePathPrefixByID($idEntity, $options = array())
    {
        $path = null;
        
        if (!empty($options['path'])) {
            $path = $options['path'];
        }
        
        $storageStep = empty($options['step']) ? 2 : $options['step'];
        $storagePathChunkCount = 2;
        if (!empty($options['chunks'])) {
            $storagePathChunkCount = $options['chunks'];
        }
        
        $minStrLength = $storageStep * $storagePathChunkCount;
    
        if (strlen($idEntity) < $minStrLength) {
            $pathString = str_pad($idEntity, $minStrLength, "0", STR_PAD_LEFT);
        } else {
            $startIndex = strlen($idEntity) - $minStrLength;
            $pathString = substr($idEntity, $startIndex);
        }
    
        $chunks = str_split($pathString, $storageStep);
        
        $pathPrefix = join(DIRECTORY_SEPARATOR, $chunks).DIRECTORY_SEPARATOR;
    
        // Hack: for folder permission from web and cron job
        if ($path) {
            foreach ($chunks as $prefix) {
                $path .= $prefix.DIRECTORY_SEPARATOR;
                if (!is_dir($path) && !mkdir($path, 0777)) {
                    $message = __l("Permission error: %s", $path);
                    throw new PermissionsException($message);
                }
            }
            
            $pathPrefix = $path;
        }
        
        return $pathPrefix;
    }

    ///////////////////////////////
    // Session
    //////////////////////////////
    
    public function isExistParam($key)
    {
        return isset($_SESSION[$key]);
    }

    public function addParam($key, $value, $isUseToken = false)
    {
        if ($isUseToken) {
            $tokenKey = $this->getOption(static::OPTION_SESSION_TOKEN_KEY);
            if (isset($_REQUEST[$tokenKey])) {
                $tokens = &$_SESSION[static::OPTION_SESSION_TOKENS];
                $tokens[$_REQUEST[$tokenKey]][$key] = $value;
                return true;
            }
        } 
        
        $_SESSION[$key] = $value;
    } // end addParam

    public function getParam($key, $isUseToken = false)
    {
        if ($isUseToken) {
            $tokenKey = $this->getOption(static::OPTION_SESSION_TOKEN_KEY);
            if (isset($_REQUEST[$tokenKey])) {
                $token = $_REQUEST[$tokenKey];

                return $_SESSION[static::OPTION_SESSION_TOKENS][$token][$key]
                       ?? false;
            }
        }
        
        return $_SESSION[$key] ?? false;
    } // end getParam

    public function popParam($key, $isUseToken = false)
    {
        if ($isUseToken) {
            $tokenKey = $this->getOption(static::OPTION_SESSION_TOKEN_KEY);
            if (isset($_REQUEST[$tokenKey])) {
                $token = $_REQUEST[$tokenKey];
                
                $value = $_SESSION[static::OPTION_SESSION_TOKENS][$token][$key]
                         ?? false;
                unset($_SESSION[static::OPTION_SESSION_TOKENS][$token][$key]);
                return $value;
            }
        }
        
        $value = $this->getParam($key);
        $this->removeParam($key);

        return $value;
    } // end popParam

    public function getSessionToken()
    {
        $tokenKey = $this->getOption(static::OPTION_SESSION_TOKEN_KEY);
        if (empty($_REQUEST[$tokenKey])) {
            return false;
        }
        
        return $_REQUEST[$tokenKey];
    } // end getSessionToken

    public function removeParam($key)
    {
        if (!isset($_SESSION[$key])) {
            return false;
        }

        unset($_SESSION[$key]);
        return true;
    }
    
    public function startSessionToken()
    {
        $key = $this->getOption(static::OPTION_SESSION_TOKEN_KEY);
        
        if (isset($_REQUEST[$key])) {
            return $_REQUEST[$key];
        }
        
        $token = bin2hex(random_bytes(32));
        
        if (!array_key_exists(static::SERVER_REQUEST_URI, $_SERVER)) {
            return false;
        }
        
        $url = $_SERVER[static::SERVER_REQUEST_URI];
        
        if (strpos($url, '?') !== false) {
            $url .= "&";
        } else {
            $url .= "?";
        }
        
        $url .= $key."=".$token;
        
        $this->redirect($url, false);
    } // end startSessionToken
    
    /**
     * Dispatch event
     *
     * @param mixed $eventType
     * @param mixed|null $data
     * @return array|bool
     */
    public function fireEvent($eventType, &$data = null)
    {
        $event = new FestiEvent($eventType, $data);
        return $this->dispatchEvent($event);
    } // end fireEvent
    
    public function isMobileRequest()
    {
        if (empty($_SERVER['HTTP_USER_AGENT'])) {
            return false;
        }
        
        $regExp = '#Mobile|iP(hone|od|ad)|Android|BlackBerry|IEMobile|Kindle|'.
              'NetFront|Silk-Accelerated|(hpw|web)OS|Fennec|Minimo|'.
              'Opera M(obi|ini)|Blazer|Dolfin|Dolphin|Skyfire|Zune#Umis';
              
        
        return preg_match($regExp, $_SERVER['HTTP_USER_AGENT']);
    } // end isMobileRequest
    
    public function getConfigValue($key)
    {
        if (!isset($this->config)) {
            throw new SystemException("Undefined config in Core");
        }
        
        if (array_key_exists($key, $this->config)) {
            return $this->config[$key];
        }
        
        return false;
    } // end getConfigValue
    
    public function setSystemPlugin(ISystemPlugin &$plugin)
    {
        $this->_systemPlugin = &$plugin;
    } // end setSystemPlugin

    /**
     * @return ISystemPlugin
     * @throws SystemException
     */
    public function &getSystemPlugin(): ISystemPlugin
    {
        if (!$this->_systemPlugin) {
            throw new SystemException("Undefined System Plugin.");
        }

        return $this->_systemPlugin;
    } // end getSystemPlugin

    public function isConsole()
    {
        return php_sapi_name() == 'cli';
    } // end isConsole
    
    public function setHooks(&$hooks)
    {
        $this->_hooks = $hooks;
    } // end setHooks
    
    public function addHookListener($name, $className)
    {
        if (array_key_exists($name, $this->_hooks)) {
            if (is_scalar($this->_hooks[$name])) {
                $this->_hooks[$name] = array($this->_hooks[$name]);
            }
            
            $this->_hooks[$name][] = $className;
        } else {
            $this->_hooks[$name] = $className;
        }
        
        return true;
    } // end addHookListener
    
    public function setHookListener($name, $className)
    {
        $this->_hooks[$name] = $className;
        
        return true;
    } // end setHookListener
    
    
    public function fireHook($name, &$data)
    {
        if (!array_key_exists($name, $this->_hooks)) {
            return false;
        }
        
        $listeners = $this->_hooks[$name];
        
        if (is_scalar($listeners) || is_callable($listeners)) {
            $listeners = array($listeners);
        }
        
        foreach ($listeners as $callback) {
            if (!is_callable($callback)) {
                $plugin = $this->getPluginInstance($callback);
                $callback = array(
                    &$plugin,
                    $name
                );
            }
            
            call_user_func_array($callback, array(&$data));
        }
    } // end fireHook
    
}
