<?php

define('PARAM_ARRAY', 100);
define('PARAM_STRING', 101);
define('PARAM_STRING_NULL', 104);
define('PARAM_FILE', 102);
define('PARAM_METHOD', 105);

class Entity extends EventDispatcher
{
    public const FIELD_TYPE_ARRAY           = 100;
    public const FIELD_TYPE_BOOL            = 111;
    public const FIELD_TYPE_FILE            = 102;
    public const FIELD_TYPE_FLOAT           = 109;
    public const FIELD_TYPE_INT             = 108;
    public const FIELD_TYPE_JSON            = 106;
    public const FIELD_TYPE_METHOD          = 105;
    public const FIELD_TYPE_OBJECT          = 107;
    public const FIELD_TYPE_SECURITY_STRING = 110;
    public const FIELD_TYPE_STRING          = 101;
    public const FIELD_TYPE_STRING_NULL     = 104;

    public const OPTION_KEY_DEFAULT     = 'default';
    public const OPTION_KEY_ERROR       = 'error';
    public const OPTION_KEY_FILTER      = 'filter';
    public const OPTION_KEY_OBJECT_TYPE = 'object_type';
    public const OPTION_KEY_REGEXP      = 'regexp';
    public const OPTION_KEY_REQUIRED    = 'required';
    public const OPTION_KEY_TYPE        = 'type';
    public const OPTION_KEY_RULE        = 'rule';


    /**
     * @deprecated
     */
    public const OPTION_DEFAULT  = self::OPTION_KEY_DEFAULT;
    /**
     * @deprecated
     */
    public const OPTION_ERROR = self::OPTION_KEY_ERROR;
    /**
     * @deprecated
     */
    public const OPTION_FILTER = self::OPTION_KEY_FILTER;
    /**
     * @deprecated
     */
    public const OPTION_REGEXP = self::OPTION_KEY_REGEXP;
    /**
     * @deprecated
     */
    public const OPTION_REQUIRED = self::OPTION_KEY_REQUIRED;
    /**
     * @deprecated
     */
    public const OPTION_TYPE = self::OPTION_KEY_TYPE;

    /**
     * Returns valid data values. All errors are written to the $errors.
     *
     * @param array $request
     * @param array $needles
     * @param array|null $errors
     * @return array
     */
    public function getPreparedData(array $request, array $needles, ?array &$errors = array()): array
    {
        $fields = array();

        foreach ($needles as $fieldName => $options) {
            $fields[$fieldName] = $this->_getDataItemValue(
                $fieldName,
                $request,
                $options,
                $errors
            );
        }

        return $fields;
    } // end getPreparedData

    /**
     * @param array $request
     * @param array $needles
     * @param array|null $errors
     * @return mixed
     */
    public function getExtendData(array $request, array $needles, ?array &$errors = array()): array
    {
        foreach ($needles as $fieldName => $options) {
            $request[$fieldName] = $this->_getDataItemValue(
                $fieldName,
                $request,
                $options,
                $errors
            );
        }

        return $request;
    } // end getExtendData


    /**
     * Returns default options for validation request field
     *
     * @param array|int $options
     * @return array
     */
    private function _getFieldOptions($options): array
    {
        $optionsAttributes = array(
            static::OPTION_KEY_REQUIRED,
            static::OPTION_KEY_REGEXP,
            static::OPTION_KEY_ERROR,
            static::OPTION_KEY_FILTER,
            static::OPTION_KEY_DEFAULT,
            static::OPTION_KEY_OBJECT_TYPE,
            static::OPTION_KEY_RULE,
        );

        if (is_numeric($options)) {
            $options = array(
                static::OPTION_KEY_TYPE => $options,
            );
        }

        if (!isset($options[static::OPTION_KEY_TYPE])) {
            $options[static::OPTION_KEY_TYPE] = self::FIELD_TYPE_STRING_NULL;
        }

        if (
            isset($options[static::OPTION_KEY_RULE]) &&
            !is_callable($options[static::OPTION_KEY_RULE])
        ) {
            unset($options[static::OPTION_KEY_RULE]);
        }

        foreach ($optionsAttributes as $attribute) {
            if (!isset($options[$attribute])) {
                $options[$attribute] = null;
            }
        }

        return $options;
    } // end _getFieldOptions

    /**
     * Returns field value by type
     *
     * @param string $name
     * @param int $type
     * @param array|null $request
     * @return mixed
     */
    private function _getDataItemValueByType(string $name, int $type, ?array &$request)
    {
        switch ($type) {
            case self::FIELD_TYPE_ARRAY:
                $value = $this->_getArrayTypeValue($name, $request);
                break;

            case self::FIELD_TYPE_FILE:
                $value = $this->_getFileTypeValue($name);
                break;

            case self::FIELD_TYPE_METHOD:
                $value = $this->_getMethodTypeValue($name, $request);
                break;

            case self::FIELD_TYPE_JSON:
                $value = isset($request[$name]) ? json_decode($request[$name], true) : null;
                break;

            case self::FIELD_TYPE_OBJECT:
                $value = $request[$name] ?? null;
                break;
                
            case self::FIELD_TYPE_SECURITY_STRING:
                $value = $this->_getSecurityStringTypeValue($name, $request);
                break;

            default:
                $value = $this->_getDefaultValue($name, $request, $type);
        }

        return $value;
    } // end _getDataItemValueByType

    /**
     * @param string $name
     * @param array $request
     * @return string
     */
    private function _getSecurityStringTypeValue(string $name, array $request): ?string
    {
        $value = null;
        if (isset($request[$name]) && $request[$name] !== '') {
            // TODO: use htmlspecialchars
            $value = filter_var($request[$name], FILTER_SANITIZE_STRING);
        }
        
        return $value;
    } // end _getSecurityStringType

    /**
     * @param string $name
     * @param array $request
     * @return mixed|null
     */
    private function _getMethodTypeValue(string $name, array $request)
    {
        $value = null;
        if (isset($request[$name]) && is_callable($request[$name])) {
            $value = $request[$name];
        }
        
        return $value;
    } // end _getMethodTypeValue

    /**
     * @param string $name
     * @return mixed|null
     */
    private function _getFileTypeValue(string $name)
    {
        $value = null;
        if ($this->_isFileUploadedSuccessfully($name)) {
            $value = $_FILES[$name];
        }
        
        return $value;
    } // end _getFileTypeValue

    /**
     * @param string $name
     * @return bool
     */
    private function _isFileUploadedSuccessfully(string $name): bool
    {
        return isset($_FILES[$name]) && $_FILES[$name][static::OPTION_KEY_ERROR] == UPLOAD_ERR_OK;
    }

    /**
     * @param string $name
     * @param array $request
     * @return array
     */
    private function _getArrayTypeValue(string $name, array $request): array
    {
        $result = array();
        if (isset($request[$name]) && is_array($request[$name])) {
            $result = $request[$name];
        }

        return $result;
    } // end _getArrayValue

    /**
     * @param string $name
     * @param array|null $request
     * @param int $type
     * @return mixed|null
     */
    private function _getDefaultValue(string $name, ?array $request, int $type)
    {
        if (!array_key_exists($name, $request) || !is_scalar($request[$name])) {
            return null;
        }

        $value = $request[$name];

        if ($type == static::FIELD_TYPE_INT) {
            $value = (int) $value;
        } else if ($type == static::FIELD_TYPE_FLOAT) {
            $value = (float) $value;
        } else if ($type == static::FIELD_TYPE_BOOL) {
            $value = filter_var(
                $value, 
                FILTER_VALIDATE_BOOLEAN, 
                FILTER_NULL_ON_FAILURE
            );
        } else if ($value === '') {
            $value = null;
        }

        return $value;
    } // end _getDefaultValue

    /**
     * Returns valid field value. If field invalid will be written error to
     * array $errors
     *
     * @param string $name
     * @param array $request
     * @param mixed $options
     * @param array|null $errors
     * @return mixed
     */
    private function _getDataItemValue(string $name, array $request, $options, ?array &$errors)
    {
        $options = $this->_getFieldOptions($options);

        $value = $this->_getDataItemValueByType($name, $options[static::OPTION_KEY_TYPE], $request);

        $hasError = $options[static::OPTION_KEY_REQUIRED] && is_null($value);

        $value = $this->_getFilterDataItemValue($value, $options, $hasError);

        $customError = null;
        if ($options[static::OPTION_KEY_RULE]) {
            $validator = $options[static::OPTION_KEY_RULE];
            $filterResult = $validator($value, $customError);
            if ($filterResult === false) {
                $hasError = true;
                $value = null;
            } else {
                $value = $filterResult;
            }
        }

        $this->_checkObjectType($value, $options, $hasError);

        if ($hasError) {
            $errors[$name] = $customError ?? $options[static::OPTION_KEY_ERROR] ?? false;
        }

        if (is_null($value) && isset($options[static::OPTION_KEY_DEFAULT])) {
            $value = $options[static::OPTION_KEY_DEFAULT];
        }

        return $value;
    } // end _getDataItemValue

    /**
     * @param $value
     * @param array $options
     * @param bool $hasError
     * @return mixed
     */
    private function _getFilterDataItemValue($value, array $options, bool &$hasError)
    {
        if ($options[static::OPTION_KEY_REGEXP] && !preg_match($options[static::OPTION_KEY_REGEXP], $value)) {
            $hasError = true;
        } else if ($options[static::OPTION_KEY_FILTER]) {
            $filterResult = filter_var($value, $options[static::OPTION_KEY_FILTER]);
            if ($filterResult === false) {
                $hasError = true;
            } else {
                $value = $filterResult;
            }
        }
        
        return $value;
    } // end _getFilterDataItemValue

    /**
     * @param $value
     * @param array $options
     * @param bool $hasError
     */
    private function _checkObjectType($value, array $options, bool &$hasError): void
    {
        $type       = $options[static::OPTION_KEY_TYPE];
        $objectType = $options[static::OPTION_KEY_OBJECT_TYPE];

        $isObject = $type === static::FIELD_TYPE_OBJECT && is_object($value);
        $hasObjectType = (bool) $objectType;

        if (!$isObject || !$hasObjectType) {
            return;
        }

        if (!is_string($objectType) || !($value instanceof $objectType)) {
            $hasError = true;
        }
    }

    /**
     * @param string|null $string
     * @param bool|array $values
     * @return string
     */
    public static function fillString(?string $string, $values = false): string
    {
        $in = array();
        $to = array();
        if ($values) {
            foreach ($values as $key => $value) {
                $in[] = '%'.$key.'%';
                $to[] = $value;
            }
        }

        // TODO: add session varibles

        if ($in) {
            return str_replace($in, $to, $string);
        }

        return $string;
    } // end fillString
}
