<?php

class MoDictionaryLocale implements IDictionaryLocale
{
    private const MAGIC_BIG    = 0xde120495;
    private const MAGIC_LITTLE = 0x950412de;

    private const ENDIAN_BIG    = 'N';
    private const ENDIAN_LITTLE = 'V';

    private const ENDIANS = array(
        self::MAGIC_BIG    => self::ENDIAN_BIG,
        self::MAGIC_LITTLE => self::ENDIAN_LITTLE,
    );

    private const ENCODING = 'ascii';

    private const WORD_SIZE = 4;

    private const META_KEY_PLURAL_FORMS = 'Plural-Forms';

    /** @var string */
    private $_moFilePath;

    /** @var array */
    private $_dictionaries = array();

    /** @var MoReader */
    private $_reader;

    /** @var string */
    private $_endian;

    /** @var int */
    private $_revision;

    /** @var int */
    private $_total;

    /** @var int */
    private $_originalsLengthsAddr;

    /** @var int */
    private $_translationsLengthsAddr;

    /** @var int */
    private $_hashLength;

    /** @var int */
    private $_hashAddr;

    public function __construct(string $file)
    {
        $this->_moFilePath = $file;
    }

    /**
     * @throws MoLocaleException
     */
    public function load(): void
    {
        $this->_loadReader();
        $this->_loadByteOrder();
        $this->_loadHeader();

        if ($this->_revision !== 0) {
            $message = 'Support revision 0 of MO format specs, only';
            throw new MoLocaleException($message);
        }

        // seek to data blocks
        $this->_reader->seekto($this->_originalsLengthsAddr);

        $originals    = $this->_getIndices($this->_originalsLengthsAddr, $this->_translationsLengthsAddr);
        $translations = $this->_getIndices($this->_translationsLengthsAddr, $this->_hashAddr);

        // transform raw data into set of indices
        $size = 2 * static::WORD_SIZE;
        $originals    = str_split($originals, $size);
        $translations = str_split($translations, $size);

        // skip hash table
        $stringsAddr = $this->_hashAddr + $this->_hashLength * static::WORD_SIZE;

        $this->_reader->seekto($stringsAddr);
        
        $strings = $this->_reader->readAll();
        $this->_reader->close();

        $this->_loadDictionary($originals, $translations, $stringsAddr, $strings);
    }

    /**
     * @throws MoLocaleException
     */
    private function _loadReader(): void
    {
        $this->_reader = new MoReader($this->_moFilePath);
        if (!$this->_reader->isResource()) {
            throw new MoLocaleException();
        }
    }

    /**
     * @return mixed
     * @throws MoLocaleException
     */
    private function _getMagic()
    {
        $bytes = $this->_reader->read(static::WORD_SIZE);
        if (mb_strlen($bytes, static::ENCODING) !== static::WORD_SIZE) {
            throw new MoLocaleException();
        }

        $result = unpack(static::ENDIAN_LITTLE, $bytes);
        if (!$result) {
            throw new MoLocaleException();
        }

        return array_shift($result);
    }

    /**
     * @throws MoLocaleException
     */
    private function _loadByteOrder(): void
    {
        $magic = $this->_getMagic();

        if (!array_key_exists($magic, static::ENDIANS)) {
            throw new MoLocaleException();
        }

        $this->_endian = static::ENDIANS[$magic];
    }

    /**
     * @throws MoLocaleException
     */
    private function _loadHeader(): void
    {
        $vars = array(
            'revision',
            'total',
            'originalsLengthsAddr',
            'translationsLengthsAddr',
            'hashLength',
            'hashAddr',
        );

        $size = count($vars) * static::WORD_SIZE;

        $header = $this->_reader->read($size);
        if (mb_strlen($header, static::ENCODING) != $size) {
            throw new MoLocaleException();
        }

        foreach ($vars as &$var) {
            $var = $this->_endian.$var;
        }

        $format = implode('/', $vars);

        $header = unpack($format, $header);
        if (!is_array($header)) {
            throw new MoLocaleException();
        }

        $this->_revision                = $header['revision'];
        $this->_total                   = $header['total'];
        $this->_originalsLengthsAddr    = $header['originalsLengthsAddr'];
        $this->_translationsLengthsAddr = $header['translationsLengthsAddr'];
        $this->_hashLength              = $header['hashLength'];
        $this->_hashAddr                = $header['hashAddr'];
    }

    /**
     * @param int $startAddr
     * @param int $endAddr
     * @return string
     * @throws MoLocaleException
     */
    private function _getIndices(int $startAddr, int $endAddr): string
    {
        $indicesLengths = $endAddr - $startAddr;
        $size = 2 * static::WORD_SIZE;
        if ($indicesLengths != $this->_total * $size) {
            throw new MoLocaleException();
        }

        $indices = $this->_reader->read($indicesLengths);
        if (mb_strlen($indices, static::ENCODING) != $indicesLengths) {
            throw new MoLocaleException();
        }

        return $indices;
    }

    /**
     * @param array $originals
     * @param array $translations
     * @param int $stringsAddr
     * @param string $strings
     * @throws MoLocaleException
     */
    private function _loadDictionary(
        array $originals, array $translations, int $stringsAddr, string $strings
    ): void
    {
        $format = "{$this->_endian}length/{$this->_endian}offset";

        for ($i = 0; $i < $this->_total; $i++) {
            $original    = unpack($format, $originals[$i]);
            $translation = unpack($format, $translations[$i]);
            if (!$original || !$translation) {
                throw new MoLocaleException();
            }

            // adjust offset due to reading strings to separate space before
            $original['offset']    -= $stringsAddr;
            $translation['offset'] -= $stringsAddr;

            $original    = mb_substr($strings, $original['offset'], $original['length'], static::ENCODING);
            $translation = mb_substr($strings, $translation['offset'], $translation['length'], static::ENCODING);

            $this->_dictionaries[$original] = $translation;
        }
    }

    public function get(string $key): ?string
    {
        return $this->_dictionaries[$key] ?? null;
    }
    
    public function getAll(): array
    {
        return $this->_dictionaries;
    }

    public function getMeta(): ?array
    {
        $meta = $this->get('');
        if (is_null($meta)) {
            return null;
        }

        $meta = explode("\n", $meta);

        foreach ($meta as $i => $item) {
            $index = mb_strpos($item, ':');
            if ($index !== false) {
                $key   = mb_substr($item, 0, $index);
                $value = trim(mb_substr($item, $index + 1));

                $meta[$key] = $value;
                unset($meta[$i]);
            }
        }

        return $meta;
    }

    public function getPluralForm(int $n): int
    {
        $meta = $this->getMeta();
        if (!isset($meta[static::META_KEY_PLURAL_FORMS])) {
            return 0;
        }

        $pluralForms = $meta[static::META_KEY_PLURAL_FORMS];

        $vars = array(
            'nplurals',
            'plural',
            'n',
        );

        foreach ($vars as &$var) {
            $var = "({$var})";
        }

        $pattern = '/'.implode('|', $vars).'/';
        $pluralForms = preg_replace($pattern, '\$$0', $pluralForms);

        eval($pluralForms);

        if (!isset($plural)) {
            return 0;
        }

        return $plural;
    }

    public function getPlural(string $idMessage, string $messagePlural, int $count): ?string
    {
        $separator = chr(0);
        $key = $idMessage.$separator.$messagePlural;

        $words = $this->get($key);
        if (is_null($words)) {
            return null;
        }

        $words = explode($separator, $words);

        $pluralForm = $this->getPluralForm($count);

        if ($pluralForm > count($words) - 1) {
            return array_shift($words);
        }

        return $words[$pluralForm];
    }
}
