<?php

use core\stage\StageException;
use core\store\IStore;

class StageBuilder
{
    public const REQUEST_KEY_BLOCK_INDEX = '_stageBlock';
    
    public const LAYOUT_ROWS = 'rows';
    public const LAYOUT_COLUMNS = 'columns';
    public const LAYOUT_MASONRY = 'masonry';

    private string $_layout;
    private $_view;
    
    private int $_maxIndex = 0;
    private array $_blocks;
    
    private int $_size;
    private int $_maxLength = 12;
    
    public function __construct(int $size = 3, ?string $layout = null, $view = false)
    {
        $this->_size = $size;
        $this->_layout = $layout ? $layout : static::LAYOUT_ROWS;
        
        if (!$view) {
            $path = Core::getInstance()->getOption(Core::OPTION_ENGINE_TEMPLATE_PATH);
            $view = new Template($path);
        }
        
        $this->_view = $view;
        $this->_blocks = array();
    }
    
    public function setSize(int $size): void
    {
        $this->_size = $size;
    }
    
    public function setLayout(string $layout): void
    {
        $this->_layout = $layout;
    }
    
    protected function getLayout(): string
    {
        return $this->_layout;
    }
    
    public function addBlock($instance, int $size = 1, ?string $className = null): int
    {
        if (!$className) {
            $className = $this->_getBlockClassName($instance);
        }
        
        $index = $this->_maxIndex;
        
        $this->_blocks[$index] = new $className(
            $instance, 
            $index, 
            $size, 
            $this->_view
        );
        
        $this->_maxIndex++;
        
        return $index;
    } // end addBlock
    
    private function _getBlockClassName($instance): string
    {
        if ($instance instanceof IStore) {
            $className = StoreStageBlock::class;
        } else if ($instance instanceof AbstractDisplayAction) {
            $className = StoreActionStageBlock::class;
        } else if (is_array($instance)) {
            if (!is_callable($instance)) {
                throw new StageException(
                    'Not found callback for stage block type.'
                );
            }

            $className = CallbackStageBlock::class;
        } else if (is_string($instance)) {
            $className = ContentStageBlock::class;
        } else {
            throw new StageException('Undefined stage block type.');
        }
        
        return $className;
    } // end _getBlockClassName

    public function addStoreBlock(IStore $store, int $size = 1): int
    {
        return $this->addBlock($store, $size, StoreStageBlock::class);
    } // end addStoreBlock
    
    public function addStoreActionBlock(AbstractDisplayAction $store, int $size = 1): int
    {
        return $this->addBlock($store, $size, StoreActionStageBlock::class);
    } // end addStoreActionBlock
    
    public function addCallbackBlock(
        callable $callback, ?array $externalParams = null, int $size = 1
    ): int
    {
        $index =  $this->addBlock($callback, $size, CallbackStageBlock::class);
        
        if ($externalParams) {
            $this->_blocks[$index]->setExternalParams($externalParams);
        }
        
        return $index;
    } // end addCallbackBlock
    
    public function addUrlBlock(string $url, int $size = 1): int
    {
        return $this->addBlock($url, $size, UrlStageBlock::class);
    } // end addUrlBlock
    
    public function onRequest(Response &$response)
    {
        if ($this->_isExecBlockRequest()) {

            $blockIndex = $this->_getExecutedBlockIndexFromRequest();
            $blockInstance = $this->_getBlockInstanceByIndex($blockIndex);

            return $blockInstance->exec($response);
        }

        $response->content = $this->fetchLayout();

        return true;
    } // end onRequest

    private function _getBlockInstanceByIndex(int $index): StageBlock
    {
        if (!array_key_exists($index, $this->_blocks)) {
            throw new StageException(__('Undefined stage block with index: %d', $index));
        }

        return $this->_blocks[$index];
    } // end _getBlockInstanceByIndex
    
    protected function fetchLayout(): string
    {
        $defaultLength = $this->_maxLength / $this->_size;
        
        $grid = array();
        
        $rowIndex = 0;
        $columnIndex = 0;
        
        $currentLength = 0;
        
        foreach ($this->_blocks as $index => $block) {
        
            $blockSize = $block->getSize();
            $lenght = $blockSize * $defaultLength;
        
            if ($currentLength >= $this->_maxLength) {
                $rowIndex++;
                $columnIndex = 0;
                $currentLength = 0;
            }
        
            $grid[$rowIndex][$columnIndex]['length'] = $lenght;
            $grid[$rowIndex][$columnIndex]['content'] = $block->exec();
        
            $currentLength += $lenght;
            $columnIndex++;
        }
        
        $this->_view->grid = $grid;
        $this->_view->defaultLength = $defaultLength;
        $this->_view->size = $this->_size;
        
        $templateName = 'stage/layout_'.$this->getLayout();
        
        return $this->_view->fetch($templateName.'.phtml');
    } // end fetchView
    
    
    private function _getExecutedBlockIndexFromRequest() : ?int
    {
        if (!array_key_exists(StageBuilder::REQUEST_KEY_BLOCK_INDEX, $_REQUEST)) {
            return null;
        }

        return (int) $_REQUEST[StageBuilder::REQUEST_KEY_BLOCK_INDEX];
    } // end _getExecutedBlockIndexFromRequest
    
    private function _isExecBlockRequest(): bool
    {
        $blockIndex = $this->_getExecutedBlockIndexFromRequest();

        if (is_null($blockIndex)) {
            return false;
        }

        if (Response::isAjaxRequest()) {
            return true;
        }

        $blockInstance = $this->_getBlockInstanceByIndex($blockIndex);

        return $blockInstance->isExec();
    } // end _isExecBlockRequest
}
