<?php

class ForeignKeyLoadAction extends AbstractAction
{
    public const REQUEST_PARAM_AJAX_CHILD        = 'ajaxChild';
    public const REQUEST_PARAM_AJAX_CHILD_VALUES = 'ajaxChildValues';
    public const REQUEST_PARAM_AJAX_PARENT       = 'ajaxParent';
    public const REQUEST_PARAM_FIELD_NAME        = 'fieldName';
    public const REQUEST_PARAM_TERM              = 'term';
    public const REQUEST_PARAM_VALUE             = 'value';
    public const REQUEST_PARAM_EXCLUDE           = 'exclude';

    private $_values;

    /**
     * @param Response $response
     * @return bool
     * @throws SystemException
     */
    public function onStart(Response &$response): bool
    {
        $data = $this->getDataFromRequest();

        if ($data[static::REQUEST_PARAM_AJAX_CHILD] && $data[static::REQUEST_PARAM_AJAX_PARENT]) {
            return $this->_onStartLoadChildFieldValues($response, $data);
        } else if ($data[static::REQUEST_PARAM_FIELD_NAME]) {
            return $this->_onStartLoadFieldValues($response, $data);
        }

        throw new StoreActionException("Undefined request", $this);
    } // end onStart

    /**
     * @param Response $response
     * @param array $data
     * @return bool
     */
    private function _onStartLoadFieldValues(Response &$response, array $data): bool
    {
        // TODO: We have to make refactoring of field name's
        $fieldName = preg_replace("#^__#Umis", "", $data[static::REQUEST_PARAM_FIELD_NAME]);
        $field = $this->model->getFieldByName($fieldName);

        if (!$field) {
            throw new StoreActionException(
                "Not found field with name: ".$data[static::REQUEST_PARAM_FIELD_NAME],
                $this
            );
        }

        if ($field instanceof ForeignKeyField) {
            $options = $field->getAttributes();
            $options['foreignLimit'] = $options['autocompleteLimit'];

        } else {
            $options = array(
                'foreignKeyField'   => 'keyValue',
                'foreignValueField' => $field->getName(),
                'foreignTable'      => $field->getStoreName(),
                'unique'            => true
            );
        }

        $term    = $data[static::REQUEST_PARAM_TERM];
        $value   = $data[static::REQUEST_PARAM_VALUE];
        $exclude = $data[static::REQUEST_PARAM_EXCLUDE];

        if (!empty($term)) {
            $store = $this->getStore();

            $options['search'] = $store->getProxy()->getLikeCondition(
                $options['foreignValueField'],
                '%'.$term.'%'
            );
        } else if (!empty($value)) {
            $fieldKeyName = $options['foreignKeyField'].'&IN';

            $options['search'] = array(
                $fieldKeyName => explode(',', $value)
            );

            unset($options['foreignLimit']);
        }

        if (!empty($exclude)) {
            $fieldKeyName = $options['foreignKeyField'].'&NOT IN';

            $options['search'][$fieldKeyName] = explode(',', $exclude);
        }

        $this->_values = $this->getStore()->getProxy()
            ->loadForeignValues($options);

        $results = array();
        foreach ($this->_values as $key => $value) {
            $results[] = array(
                'key'   => $value,
                'value' => $key
            );
        }

        // TODO: Double check
        $response->setType(Response::JSON);

        $response->results = $results;

        return true;
    } // end _onStartLoadFieldValues

    /**
     * @param Response $response
     * @param array $data
     * @return bool
     */
    private function _onStartLoadChildFieldValues(Response &$response, array $data): bool
    {
        $view = $this->store->getView();
        $view->info = $data;

        if (!array_key_exists(static::REQUEST_PARAM_VALUE, $_REQUEST)) {
            throw new StoreActionException("Undefined value param");
        }

        $data[static::REQUEST_PARAM_VALUE] = $_REQUEST[static::REQUEST_PARAM_VALUE];

        $parentField = $this->model->getFieldByName($data[static::REQUEST_PARAM_AJAX_PARENT]);

        $childsFields = explode('|', $data[static::REQUEST_PARAM_AJAX_CHILD]);

        $childFieldsValues = $this->store->getRequestParam(static::REQUEST_PARAM_AJAX_CHILD_VALUES);

        $content = '';
        foreach ($childsFields as $fieldName) {
            $childField = $this->model->getFieldByName($fieldName);

            $options = $childField->getAttributes();

            if (!empty($options['ajaxParentField'])) {
                $options['ajaxParentColumn'] = $options['ajaxParentField'];
            }

            $options['ajaxParentValue'] = $data[static::REQUEST_PARAM_VALUE];

            $this->_values = $this->store->getProxy()->loadForeignValues($options);

            $view->values = $this->_values;

            $childrenValue = null;
            if (is_array($childFieldsValues) && array_key_exists($fieldName, $childFieldsValues)) {
                $childrenValue = $childFieldsValues[$fieldName];
            }

            $view->childValue = $childrenValue;
            $view->fieldName = $fieldName;

            // FIXME: actions/foreign_key_load.php
            $content .= $view->fetch('actions/foreign_key_load.php');
        }

        $response->content = $content;
        $response->flush(true);

        return true;
    } // end _onStartChildValuesLoad

    /**
     * Returns values.
     *
     * @return array|null
     */
    public function getValues(): ?array
    {
        return $this->_values;
    } // end getValues

    /**
     * @override
     * @return array|null
     */
    protected function getRequestFields(): ?array
    {
        return array(
            static::REQUEST_PARAM_AJAX_CHILD  => static::FIELD_TYPE_STRING_NULL,
            static::REQUEST_PARAM_AJAX_PARENT => static::FIELD_TYPE_STRING_NULL,
            static::REQUEST_PARAM_FIELD_NAME  => static::FIELD_TYPE_STRING_NULL,
            static::REQUEST_PARAM_TERM        => static::FIELD_TYPE_STRING_NULL,
            static::REQUEST_PARAM_VALUE       => static::FIELD_TYPE_STRING_NULL,
            static::REQUEST_PARAM_EXCLUDE     => static::FIELD_TYPE_STRING_NULL,
        );
    }

    /**
     * @override
     */
    public function getActionName(): string
    {
        return Store::ACTION_FOREIGN_KEY_LOAD;
    }

    /**
     * @override
     */
    public function isExec()
    {
        return true;
    } // end isExec
    
}
