<?php

use core\store\field\IStoreForeignField;

class RelationAction extends AbstractAction
{
    private string $_relationStoreName;
    private string $_type;

    protected $session;

    /**
     * @param Response $response
     * @return bool
     * @throws SystemException
     */
    public function onStart(Response &$response): bool
    {
        $this->session = &$this->store->getSession();

        $data = $this->getDataFromRequest();

        $relation = $this->_getRelation($data);

        $relationStore = new Store(
            $this->store->getConnection(),
            $relation['foreignTable'],
            $this->store->getOptions()
        );

        $this->_relationStoreName = $relationStore->getIdent();
        $this->_type = $relation['type'];

        if ($relation['type'] == Store::RELATION_TYPE_CHILD) {
            $this->_doChildRelation($relation);
        } else if ($relation['type'] == Store::RELATION_TYPE_PARENT) {
            $this->_doParentRelation($relation);
        }

        // FIXME:
        $params = array();
        if (Response::isAjaxRequest()) {
            $params[Response::REQUEST_KEY_AJAX_MODE] = 'true';
        }

        $storeName = $relationStore->getIdent();

        $options = $relationStore->getOption('handler_options');

        $url = $this->_getRedirectUrl($storeName, $relation, $options);

        $redirectUrl = Core::getInstance()->getUrl($url, $params);
        $response->url = $redirectUrl;
        $response->setAction(Response::ACTION_REDIRECT);

        return true;
    } // end onStart

    private function _getRedirectUrl(string $storeName, array $relation, array $options): string
    {
        if (empty($relation['url'])) {
            $currentUrl = $this->store->getOption('current_url');

            // TODO: Move to config system url prefix
            $regExp = "#^/festi/#Umis";
            if (preg_match($regExp, $currentUrl)) {

                $url = "/festi/" . $storeName . "/";
                if (!empty($options['plugin'])) {
                    $url .= $options['plugin'] . "/";
                }

            } else if ($storeName == $this->store->getIdent()) { // XXX: The child is the same DGS
                $url = $this->getUrl();
            } else {
                $msg = __l(
                    'Relation "%s" has an error: atrribute "url" is undefined into the relation %s.',
                    $this->_type,
                    $this->_relationStoreName
                );

                throw new StoreActionException($msg);
            }
            
        } else {
            $url = $relation['url'];
        }

        $primaryKeyValue = $this->store->getPrimaryKeyValueFromRequest();
        $vars = array(
            Store::PRIMARY_KEY_IN_REQUEST => $primaryKeyValue
        );

        $url = self::fillString($url, $vars);

        return $url;
    } // end _getRedirectUrl

    private function _doParentRelation(array $relation): bool
    {
        if ($this->store->hasSessionValue(Store::CAPTION_IN_SESSION)) {
            $caption = $this->store->getSessionValue(Store::CAPTION_IN_SESSION);
            $pos = strrpos($caption, ' / ');

            if ($pos === false) {
                $caption = '';
            } else {
                $caption = substr($caption, 0, $pos);
            }

            $this->store->setSessionValue(Store::CAPTION_IN_SESSION, $caption, $this->_relationStoreName);
        }
        
        $parentValue = null;
        if ($this->_isParentRelation($relation)) {
            $primaryKeyValue = $this->store->getSessionValue(
                Store::PARENT_PRIMARY_KEY_IN_SESSION
            );
            $data = $this->store->loadRowByPrimaryKey($primaryKeyValue);
            
            $field = $this->model->getField($relation['field']);
            if ($field && $field instanceof IStoreForeignField) {
                $parentValueKey = '_foreign_'.$relation['field'];
            } else {
                $parentValueKey = $relation['field'];
            }

            if (isset($data[$parentValueKey])) {
                $parentValue = $data[$parentValueKey];
            }

        }

        $this->store->setSessionValue(
            Store::PARENT_PRIMARY_KEY_IN_SESSION,
            $parentValue
        );

        return true;
    } // end _doParentRelation

    private function _isParentRelation(array $relation): bool
    {
        $hasPrimaryKeyInSession = $this->store->hasSessionValue(
            Store::PARENT_PRIMARY_KEY_IN_SESSION
        );
        return $hasPrimaryKeyInSession && 
               $this->store->getIdent() == $relation['foreignTable'];
    }

    private function _doChildRelation(array $relation): bool
    {
        $id = $this->store->getPrimaryKeyValueFromRequest();

        if (!$id) {
            return false;
        }
        
        $rowData = $this->store->loadRowByPrimaryKey($id);

        $this->_setParentTableRowForChild(
            $this->_relationStoreName,
            $rowData
        );

        if (!array_key_exists($relation['field'], $rowData)) {

            $msg = __l(
                'Relation "%s" has an error: field "%s" is undefined into "%s". '.
                'You have to use a field defined in a fields section.',
                $this->_type,
                $relation['field'],
                $this->_relationStoreName
            );

            throw new StoreActionException($msg);
        }

        // FIXME: If open two tab with table in browser and modify we have bug
        $this->store->setSessionValue(
            Store::PARENT_PRIMARY_KEY_IN_SESSION,
            $rowData[$relation['field']],
            $this->_relationStoreName
        );

        $captionValue = null;
        if (!empty($relation['treeCaption'])) {
            $field = $this->model->getField($relation['treeCaption']);

            if (!$field) {
                $tableName = $this->store->getIdent();
                throw new SystemException(
                    __l('Not found column "%s" in "%s" for treeCaption', $relation['treeCaption'], $tableName)
                );
            }

            if ($field instanceof IStoreForeignField) {
                $this->store->loadForeignKeyValues($field);
                $caption = $field->getValuesList()[$rowData[$relation['treeCaption']]];
            } else {
                $caption = $rowData[$relation['treeCaption']];
            }

            if ($this->store->hasSessionValue(Store::CAPTION_IN_SESSION)) {
                $mainCaption = $this->store->getSessionValue(Store::CAPTION_IN_SESSION);
                $caption = $mainCaption.' / '.$caption;
            }
            
            $captionValue = $caption;
        }

        $this->store->setSessionValue(Store::CAPTION_IN_SESSION, $captionValue, $this->_relationStoreName);

        return true;
    } // end _doChildRelation

    /**
     * Save parent table row values
     *
     * @param string $childTableName
     * @param array $values
     */
    private function _setParentTableRowForChild(string $childTableName, array $values): void
    {
        $this->store->setSessionValue(
            Store::PARENT_ROW_VALUES_IN_SESSION, 
            $values, 
            $childTableName
        );
    }

    private function _getRelation(array $data): array
    {
        $action = $this->model->getAction($data['action']);

        if (!$action) {
            $msg = __l(
                'Relation has an error: action "%s" is undefined into the actions section.',
                $data['action']
            );

            throw new StoreActionException($msg);
        }

        $relations = $this->model->getRelation($action['type']);

        if (!$relations) {
            $relations = $this->model->getRelation($action['relationType']);
        }

        if (!$relations) {
            $msg = 'Not found relations for action '.$action['type'];
            throw new StoreActionException($msg);
        }

        if (!isset($relations[$action['relation']])) {
            $msg = 'Not found relation '.$action['relation'];
            throw new StoreActionException($msg);
        }

        $relation = $relations[$action['relation']];

        if (!is_null($action['link'])) {
            $relation['url'] = $action['link'];
        }

        return $relation;
    } // end _getRelation

    protected function getRequestFields(): ?array 
    {
        $requestFields = array(
           'action' => array(
                'type'     => self::FIELD_TYPE_STRING_NULL,
                'required' => true
            )
        );

        return $requestFields;
    } // end getRequestFields
    
    /**
     * @override
     */
    public function getActionName(): string
    {
        return Store::ACTION_RELATION;
    }

    /**
     * @override
     */
    public function isExec(): bool
    {
        return true;
    }
}
