<?php

use core\store\event\StoreFieldEvent;
use core\store\event\UrlStoreActionEvent;
use core\store\field\IStoreField;

/**
 * Class AbstractField
 */
class AbstractField extends Template implements IStoreField
{
    public const OPTION_AUTOCOMPLETE            = 'autocomplete';
    public const OPTION_AUTOCOMPLETE_MIN_LENGTH = 'autocompleteMinLength';
    public const OPTION_AUTOCOMPLETE_LIMIT      = 'autocompleteLimit';
    public const OPTION_CONTAINER_CSS           = 'containerCss';
    public const OPTION_EXPRESSION              = 'expression';
    public const OPTION_FILTER                  = 'filter';
    public const OPTION_REQUIRED                = 'required';
    public const OPTION_STORE                   = 'store';
    public const OPTION_WHERE                   = 'where';
    public const OPTION_PERMISSION              = 'permission';

    public const FILTER_TYPE_RANGE    = 'range';
    public const FILTER_TYPE_SELECT   = 'select';
    public const FILTER_TYPE_MULTIPLE = 'multiple';
    public const FILTER_TYPE_EXACT    = 'exact';
    public const FILTER_TYPE_TEXT     = 'text';


    /**
     * Reference to store.
     *
     * @var Store
     */
    protected Store $store;
    
    /**
     * Reference to current action.
     * 
     * @var AbstractAction
     */
    protected $action;

    protected array $attributes = array();

    private string $_name = '';

    private string $_selectorPostfix = '';
    
    protected ?int $index = null;
    
    protected $itemValue;
    
    protected $filterValues = null;
    
    public ?string $lastErrorMessage = null;

    /**
     * AbstractField constructor.
     * @param Store $store
     * @throws SystemException
     */
    public function __construct(Store &$store)
    {
        $this->store = &$store;

        parent::__construct($this->getTemplatePath());
    } // end __construct

    /**
     * Set reference to current DGS action.
     * 
     * @param AbstractAction $action
     */
    public function setAction(AbstractAction &$action): void
    {
        $this->action = &$action;
    } // end setAction
    
    /**
     * Returns reference to current DGS action.
     * 
     * @return AbstractAction
     */
    public function &getAction(): AbstractAction
    {
        return $this->action;
    } // end getAction
    
    /**
     * @return Store
     */
    public function &getStore(): Store
    {
        return $this->store;
    } // end getStore
    
    /**
     * Returns path to template folders.
     *
     * @return string
     * @throws SystemException
     */
    protected function getTemplatePath(): string
    {
        return $this->getBaseTemplatePath();
    } // end getTemplatePath

    /**
     * Returns path to base template folders.
     *
     * @return string
     * @throws SystemException
     */
    protected function getBaseTemplatePath(): string
    {
        return Core::getInstance()->getOption(Core::OPTION_ENGINE_TEMPLATE_PATH).'fields'.DIRECTORY_SEPARATOR;
    } // end getBaseTemplatePath
    
    /**
     * Returns encrypted value.
     * @param $value
     * @return string
     * @throws SystemException
     */
    public function getEncryptValue($value): string
    {
        $publicKey = $this->store->getPublicSslKey();
        
        $cryptedValue = null;
        openssl_public_encrypt(
            $value,
            $cryptedValue,
            $publicKey
        );
        
        $value = base64_encode($cryptedValue);
        
        return $value;
    } // end getEncryptValue

    /**
     * Returns decrypted value.
     *
     * @param string $value
     * @return string
     * @throws SystemException
     */
    public function getDecryptedValue(string $value): ?string
    {
        $privateKey = $this->store->getPrivateSslKey();

        $cryptedValue = base64_decode($value);

        $decryptedValue = null;
        $res = openssl_private_decrypt(
            $cryptedValue,
            $decryptedValue,
            $privateKey
        );

        if ($res === false) {
            throw new FieldException(openssl_error_string());
        }

        return $decryptedValue;
    } // end getDecryptedValue
    
    /**
     * @param $value
     * @return string
     * @throws SystemException
     */
    protected function getViewValue($value)
    {
        if ($value && $this->get('crypt')) {
            $value = $this->getDecryptedValue($value);
        }
        
        return $value;
    } // end getViewValue

    /**
     * Init field attributes
     *
     * @param FieldModel $scheme
     * @throws SystemException
     */
    public function onInit(FieldModel $scheme): void
    {
        $fields = $this->getAttributesFields();

        $attributes = $scheme->getAttributes();
        $this->attributes = $this->getExtendData($attributes, $fields, $errors);
        
        if ($errors) {
            $message = array_shift($errors);
            throw new SystemException($message);
        }

        $this->_name = $this->get('name') ?? '';

        if ($this->_name && StoreProxy::isComplexField($this->_name)) {
            $pieces = explode('as', $this->_name);
            list($aggregateExpression, $fieldName) = array_map('trim', $pieces);
            $this->_name = $fieldName;
            $this->set(static::OPTION_EXPRESSION, $aggregateExpression);
        }

        $this->_selectorPostfix = str_replace(
            array('[', ']'),
            array('-', ''),
            $this->_name
        );
    } // end onInit

    /**
     * Returns list of attributes for field.
     *
     * @return array
     */
    protected function getAttributesFields(): array
    {
        $fields = array(
            'caption' => array(
                'type'     => static::FIELD_TYPE_STRING,
                'error'    => 'Undefined caption in field',
                'required' => true
            ),
            'type' => array(
                'type'     => static::FIELD_TYPE_STRING,
                'error'    => 'Undefined type in field',
                'required' => true
            ),
            'name' => array(
                'type'     => static::FIELD_TYPE_STRING,
                'error'    => 'Undefined name in field',
                'required' => true
            ),

            'link'       => static::FIELD_TYPE_STRING_NULL,
            'clicable'   => static::FIELD_TYPE_STRING_NULL,
            'width'      => static::FIELD_TYPE_STRING_NULL,
            'filter'     => static::FIELD_TYPE_STRING_NULL,
            'sorting'    => static::FIELD_TYPE_STRING_NULL,
            'table'      => static::FIELD_TYPE_STRING_NULL,
            static::OPTION_REQUIRED => static::FIELD_TYPE_STRING_NULL,
            'format'     => static::FIELD_TYPE_STRING_NULL,
            'hide'       => static::FIELD_TYPE_STRING_NULL,
            'allowEmpty' => static::FIELD_TYPE_BOOL,
            static::OPTION_ONLY_LIST   => static::FIELD_TYPE_BOOL,
            static::OPTION_CONTAINER_CSS => static::FIELD_TYPE_STRING,
            'onValidateValue' => static::FIELD_TYPE_STRING_NULL,

            static::OPTION_AUTOCOMPLETE => static::FIELD_TYPE_STRING_NULL,
            static::OPTION_AUTOCOMPLETE_MIN_LENGTH => array(
                'type'    => static::FIELD_TYPE_INT,
                'default' => 2
            ),
            static::OPTION_AUTOCOMPLETE_LIMIT => array(
                'type'    => static::FIELD_TYPE_INT,
                'default' => 10
            )
        );

        return $fields;
    } // end getAttributesFields

    /**
     * Returns attribute value or false if that not found.
     *
     * @param string $name
     * @return mixed
     */
    public function get(string $name)
    {
        $result = false;
        if (array_key_exists($name, $this->attributes)) {
            $value = $this->attributes[$name];
            if ($value === 'true') {
                $result = true;
            } else if ($value !== 'false') {
                $result = $value;
            }
        }

        return $result;
    } // end get
    
    /**
     * @param $key
     * @param $value
     */
    public function set(string $key, $value): void
    {
        $this->attributes[$key] = $value;
    } // end set

    public function getAttributes(): array
    {
        return $this->attributes;
    } // end getAttributes
    
    /**
     * @return string
     */
    public function getName(): string
    {
        return $this->_name;
    } // end getName

    public function getSelectorPostfix(): string
    {
        return $this->_selectorPostfix;
    }

    public function setName(string $name): void
    {
        $this->_name = $name;
    } // end getName

    public function getType(): string
    {
        return $this->get('type');
    } // end getType
    
    public function setIndex(int $index): void
    {
        $this->index = $index;
    } // end setIndex
    
    public function getIndex(): int
    {
        return $this->index;
    } // end getIndex
    
    // FIXME:
    /**
     * @param $value
     * @return string
     */
    public function getFilter($value): string
    {
        return "= '".$value."'";
    }

    public function getEditInput(?string $value = '', $inline = null): ?string
    {
        $value = $this->getViewValue($value);
        $value = !is_null($value) ? htmlspecialchars($value) : '';

        $data = array(
            'field'   => &$this,
            'value'   => &$value,
            'content' => null
        );

        $event = new StoreFieldEvent(
            StoreFieldEvent::EVENT_ON_GET_EDIT_VIEW,
            $data
        );
        $this->dispatchEvent($event);

        if (!$data['content']) {
            $this->assign('value', $data['value']);
            $data['content'] = $this->fetch('edit.phtml');
        }

        return $data['content'];
    } // end getEditInput

    public function isVirtualField(?string $actionName = null): bool
    {
        return false;
    } // end isVirtualField

    public function isCustom(?string $section = null): bool
    {
        return $this->get('isCustom') === true || $this->get('isCustom') === 'true';
    } // end isCustom

    /**
     * Returns true if field is updated store logic.
     *
     * @return boolean
     */
    public function isEditable(): bool
    {
        return !$this->isVirtualField();
    } // end isVirtualField
    
    /**
     * @return bool
     */
    public function isForeignKey(): bool
    {
        return false;
    }

    public function displayValue(?string $value, array $row = array()): ?string
    {
        return $this->getDefaultCellValue($value, $row);
    } // end displayValue
    
    /**
     * @param string|null $value
     * @param array $row
     * @return string
     * @throws SystemException
     */
    protected function getDefaultCellValue(
        ?string $value, array $row = array()
    ): string
    {
        if ($value && $this->get('crypt')) {
            $value = $this->getDecryptedValue($value);
        }

        if (
            isset($this->attributes['trim']) && 
            ($this->attributes['trim'] != 'false')
        ) {
            $trimValue = $this->get('trim');
            $trimMode = $this->get('trimMode');
            // need to trim
            if (is_numeric($trimValue) || $trimValue == "button") {
                $key = $this->store->getPrimaryKey();

                $this->assign('trimValue', $trimValue);
                $this->assign('trimMode', $trimMode);
                $this->assign('value', $value);
                $this->assign('primaryKeyValue', $row[$key]);
        
                $value = $this->fetch('cell_value_trim.phtml');
            } elseif ($this->attributes['trim'] == 'link') {
                if (strlen($value) > 36) {
                    $value = '<a href="'.$value.'" target="_blank">'.
                    htmlspecialchars(substr($value, 0, 32)).'...'.
                    substr($value, -4).'</a>';
                } else {
                    $value = '<a href="'.$value.'" target="_blank">'.
                    htmlspecialchars($value).'</a>';
                }
            }
        }
        
        $value = $this->getFormattedValue($value);

        $this->assign('rowValues', $row);
        $this->assign('currentValue', $value);

        $value = $this->fetch('cell_value.phtml');
        
        return $value;
    } // end getDefaultCellValue

    /**
     * Returns formatted filter value.
     * Formatted value uses for display to a user.
     *
     * @param string|array|null $value
     * @return string|null
     * @throws FieldException
     */
    public function getFormattedFilterValue($value): ?string
    {
        $filterType = $this->getFilterType();
    
        if (!$value || !$filterType) {
            return null;
        }
    
        if ($filterType != static::FILTER_TYPE_RANGE) {
            
            if (is_array($value)) {
                return join(', ', $value);
            }
            
            return $this->getFormattedValue($value);
        }
    
        if (!is_array($value)) {
            throw new FieldException('Range filter value has type error.');
        }
    
        if (count($value) == 1) {
            if (array_key_exists(0, $value)) {
                $operation = ' >= ';
                $value = $value[0];
            } else if (array_key_exists(1, $value)) {
                $operation = ' <= ';
                $value = $value[1];
            } else {
                throw new FieldException(
                    __('Range filter value has structure error.')
                );
            }
    
            return $operation.' '.$this->getFormattedValue($value);
        }
    
        return __l(
            'from %s to %s',
            $this->getFormattedValue($value[0]),
            $this->getFormattedValue($value[1])
        );
    } // end getFormattedFilterValue
    
    /**
     * @param $value
     * @return string|null
     */
    public function getFormattedValue($value): ?string
    {
        $format = $this->get('format');
        if (!$format) {
            return $value;
        }

        $methodName = 'getFormatValue'.ucfirst($format);

        if (is_callable(array($this, $methodName))) {
            return $this->$methodName($value);
        }

        if (is_null($value)) {
            return $value;
        }

        return sprintf($format, $value);
    } // end getFormattedValue

    // FIXME:
    /**
     * @param $value
     * @return string
     */
    public function displayRow(string $value): string
    {
        return '<input type="text" readonly disabled name="'.
                $this->_name.'" value="'.$value.'" class="'.
                $this->getCssClassName().'">';
    }

    /**
     * @param string $value
     * @return string
     * @throws SystemException
     */
    public function displayRO($value): ?string
    {
        // FIXME:
        if (empty($value)) {
            $attrValue = $this->get('value');

            if ($attrValue !== false) {
                $value = $attrValue;
            }
        } else {
            $value = $this->getViewValue($value);
        }
        
        $format = $this->get('format');
        if ($format) {
            $methodName = 'getFormatValue'.ucfirst($format);
            
            if (method_exists($this, $methodName)) {
                $value = $this->$methodName($value);
            } else {
                $value = $this->getFormatValue($value, $format);
            }
        }

        return !is_null($value) ? nl2br($value) : '';
    }

    /**
     * @param string $value
     * @param array $row
     * @return string|null
     */
    public function getInfoValue($value, $row = null): ?string
    {
        return $this->displayRO($value);
    }
    
    /**
     * @param mixed $value
     * @param string $format
     * @return string
     */
    protected function getFormatValue($value, string $format): string
    {
        return sprintf($format, $value);
    }

    /**
     * @param $value
     * @return string|null
     */
    protected function getFormatValuePrice($value): ?string
    {
        if (is_null($value) || !is_numeric($value)) {
            return $value;
        }

        $locale = $this->get('locale');
        if (!$locale) {
            $locale = 'en_US.UTF-8';
        }
        setlocale(LC_MONETARY, $locale);

        $currency = $this->get('currency');
        if (!$currency) {
            $currency = 'USD';
        }

        if (class_exists('NumberFormatter')) {
            $formatter = new NumberFormatter($locale, NumberFormatter::CURRENCY);
            return $formatter->formatCurrency($value, $currency);
        }

        if (function_exists('money_format')) {
            return money_format('%.2n', $value);
        }

        return number_format($value, 2).' '.$currency;
    } // end getFormatValuePrice
    
    /**
     * @param $value
     * @return string
     */
    protected function getFormatValuePercent($value)
    {
        if (is_null($value)) {
            return $value;
        }

        // Move $,2 to attr
        return number_format($value, 2).'%';
    } // end getFormatValuePercent
    
    /**
     * @param $value
     * @return string
     */
    protected function getFormatValueNumber($value)
    {
        if (is_null($value)) {
            return $value;
        }

        $decimals = (int) $this->get('decimals');
    
        return number_format($value, $decimals);
    } // end getFormatValueNumber

    /**
     * @param array $request
     * @return bool|string|mixed|null
     * @throws \SystemException
     */
    public function getValue(array $request = array())
    {
        $value = array_key_exists($this->_name, $request) 
            ? $request[$this->_name] 
            : '';

        if (!$value && $this->get('isnull')) {
            $value = null;
        }

        if (!$this->isValidValue($value)) {
            return false;
        }
        
        if ($this->get('crypt')) {
            $value = $this->getEncryptValue($value);
        }
        
        if ($value) {
            // XXX: XSS Protection
            $this->onPrepareValue($value);
        }

        return $value;
    } // end getValue
    
    /**
     * @return mixed
     * @throws SystemException
     */
    protected function doEventCallback(): bool
    {
        $params = func_get_args();
        $callback = array_shift($params);

        if (!is_string($callback)) {
            throw new SystemException('Undefined callback type.');
        }

        return $this->_doEventCallbackByString($callback, $params);
    } // end doEventCallback
    
    /**
     * @param $callback
     * @param $params
     * @return mixed
     * @throws SystemException
     */
    private function _doEventCallbackByString($callback, array $params): bool
    {
        $data = explode('::', $callback);
        if (empty($data[0])) {
            throw new SystemException('Undefined plugin name for callback.');
        }

        if (empty($data[1])) {
            throw new SystemException('Undefined plugin method for callback.');
        }

        $plugin = Core::getInstance()->getPluginInstance($data[0]);

        $method = array(
            $plugin,
            $data[1]
        );

        if (!is_callable($method)) {
            throw new SystemException('Undefined callback: '.$callback);
        }

        // XXX: We could NOT use array_merge, array_unshift,
        // etc to be compatibility with reference processing PHP 7+
        $callbackParams = array(&$this);
        foreach ($params as $param) {
            $callbackParams[] = $param;
        }

        return call_user_func_array($method, $callbackParams);
    } // end _doEventCallbackByString
    
    /**
     * @param $value
     * @return bool
     * @throws SystemException
     */
    public function isValidValue($value): bool
    {
        $onValidateValue = $this->get('onValidateValue');
        if ($onValidateValue) {
            return $this->doEventCallback($onValidateValue, $value);
        }

        $isRequired = $this->get(static::OPTION_REQUIRED);
        if ($isRequired && empty($value)) {
            $this->lastErrorMessage = __l(
                '%s is required field', 
                $this->attributes['caption']
            );
            return false;
        }

        $regExp = $this->get('regexp');
        if ($regExp && !preg_match('#'.$regExp.'#', $value)) {
            
            $allowEmpty = $this->get('allowEmpty');
            $isNull = $this->get('isnull');

            if (!$value && ($allowEmpty || $isNull)) {
                return true;
            }

            $this->lastErrorMessage = __l(
                'Value is not valid in "%s"',
                $this->attributes['caption'],
                $regExp
            );

            return false;
        }
        
        $valueFilter = $this->get('valueFilter');
        if ($valueFilter && !empty($value)) {
            $res = filter_var($value, constant($valueFilter));
            if ($res === false) {
                $this->lastErrorMessage = __l(
                    'Value is not valid in "%s"',
                    $this->attributes['caption'],
                    $regExp
                );
                
                return false;
            }
        }

        return true;
    } // end isValidValue

    public function getFilterType(): ?string
    {
        $type = $this->get('filter');
        
        return $type ? strtolower($type) : null;
    } // end getFilterType

    /**
     * Returns a CSS names for DGS Filter HTML elements.
     * 
     * @return string
     * @return string
     */
    public function getFilterCssClassName(): string
    {
        $cssNames = $this->getCssClassName('filter');
        
        return $cssNames.' e-dgs-field-filter-'.$this->getFilterType();
    } // end getFilterCssClassName

    public function getRowCssClassName(): string
    {
        $classes = array();

        if ($this->isReadonly()) {
            $classes[] = 'm-readonly';
        }

        return join(' ', $classes);
    } // end getRowCssClassName

    /**
     * Returns a CSS names for UI element.
     *
     * @param bool|string $postfix = false
     * @return string
     */
    public function getCssClassName($postfix = false): string
    {
        $basePostfix = $postfix ? '-'.$postfix : '-input';
        
        $classes = array(
            'e-dgs-field'.$basePostfix,
            'e-dgs-field'.$basePostfix.'-'.$this->getType(),
            'e-dgs-field-element-'.$this->getElementName(),
            'e-dgs-field'.$basePostfix.'-'.$this->getElementName()
        );
        
        $externalClasses = $this->getExternalCssClassName($postfix);
        
        $classes = array_merge($classes, $externalClasses);
        
        return implode(' ', $classes);
    } // end getCssClassName

    /**
     * Returns additional a CSS names.
     *
     * return array
     */
    protected function getExternalCssClassName($postfix = false)
    {
        return array();
    } // end getExternalCssClassName

    public function getCssSelector(): string
    {
        return '.e-dgs-field-element-'.$this->getName();
    } // end getCssSelector

    public function getElementAttributes(): string
    {
        $attributes = array();
        
        if ($this->isReadonly()) {
            $attributes[] = 'readonly="readonly"';
        }
        
        if ($this->get(static::OPTION_REQUIRED)) {
            $attributes[] = 'required';
        }
        
        $errorMessage = $this->get('error');
        if ($errorMessage) {
            $errorMessage = htmlspecialchars($errorMessage);
            $attributes[] = 'data-error-message="'.$errorMessage.'"';
        }
        
        $regExp = $this->get('regexp');
        if ($regExp) {
            $attributes[] = 'pattern="'.$regExp.'"';
        }

        $autocomplete = $this->get(static::OPTION_AUTOCOMPLETE);
        if ($autocomplete === 'off') {
            $attributes[] = 'autocomplete="off"';
        }

        $mask = $this->get('mask');
        if (!empty($mask)) {
            // XXX: https://github.com/RobinHerbots/jquery.inputmask
            $attributes[] = 'data-inputmask="'.htmlspecialchars($mask).'"';
        }

        return implode(' ', $attributes);
    } // end getElementAttributes

    /**
     * Returns true if field only for readonly
     * @return bool
     */
    public function isReadonly(): bool
    {
        $attributeValue = $this->get(static::OPTION_READONLY);
        return $attributeValue === true || $attributeValue === 'true';
    }

    public function getLastErrorMessage(): ?string
    {
        return $this->lastErrorMessage;
    } // end getLastErrorMessage

    public function setErrorMessage(string $message): void
    {
        $this->lastErrorMessage = $message;
    } // end setErrorMessage

    public function getQueryWhere(): ?string
    {
        return $this->get(static::OPTION_WHERE) ?: null;
    } // end getQueryWhere

    public function getStoreName(): string
    {
        $storeName = $this->get('store');
        if (!$storeName) {
            $storeName = $this->store->getName();
        }

        return $storeName;
    } // end getStoreName

    public function isShow(): bool
    {
        $action = $this->store->getAction();
        if ($action !== Store::ACTION_LIST && $this->getName()) {
            return true;
        }

        if ($this->get('hide') || !$this->getName()) {
            return false;
        }

        return true;
    } // end isShow

    public function isSorting(): bool
    {
        if (!$this->get('sorting')) {
            return false;
        }

        return !!$this->getName();
    } // end isSorting

    public function getFilterKey(): string
    {
        $fieldName = $this->getName();
        if (!$fieldName) {
            $fieldName = $this->getIndex();
        }

        return $this->store->getIdent().'_'.$fieldName;
    } // end getFilterKey

    /**
     * Returns template name for filter
     *
     * @return string
     */
    protected function getFilterTemplateName()
    {
        $filterType = $this->get('filter');
        $teplateName = strtolower($filterType);

        return $teplateName.'.phtml';
    } // end getFilterTemplateName

    /**
     * Returns path to template folder
     *
     * @return string
     */
    protected function getFilterTemplatePath(): string
    {
        $path = $this->store->getOption(Core::OPTION_ENGINE_TEMPLATE_PATH).'filters';

        return $path.DIRECTORY_SEPARATOR;
    } // end getFilterTemplatePath
    
    /**
     * Returns filter template
     * @return string
     * @throws SystemException
     */
    protected function fetchFilterTemplate()
    {
        $view = new Template($this->getFilterTemplatePath());

        $view->assign('field', $this);

        return $view->fetch($this->getFilterTemplateName());
    } // end fetchFilterTemplate

    /**
     * Override this method if you need add custom logic into filter template.
     */
    protected function onFilterFetch(): void
    {
        $type = $this->getFilterType();
        $typesForLoadValues = array("multiple", "select");
        if (in_array($type, $typesForLoadValues)) {
            $options = array(
                'unique' => true,
                'foreignKeyField' => 'keyValue',
                'foreignValueField' => $this->getName(),
                'foreignTable' => $this->getStoreName()
            );
            
            $this->filterValues = $this->store->getProxy()->loadForeignValues(
                $options
            );
        }
    } // end onFilterFetch

    public function fetchFilter(): ?string
    {
        $filterType = $this->getFilterType();
        if (!$filterType) {
            return null;
        }

        $value = $this->store->getFieldFilterValueInSession($this);
        $this->filterValue = $value;

        $this->onFilterFetch();

        $store = $this->getStore();
        $target = array(
            'instance' => &$this,
            'field'    => $this->getName(),
            'value'    => &$this->filterValue,
            'values'   => &$this->filterValues,
            'action'   => $store->getAction()
        );

        $event = new FestiEvent(Store::EVENT_ON_FETCH_FIELD_FILTER, $target);
        $store->dispatchEvent($event);

        return $this->fetchFilterTemplate();
    } // end fetchFilter
    
    /**
     * @param $value
     */
    public function setItemValue($value): void
    {
        $this->itemValue = $value;
    }
    
    /**
     * @return mixed
     */
    public function getItemValue()
    {
        return $this->itemValue;
    }
    
    /**
     * Returns field input type.
     * 
     * @return string
     */
    public function getInputType()
    {
        $inputType = 'text';
        
        $numberTypes = array("price", "number");
        if (in_array($this->getType(), $numberTypes)) {
            $inputType = 'number';
        }
        
        $format = $this->get('format');
        if (in_array($format, $numberTypes)) {
            $inputType = 'number';
        }
        
        return $inputType;
    } // end getInputType
    
    /**
     * Returns values to field filter.
     * 
     * @return array
     */
    public function getFilterValues()
    {
        return $this->filterValues;
    } // end getFilterValues

    public function getCaption(): string
    {
        return $this->get('caption');
    } // end getCaption
    
    /**
     * @param $value
     * @return bool
     */
    protected function onPrepareValue(&$value)
    {
        // TODO: use htmlspecialchars
        $value = filter_var(
            $value, 
            FILTER_SANITIZE_STRING, 
            FILTER_FLAG_NO_ENCODE_QUOTES
        );
        
        return true;
    } // end onPrepareValue

    public function hasUserPermission(): bool
    {
        $sectionName = $this->get(static::OPTION_PERMISSION);

        if (!$sectionName) {
            return true;
        }

        $systemPlugin = Core::getInstance()->getSystemPlugin();

        return $systemPlugin->hasUserPermissionToSection($sectionName);
    } // end hasUserPermission
    
    /**
     * Returns the external value of this field.
     *
     * @return string
     */
    protected function getExternalValue()
    {
        $externalValues = $this->getStore()->getModel()->getExternalValues();
        if (!array_key_exists($this->getName(), $externalValues)) {
            return false;
        }
        
        return $externalValues[$this->getName()];
    }

    public function hasAutocomplete(): bool
    {
        $autocomplete = $this->get(static::OPTION_AUTOCOMPLETE);
        if ($autocomplete === 'off') {
            return false;
        }
        return (bool) $autocomplete;
    } // end hasAutocomplete

    public function getAutocompleteUrl(): string
    {
        $autocomplete = $this->get('autocomplete');
        if (!$autocomplete) {
            $message = __('Undefined autocomplete attribute in %s', $this->getName());
            throw new FieldException($message);
        }

        if ($autocomplete !== true && is_string($autocomplete)) {
            $primaryKeyField = $this->store->getPrimaryKey();
            $primaryKeyValue = $this->store->getPrimaryKeyValueFromRequest();
            if (isset($primaryKeyValue)) {
                $rowData = array(
                    $primaryKeyField => $primaryKeyValue,
                );
                $autocomplete = Entity::fillString($autocomplete, $rowData);
            }

            return $autocomplete;
        }

        $params = array(
            Store::ACTION_KEY_IN_REQUEST => Store::ACTION_FOREIGN_KEY_LOAD,
            'fieldName'                  => $this->getElementName(),
        );

        $ident = $this->getStore()->getIdent();
        $params = array(
            $ident => $params,
        );

        $url = $this->getStore()->getCurrentUrl();

        $event = new UrlStoreActionEvent($url, $params, $this->store);
        $this->store->dispatchEvent($event);

        return Core::getInstance()->getUrl($url, $params);
    } // end getAutocompleteUrl

    public function getElementName(): string
    {
        $fieldName = $this->getName();
        if (!$fieldName) {
            $fieldName = '__'.$this->getIndex();
        }

        return $fieldName;
    } // end getElementName

    public function getKeyInRequest(): string
    {
        return $this->getName();
    } // end getElementName

    public function isRequired(): bool
    {
        $isRequired = $this->get(static::OPTION_REQUIRED);
        return $isRequired === true || $isRequired === 'true';
    }
}
