<?php

/**
 * DateTime Field
 *
 * @package Festi
 * @subpackage Fields
 */
class DatetimeField extends AbstractField
{
    public const DEFAULT_DB_DATETIME_FORMAT = 'Y-m-d H:i:s';
    
    public $needTime;
    
    /**
     * @override
     */
    protected function getAttributesFields(): array
    {
        $fields = parent::getAttributesFields();

        $fields['format'] = array(
            'type'     => PARAM_STRING,
            'default'  => '%m/%d/%Y %I:%M %P',
        );

        $fields['time'] = static::FIELD_TYPE_STRING_NULL;

        return $fields;
    } // end getAttributesFields
    
    /**
     * @override
     */
    public function getEditInput(?string $value = '', $inline = null): ?string
    {
        $format = $this->getFormat();

        if ($this->isHtmlFiveVersion()) {
            $format = '%Y-%m-%d';
            if ($this->get('time')) {
                $format = '%Y-%m-%dT%H:%M';
            }
        }

        if (empty($value)) {
            $default = $this->get('default');
            if ($default) {
                $value = $default;
            } else if (!$this->get('isnull')) {
                // TODO: Use date function (another format)
                $value = strftime($format);
            }
        } else {
            $value = $this->getPreparedValueForEdit($value, $format);
        }

        if (!empty($this->attributes['readonly'])) {
            //return $this->displayRO($value) ;
        }

        return $this->getHtml($value, $format);
    } // end getEditInput

    /**
     * Returns date format according to locale settings.
     */
    public function getFormat()
    {
        $format = $this->get('format');
        if ($format) {
            return $format;
        }

        $this->needTime = 'false';
        $format = '%Y-%m-%d';

        return $format;
    } // end getFormat
    
    /**
     * @override
     */
    public function displayValue(?string $value, array $row = array()): ?string
    {
        return $this->getFormattedValue($value);
    } // end displayValue

    /**
     * @override
     */
    public function getFormattedValue($value): ?string
    {
        $format = $this->getFormat();

        if (!empty($value)) {
            $value = $this->_getPreparedViewValue($value);
            // TODO: Use date function (another format)
            $value = strftime($format, strtotime($value));
        }

        return $value;
    } // end getFormattedValue

    /**
     * @override
     */
    public function displayRO($value): ?string
    {
        return $this->getFormattedValue($value);
    } // end displayRO

    private function getHtml($value, $format)
    {
        $this->assign('value', htmlspecialchars($value));
        $this->assign('format', $format);

        return $this->fetch('edit.phtml');
    } // end getHtml

    protected function getPreparedValueForEdit($value, $format)
    {
        $value = $this->_getPreparedViewValue($value);
        // TODO: Use date function (another format)
        return strftime($format, strtotime($value));
    } // end getPreparedValueForEdit

    /**
     * @override
     */
    public function getValue(array $request = array())
    {
        $value = parent::getValue($request);

        if (!$value && isset($this->attributes['isnull'])) {
            $value = null;
            return $value;
        }

        if ($this->getLastErrorMessage()) {
            return false;
        }
    
        $dateTime = new DateTime($value);
    
        if ($this->get('time')) {
            $userTimezone = $this->getStore()->getModel()->get('userTimezone');
            $sourceTimezone = $this->getStore()->getModel()->get('sourceTimezone');
    
            $dateTime = new DateTime($value, new DateTimeZone($userTimezone));
            $dateTime->setTimezone(new DateTimeZone($sourceTimezone));
        }
        
        return $dateTime->format(static::DEFAULT_DB_DATETIME_FORMAT);
    } // end getValue

    /**
     * @override
     */
    protected function onFilterFetch(): void
    {

    } // end onFilterFetch

    protected function getFilterTemplateName()
    {
        if ($this->getFilterType() == static::FILTER_TYPE_RANGE) {
            return 'datetime_range.phtml';
        }

        return 'datetime.phtml';
    } // end getFilterTemplateName

    /**
     * @override
     */
    protected function getTemplatePath(): string
    {
        return parent::getTemplatePath().'datetime'.DIRECTORY_SEPARATOR;
    } // end getTemplatePath

    protected function isHtmlFiveVersion(): bool
    {
        $htmlFive = $this->get('html5');

        return $htmlFive && $htmlFive == "true";
    } // end isHtmlFiveVersion
    
    private function _getPreparedViewValue(string $value): string
    {
        $sourceTimezone = $this->getStore()->getModel()->get('sourceTimezone');
        $dateTime = new DateTime($value, new DateTimeZone($sourceTimezone));
        
        if ($this->get('time')) {
            $userTimezone = $this->getStore()->getModel()->get('userTimezone');
            $dateTime->setTimezone(new DateTimeZone($userTimezone));
        }
        
        return $dateTime->format(static::DEFAULT_DB_DATETIME_FORMAT);
    } // end _getPreparedViewValue
}
