<?php

use core\store\field\IStoreForeignField;

class ForeignKeyField extends AbstractSelectField implements IStoreForeignField
{
    public const OPTION_VALUES_WHERE         = 'valuesWhere';
    public const OPTION_FOREIGN_KEY_FIELD    = 'foreignKeyField';
    public const OPTION_FOREIGN_VALUES_FIELD = 'foreignValueField';
    public const OPTION_FOREIGN_TABLE        = 'foreignTable';
    public const OPTION_SHOW_SEARCH_INPUT    = 'showSearchInput';

    private array $keyData = array();

    /**
     * Set list of values.
     *
     * @param array $values
     */
    public function setValuesList(array $values): void
    {
        $this->keyData = $values;
    }

    public function getValuesList(): array
    {
        return $this->keyData;
    }

    public function isCustomLoadValues(): bool
    {
        return $this->hasAutocomplete();
    }

    /**
     * @override
     */
    public function getEditInput(?string $value = '', $rowData = null): ?string
    {
        $this->assign('fieldAttributes', $this->getAttributes());

        $foreignValue = $value;
        if ($rowData) {
            $foreignValue = $rowData['_foreign_'.$this->getName()];
        }
        $this->assign('foreignValue', $foreignValue);

        if (!empty($this->attributes['readonly'])) {
            return $this->displayRO($value);
        }

        if ($this->get('ajaxChild') && $rowData) {
            $ajaxChildValues = $this->getAjaxChildValues($rowData);
            $this->assign('ajaxChildValues', $ajaxChildValues);
        }

        $this->assign('value', $value);

        return $this->fetch('edit.phtml');
    } // end getEditInput

    /**
     * @override
     */
    public function displayRO($value): ?string
    {
        return $value;
    }

    public function getForeignStoreName(): string
    {
        $foreignTableName = $this->get(static::OPTION_FOREIGN_TABLE);

        if (preg_match("/\sas\s(.+)$/", $foreignTableName, $tmp)) {
            $foreignTableName = $tmp[1];
        }

        return $foreignTableName;
    } // end getForeignStoreName

    public function getForeignTable(): string
    {
        return $this->get(static::OPTION_FOREIGN_TABLE);
    } // end getForeignStoreName

    public function getValuesWhere(): ?string
    {
        return $this->get(static::OPTION_VALUES_WHERE) ?: null;
    } // end getForeignStoreName

    /**
     * Returns key for foreign field.
     *
     * @return string
     */
    public function getForeignFieldKey(): string
    {
        return $this->get(static::OPTION_FOREIGN_KEY_FIELD);
    } // end getForeignFieldKey

    /**
     * Returns value for foreign field.
     *
     * @return string
     */
    public function getForeignFieldValue(): string
    {
        return $this->get(static::OPTION_FOREIGN_VALUES_FIELD);
    } // end getForeignFieldValue

    public function setForeignFieldValue(string $value): void
    {
        $this->set(static::OPTION_FOREIGN_VALUES_FIELD, $value);
    } // end setForeignFieldValue


    public function getFormattedFilterValue($value): ?string
    {
        $filterType  = $this->getFilterType();
     
        if ($this->_isMultipleFilter($filterType, $value)) {
            $value = $this->_getMultipleFilterValue($value);
        }
        
        if ($this->_isSelectFilter($filterType, $value)) {
            $value = $this->_getSelectFilterValue($value);
        }
        
        return parent::getFormattedFilterValue($value);
    } // end getFormattedFilterValue

    /**
     * @override
     */
    protected function onFilterFetch(): void
    {
        $this->store->loadForeignKeys();
        $values = $this->keyData;

        $this->filterValues = $values;
    } // end onFilterFetch
    
    /**
     * @override
     */
    protected function getAttributesFields(): array
    {
        $fields = parent::getAttributesFields();

        $fields[static::OPTION_FOREIGN_TABLE] = array(
            'type'     => static::FIELD_TYPE_STRING,
            'error'    => 'Undefined "foreignTable" attribute in field',
            'required' => true
        );

        $fields[static::OPTION_FOREIGN_KEY_FIELD] = array(
            'type'     => static::FIELD_TYPE_STRING,
            'error'    => 'Undefined "foreignKeyField" attribute in field',
            'required' => true
        );

        $fields[static::OPTION_FOREIGN_VALUES_FIELD] = array(
            'type'     => static::FIELD_TYPE_STRING,
            'error'    => 'Undefined "foreignValueField" attribute in field',
            'required' => true
        );

        $fields[static::OPTION_SHOW_SEARCH_INPUT] = array(
            static::OPTION_KEY_TYPE    => static::FIELD_TYPE_BOOL,
            static::OPTION_KEY_DEFAULT => true,
        );

        $fields[static::OPTION_VALUES_WHERE] = static::FIELD_TYPE_STRING_NULL;
        $fields['foreignOrderBy'] = static::FIELD_TYPE_STRING_NULL;
        $fields['foreignLimit'] = static::FIELD_TYPE_STRING_NULL;
        $fields['unique'] = static::FIELD_TYPE_BOOL;

        return $fields;
    } // end getAttributesFields

    /**
     * @override
     * @throws SystemException
     */
    protected function getTemplatePath(): string
    {
        return parent::getTemplatePath().'foreign_key'.DIRECTORY_SEPARATOR;
    } // end getTemplatePath

    private function _getMultipleFilterValue($value): array
    {
        $filterValue = array();
    
        foreach ($value as $item) {
            if (!array_key_exists($item, $this->filterValues)) {
                continue;
            }
            $filterValue[] = $this->filterValues[$item];
        }
        
        if (!$filterValue) {
            $this->setErrorMessage("Can't find filter value.");
        }
        
        return $filterValue;
    } // end _getMultipleFilterValue
    
    private function _getSelectFilterValue($value)
    {
        if (!array_key_exists($value, $this->filterValues)) {
            $this->setErrorMessage("Can't find filter value.");
            return false;
        }
        
        return $this->filterValues[$value];
    } // end _getSelectFilterValue
    
    private function _isMultipleFilter($filterType, $value): bool
    {
        return $filterType == static::FILTER_TYPE_MULTIPLE &&
               is_array($value) &&
               $value;
    } // end _isMultipleFilter
    
    private function _isSelectFilter($filterType, $value): bool
    {
        return $filterType == static::FILTER_TYPE_SELECT &&
               is_scalar($value) &&
               $value;
    } // end _isSelectFilter
}
