<?php

namespace core\store\field;

use AbstractAction;
use Store;

interface IStoreField
{
    public const OPTION_ONLY_LIST = 'onlyList';
    public const OPTION_READONLY  = 'readonly';

    /**
     * Returns attribute value or false if that not found.
     *
     * @param string $name
     * @return mixed
     */
    public function get(string $name);

    /**
     * Set attribute value to field.
     *
     * @param $key
     * @param $value
     */
    public function set(string $key, $value): void;

    public function getCaption(): string;

    public function getName(): string;
    public function getIndex(): int;
    public function setIndex(int $index): void;

    public function getType(): string;

    /**
     * Returns field name for UI.
     *
     * @return string
     */
    public function getElementName(): string;

    /**
     * Returns field key in requests.
     *
     * @return string
     */
    public function getKeyInRequest(): string;

    public function getSelectorPostfix(): string;

    /**
     * @return Store
     */
    public function &getStore(): Store;

    /**
     * Return table name for current field
     *
     * @return string
     */
    public function getStoreName(): string;

    /**
     * @return mixed
     */
    public function getItemValue();

    /**
     * @param mixed $value
     */
    public function setItemValue($value): void;

    /**
     * @param string $value
     * @param array $row
     * @return string|null
     */
    public function getInfoValue($value, $row = null): ?string;

    /**
     * Returns filter key for request parameter and session key.
     *
     * @return string
     */
    public function getFilterKey(): string;

    /**
     * Returns filter type. Returns false if the filter does not set.
     *
     * @return string|null
     */
    public function getFilterType(): ?string;

    /**
     * Returns true if field must be ignored in update table logic.
     *
     * @param string|null $actionName DGS Action Name
     * @return boolean
     */
    public function isVirtualField(?string $actionName = null): bool;

    /**
     * Returns true if field must be ignored in queries logic.
     *
     * @return boolean
     */
    public function isCustom(?string $section = null): bool;

    /**
     * @return bool
     */
    public function isShow(): bool;

    /**
     * Returns string with attributes of UI element.
     *
     * @return string
     */
    public function getElementAttributes(): string;

    /**
     * @param mixed $value
     * @return string|null
     */
    public function getFormattedValue($value): ?string;

    /**
     * @param array $request
     * @return bool|string|mixed|null
     * @throws \SystemException
     */
    public function getValue(array $request = array());

    /**
     * @param string $value
     * @return string|null
     * @throws \SystemException
     */
    public function displayRO($value): ?string;

    /**
     * Returns edit html input for field.
     *
     * @param string|null $value
     * @param array|null $inline
     * @return string|null
     * @throws \SystemException
     */
    public function getEditInput(?string $value = '', $inline = null): ?string;

    /**
     * @param string|null $value
     * @param array $row
     * @return string|null
     * @throws \SystemException
     */
    public function displayValue(?string $value, array $row = array()): ?string;

    public function getLastErrorMessage(): ?string;
    public function setErrorMessage(string $message): void;

    public function getQueryWhere(): ?string;

    /**
     * Returns true if field can be sorting on data list.
     *
     * @return boolean
     */
    public function isSorting(): bool;

    /**
     * Returns true if user has permission to the field.
     *
     * @return bool
     * @throws \SystemException
     */
    public function hasUserPermission(): bool;

    public function getAttributes(): array;

    /**
     * Returns formatted filter value.
     * Formatted value uses for display to a user.
     *
     * @param string|array|null $value
     * @return string|null
     * @throws \FieldException
     */
    public function getFormattedFilterValue($value): ?string;

    /**
     * Returns html template of filed filter.
     *
     * @return string|null
     */
    public function fetchFilter(): ?string;

    /**
     * Returns a CSS names for DGS row HTML elements.
     *
     * @return string
     */
    public function getRowCssClassName(): string;

    /**
     * Returns css selector for JS.
     *
     * @return string
     */
    public function getCssSelector(): string;

    /**
     * Set reference to current DGS action.
     *
     * @param AbstractAction $action
     */
    public function setAction(AbstractAction &$action): void;

    /**
     * Returns encrypted value.
     * @param $value
     * @return string
     * @throws \SystemException
     */
    public function getEncryptValue($value): string;

    /**
     * Returns decrypted value.
     *
     * @param string $value
     * @return string
     * @throws \SystemException
     */
    public function getDecryptedValue(string $value): ?string;

    /**
     * Returns true if field has autocomplete.
     *
     * @return bool
     */
    public function hasAutocomplete(): bool;

    /**
     * Returns autocomplete url.
     *
     * @return string
     * @throws \FieldException
     * @throws \SystemException
     */
    public function getAutocompleteUrl(): string;
}
