<?php

use core\store\field\IStoreField;

/**
 * Class StoreModel
 */
class StoreModel extends Entity
{
    public const TABLE_ATTRIBUTE_ERROR_MESSAGE = 'errorMessage';
    public const TABLE_ATTRIBUTE_PERMISSION    = 'permission';
    public const TABLE_ATTRIBUTE_FIXED_HEADER  = 'fixedHeader';

    public const ACTION_ATTRIBUTE_REDIRECT_URL = 'redirectUrl';

    public const OPTION_ACTIONS_MODE = 'actions.mode';

    public const OPTION_ACTIONS_MODE_LIST    = 'list';
    public const OPTION_ACTIONS_MODE_BUTTONS = 'buttons';

    public const OPTION_FILTERS_MODE = 'filters.mode';

    public const OPTION_FILTERS_MODE_AJAX    = 'ajax';
    public const OPTION_FILTERS_MODE_TOP     = 'top';
    public const OPTION_FILTERS_MODE_RIGHT   = 'right';
    public const OPTION_FILTERS_MODE_DEFAULT = 'default';

    public const OPTION_SECTIONS_MODE = 'sections.mode';

    public const OPTION_SECTIONS_MODE_TABS    = 'tabs';
    public const OPTION_SECTIONS_MODE_WIDGETS = 'widgets';

    public const MODE_WEB = 'web';
    public const MODE_API = 'api';

    /**
     * Store attributes.
     */
    protected array $attributes = array();

    /**
     * Store fileds.
     *
     * @var IStoreField[]
     */
    protected array $fields = array();

    /**
     * Store actions.
     */
    protected array $actions = array();

    /**
     * Store relations.
     */
    protected array $relations = array();

    /**
     * List of group actions.
     */
    protected array $grouped = array();

    /**
     * Condition filters.
     */
    protected array $filters = array();

    /**
     * Store sections.
     */
    protected array $sections = array();

    protected array $externalValues = array();

    protected array $routers = array();

    protected array $search = array();

    protected array $listeners = array();

    protected array $highlights = array();

    /**
     * List of a condition for a total.
     */
    protected array $aggregations = array();

    /**
     * Store name.
     */
    private string $_name;

    /**
     * Primary key name.
     */
    protected string $primaryKey;

    /**
     * Reference to store.
     */
    protected Store $store;

    /**
     * DGS configuration options.
     */
    protected array $options = array();

    protected string $stream = 'php://input';
    
    /**
     * StoreModel constructor.
     * @param Store $store
     */
    public function __construct(Store &$store)
    {
        $this->store = &$store;
        $this->fields = array();

    } // end __construct

    /**
     * Returns name of table in database.
     * Table name in system not always the same real name of table in database.
     *
     * @return string
     */
    public function getName(): string
    {
        return $this->_name;
    } // end getName
    
    /**
     * @param $name
     */
    public function setName(string $name): void
    {
        $this->_name = $name;
    } // end setName

    /**
     * Returns attribute value.
     *
     * @param string $name
     * @return mixed
     */
    public function get(string $name)
    {
        $value = false;
        if (array_key_exists($name, $this->attributes)) {
            $value = $this->attributes[$name];
            if ($value === 'false') {
                $value = false;
            }
        }

        return $value;
    } // end get

    /**
     * Set attribute value.
     *
     * @param string $key
     * @param $value
     */
    public function set(string $key, $value): void
    {
        $this->attributes[$key] = $value;
    } // end set
    
    /**
     * @param $actionName
     * @return bool
     */
    public function hasAction(string $actionName): bool
    {
        return isset($this->actions[$actionName]);
    }
    
    /**
     * @param $actionName
     * @return array|null
     */
    public function getAction(string $actionName): ?array
    {
        if (!$this->hasAction($actionName)) {
            return null;
        }

        return $this->actions[$actionName];
    } // end getAction
    
    /**
     * @param $actionName
     * @return array|null
     */
    public function &getActionByRef(string $actionName): ?array
    {
        if (!$this->hasAction($actionName)) {
            $res = null;
            return $res;
        }

        return $this->actions[$actionName];
    } // end getAction
    
    /**
     * @param $actionName
     */
    public function removeAction(string $actionName): void
    {
        unset($this->actions[$actionName]);
    } // end removeAction
    
    /**
     * @return array
     */
    public function &getActions(): array
    {
        return $this->actions;
    } // end getActions
    
    /**
     * @return array
     */
    public function &getHighlights(): array
    {
        return $this->highlights;
    } // end getHighlights
    
    /**
     * @param $name
     * @return ?array
     */
    public function &getRelation(string $name): ?array
    {
        if (!isset($this->relations[$name])) {
            $res = null;
            return $res;
        }

        return $this->relations[$name];
    } // end getRelation
    
    /**
     * @return IStoreField[]
     */
    public function &getFields(): array
    {
        return $this->fields;
    }
    
    /**
     * @return array
     */
    public function &getSections(): array
    {
        return $this->sections;
    }

    /**
     * Returns reference to field by name.
     *
     * @param $name field name
     * @return IStoreField|null
     */
    public function &getFieldByName(string $name) : ?IStoreField
    {
        if (!array_key_exists($name, $this->fields) || !$this->fields[$name]) {
            // XXX: Cheat for "Only variable references should be returned by reference"
            $ref = null;
            return $ref;
        }

        return $this->fields[$name];
    } // end getFieldByName

    /**
     * Returns reference to field by option.
     *
     * @param string $optionName
     * @param $optionValue
     * @return IStoreField|null
     */
    public function &getFieldByOption(string $optionName, $optionValue) : ?IStoreField
    {
        $result = null;
        foreach ($this->fields as &$field) {
            $value = $field->get($optionName);
            if ($optionValue == $value) {
                $result = &$field;
                break;
            }
        }

        return $result;
    } // end getFieldByOption

    /**
     * Remove filed from DGS.
     *
     * @param string $name
     * @return bool
     */
    public function removeFieldByName(string $name): bool
    {
        if (!array_key_exists($name, $this->fields)) {
            return false;
        }

        unset($this->fields[$name]);
        return true;
    } // end removeFieldByName

    /**
     * Returns field
     *
     * @param string $name
     * @return IStoreField|null
     */
    public function getField(string $name): ?IStoreField
    {
        return isset($this->fields[$name]) ? $this->fields[$name] : null;
    } // end getField

    public function getPrimaryKey(): string
    {
        return $this->primaryKey;
    }

    public function getGroupActions(): array
    {
        return $this->grouped;
    }
    
    /**
     * @throws SystemException
     */
    public function load(): void
    {
        throw new SystemException('Undefined load method');
    } // end load
    
    /**
     * @return array
     */
    protected function getAttributesOptions(): array
    {
        $attributesFields = array(
           'name' => static::FIELD_TYPE_STRING_NULL,

            'primaryKey' => array(
                'type' => static::FIELD_TYPE_STRING_NULL,
                'required' => 1
            ),

            'rowsForPage' => array(
                'type' => static::FIELD_TYPE_STRING_NULL,
                'default' => 50
            ),

            'paging' => array(
                'type' => static::FIELD_TYPE_STRING_NULL,
                'default' => 'normal'
            ),

            'proxy' => array(
                'type' => static::FIELD_TYPE_STRING_NULL,
                'default' => 'sql'
            ),

            'mode' => array(
                'type' => static::FIELD_TYPE_STRING_NULL,
                'default' => static::MODE_WEB
            ),

            'filter' => array(
                'type' => static::FIELD_TYPE_STRING_NULL,
                'default' => static::OPTION_FILTERS_MODE_DEFAULT
            ),
           'sourceTimezone'        => array(
               'type' => static::FIELD_TYPE_STRING_NULL,
               'default' => date_default_timezone_get()
           ),
           'userTimezone'          => array(
               'type' => static::FIELD_TYPE_STRING_NULL,
               'default' => date_default_timezone_get()
           ),

           'charset'               => static::FIELD_TYPE_STRING_NULL,
           'defaultOrderField'     => static::FIELD_TYPE_STRING_NULL,
           'defaultOrderDirection' => static::FIELD_TYPE_STRING_NULL,
           'plugin'                => static::FIELD_TYPE_STRING_NULL,
           'emptyMessage'          => static::FIELD_TYPE_STRING_NULL,
           'fastAdd'               => static::FIELD_TYPE_STRING_NULL,
           'join'                  => static::FIELD_TYPE_STRING_NULL,
           'additionalWhere'       => static::FIELD_TYPE_STRING_NULL,
           'grideditor'            => static::FIELD_TYPE_STRING_NULL,
           'exceptionMode'         => static::FIELD_TYPE_STRING_NULL,

           static::TABLE_ATTRIBUTE_ERROR_MESSAGE => static::FIELD_TYPE_STRING_NULL,
           static::TABLE_ATTRIBUTE_PERMISSION => static::FIELD_TYPE_STRING_NULL,
           static::TABLE_ATTRIBUTE_FIXED_HEADER => static::FIELD_TYPE_STRING_NULL
        );

        return $attributesFields;
    } // end getAttributesFields

    /**
     * Returns field instance created by field information
     *
     * @param FieldModel $model
     * @param int|null $index
     * @return IStoreField
     */
    public function createFieldInstance(FieldModel $model, int $index = null): IStoreField
    {
        $attributes = $model->getAttributes();
        $className = ucfirst($attributes['type']).'Field';

        if (!class_exists($className)) {
            require_once 'bundle/store/field/'.$className.'.php';
        }

        $field = new $className($this->store);

        if (!is_null($index)) {
            $field->setIndex($index);
        }

        if ($model) {
            $field->onInit($model);
        }

        return $field;
    } // end createFieldInstance
    
    /**
     * @throws SystemException
     */
    protected function doPrepareActions(): void
    {
        $attributes = $this->getActionAttributesOptions();

        foreach ($this->actions as $key => &$action) {
            $action = $this->getExtendData($action, $attributes, $errors);
            if ($errors) {
                $message = current($errors);
                throw new SystemException($message);
            }

            if ($action['permission']) {
                if (!$this->_hasUserActionPermission($action['permission'])) {
                    unset($this->actions[$key]);
                }
            }
        }
        unset($action);
    } // end doPrepareActions
    
    /**
     * @param $section
     * @return mixed
     * @throws SystemException
     */
    private function _hasUserActionPermission(string $section): bool
    {
        $systemPlugin = Core::getInstance()->getSystemPlugin();

        return $systemPlugin->hasUserPermissionToSection($section);
    }

    /**
     * Returns list of all known attributes for action
     *
     * @return array
     */
    protected function getActionAttributesOptions(): array
    {
        $attributes = array(
            'type' => array(
                'type'     => PARAM_STRING,
                'error'    => 'Undefined type in action',
                'required' => true
            ),
            'mode' => array(
                'type'    => PARAM_STRING,
                'default' => Store::ACTION_VIEW_MODE_DEFAULT
            ),
            'confirmDialog' => PARAM_STRING_NULL,
            'link'          => PARAM_STRING_NULL,
            'src'           => PARAM_STRING_NULL,
            'caption'       => PARAM_STRING_NULL,
            'js'            => PARAM_STRING_NULL,
            'view'          => PARAM_STRING_NULL,
            'addon'         => PARAM_STRING_NULL,
            'html'          => PARAM_STRING_NULL,
            'permission'    => static::FIELD_TYPE_STRING_NULL,
        );

        return $attributes;
    } // end getActionAttributesOptions

    /**
     * Returns list of all known attributes for group action
     *
     * @return array
     */
    protected function getGroupActionFields(): array
    {
        $attributes = array(
            'caption' => array(
                'type'     => PARAM_STRING,
                'error'    => 'Undefined caption in group action',
                'required' => true
            ),
            'type' => array(
                'type'     => PARAM_STRING,
                'error'    => 'Undefined type in group action',
                'required' => true
            ),
            'link'    => PARAM_STRING_NULL,
            'js'      => PARAM_STRING_NULL
        );

        return $attributes;
    } // end getGroupActionFields

    /**
     * Returns list of all known attributes for actions
     *
     * @return array
     */
    protected function getActionsAttributesOptions(): array
    {
        $attributes = array(
            'mode' => array(
                'type'    => PARAM_STRING,
                'default' => static::OPTION_ACTIONS_MODE_LIST
            )
        );

        return $attributes;
    } // end getActionsAttributesOptions

    /**
     * Returns list of all known attributes for sections
     *
     * @return array
     */
    protected function getSectionsAttributesOptions(): array
    {
        $attributes = array(
            'mode' => array(
                'type'    => PARAM_STRING,
                'default' => static::OPTION_SECTIONS_MODE_WIDGETS
            )
        );

        return $attributes;
    } // end getSectionsAttributesOptions
    
    /**
     * @return array
     */
    public function &getFilters(): array
    {
        return $this->filters;
    } // end getFilters
    
    /**
     * @return mixed
     */
    public function &getRouters(): array
    {
        return $this->routers;
    } // end getRouters
    
    /**
     * @return array
     */
    public function &getSearch(): array
    {
        return $this->search;
    } // end getSearch
    
    /**
     * @return array
     */
    public function &getExternalValues(): array
    {
        return $this->externalValues;
    } // end getExternalValue
    
    /**
     * @return bool
     */
    public function isApiMode(): bool
    {
        return $this->get('mode') == StoreModel::MODE_API;
    } // end isApiMode
    
    /**
     * @return bool
     */
    public function isWebMode(): bool
    {
        return $this->get('mode') == StoreModel::MODE_WEB;
    } // end isApiMode

    /**
     * Returns DGS action name from REQUEST_METHOD.
     *
     * @return string
     * @throws StoreException
     */
    public function getActionByRequestMethod(): string
    {
        $requestMethod = $_SERVER['REQUEST_METHOD'];

        switch ($requestMethod) {
            case 'GET':
                $action = $this->_getActionByRequestMethodGet();
                break;

            case 'POST':
                $action = $this->_getActionByRequestMethodPost();
                break;

            case 'PUT':
                $action = Store::ACTION_EDIT;
                break;

            case 'DELETE':
                $action = Store::ACTION_REMOVE;
                break;

            default:
                $action = Store::ACTION_LIST;
        }

        $actions = $this->getActions();

        if (!array_key_exists($action, $actions)) {
            throw new StoreException(__('The action "%s" is undefined in DGS model', $requestMethod));
        }

        return $action;
    } // end getActionByRequestMethod
    
    /**
     * @return string
     */
    private function _getActionByRequestMethodGet(): string
    {
        $action = Store::ACTION_LIST;
        $primaryKey = $this->store->getPrimaryKeyValueFromRequest();

        if ($primaryKey) {
            $action = Store::ACTION_INFO;
        }

        return $action;
    } // end _getActionByRequestMethodGet
    
    /**
     * @return string
     */
    private function _getActionByRequestMethodPost(): string
    {
        $action = Store::ACTION_INSERT;

        $request = $this->getRequest();

        if (is_array($request) && is_array(array_pop($request))) {
            $action = Store::ACTION_BATCH_INSERT;
        }

        return  $action;
    } // end _getActionByRequestMethodPost
    
    /**
     * @return array|null
     */
    public function getRequest(): ?array
    {
        $request = $_POST;

        if (!$request) {
            $request = $_REQUEST;
        }

        if ($this->isApiMode()) {
            $stream = $this->getStream();

            $request = file_get_contents($stream);

            if (!$request) {
                return null;
            }
            
            $request = json_decode($request, true);
        }

        return $request;
    } // end getRequest
    
    /**
     * @return string
     */
    public function &getStream(): string
    {
        return $this->stream;
    } // getStream

    /**
     * Returns reference to aggregations conditions.
     *
     * @return array &
     */
    public function &getAggregations(): array
    {
        return $this->aggregations;
    } // end getAggregations

    /**
     * Returns model option value or NULL if undefined option.
     *
     * @param string $key
     * @return mixed|null
     */
    public function getOption(string $key)
    {
        if (!array_key_exists($key, $this->options)) {
            return null;
        }

        return $this->options[$key];
    } // end getOption

    /**
     * Returns reference to DGS configuration options.
     *
     * @return array &
     */
    public function &getOptions(): array
    {
        return $this->options;
    }
    
    /**
     * @param $section
     * @param $options
     * @param $attributesOptions
     * @return bool
     */
    protected function loadSectionOptions(
        string $section, ?array $options, array $attributesOptions
    ): void
    {
        if (empty($options) && !is_array($options)) {
            $options = array();
        }

        $preparedOptions = $this->getExtendData(
            $options,
            $attributesOptions,
            $errors
        );

        foreach ($preparedOptions as $key => $value) {
            $this->options[$section.'.'.$key] = $value;
        }
    } // end loadSectionOptions

    /**
     * Returns table attributes.
     *
     * @return array
     */
    public function &getAttributes(): array
    {
        return $this->attributes;
    } // end getAttributes
}
