<?php

use core\store\field\IStoreField;

class MssqlProxy extends SqlProxy
{
    /**
     * @param bool $isAllColumns
     * @return array
     * @throws Exception
     */
    public function loadListValues(bool $isAllColumns = false): array
    {
        $columns = $this->getQueryColumns($isAllColumns);
        $columns = array_unique($columns);
        
        $joins   = $this->getQueryJoins($columns);
        $where   = $this->getQueryWhere();
        $having  = $this->getQueryHaving();
        $groupBy = $this->getQueryGroupBy();
        
        $columnsSection = join(', ', $columns);
        $fromSection    = join(' ', $joins);

        $orderBy = $this->getQueryOrderDirection();

        $sql = "SELECT ".$columnsSection.", ".
                        "ROW_NUMBER() OVER (ORDER BY ".$orderBy.") AS __row_num ".
                "FROM ".
                    $fromSection;
                    
        if ($where) {
            $sql .= " WHERE ".join(' AND ', $where);
        }
        
        if ($groupBy) {
            $sql .= $groupBy;
        }

        $limit = $this->getQueryLimit();
        $limit = explode(", ", $limit);
        $endLimit = $limit[0] + $limit[1];

        $having['t.__row_num&BETWEEN'] = array($limit[0], $endLimit);
        $havingStatements = $this->connection->getSqlCondition($having);
        
        $sqlWrapper = "SELECT ".
                            "t.* ".
                        "FROM ".
                            "(".$sql.") AS t ".
                        "WHERE 
                            ".join(' AND ', $havingStatements).
                        " ORDER BY ".
                            "t.__row_num";
        $rows = $this->connection->getAll($sqlWrapper);

        $sql = "SELECT COUNT(*) OVER () AS TotalRecords FROM ".$fromSection;
        
        if ($where) {
            $sql .= " WHERE ".join(' AND ', $where);
        }
        
        if ($groupBy) {
            $sql .= $groupBy;
        }

        $total = (int) $this->connection->getOne($sql);
        $this->setCount($total);
        
        return $rows;
    } // end loadListValues

    protected function getQueryOrderDirection(): ?string
    {
        $fieldName = $this->store->getOrderByFieldName();

        if (!$fieldName) {
            return $this->store->getName().static::STORE_NAME_AND_COLUMN_SEPARATOR.$this->store->getPrimaryKey();
        }

        $orderByDirection = $this->store->getOrderByDirection();

        $field = $this->model->getFieldByName($fieldName);
        if ($field) {
            $fieldName = $this->getColumnNameByOrderField($field);
        }

        if (!static::hasStoreName($fieldName)) {
            $fieldName = $this->store->getName().static::STORE_NAME_AND_COLUMN_SEPARATOR.$fieldName;
        }

        return $fieldName." ".$orderByDirection;
    }
    
    /**
     * @param Many2manyField $field
     * @return string
     * @throws SystemException
     */
    protected function getMany2manyQuery(Many2manyField $field): string
    {
        $foreignTable = $field->getForeignTable();
        $foreignTableName = $this->connection->quoteTableName($field->getForeignStoreName());
        $foreignValueField = $field->getForeignFieldValue();
    
        if (static::isExpression($foreignValueField)) {
            $valuesColumn = $foreignValueField;
        } else {
            $valuesColumn = $foreignTableName.static::STORE_NAME_AND_COLUMN_SEPARATOR.$foreignValueField;
        }
        
        $where = $this->getMany2manyWhere($field, $valuesColumn);
        $linkTable = $field->getLinkTable();

        $sql = "STUFF((
                    SELECT
                        '::' + (".$valuesColumn.")".
                  ' FROM '.
                        $foreignTable.
                  ' JOIN '.$linkTable.
                    ' ON ('.$foreignTable.'.'.$field->getForeignFieldKey().
                        ' = '.$linkTable.'.'.$field->getLinkForeignField().') 
                    WHERE '.
                        $field->getLinkField().' = '.
                        $this->store->getName().'.'.$this->store->getPrimaryKey().
                        $where." FOR XML PATH(''), TYPE).value('.', 'NVARCHAR(MAX)'), 1, 2, '')";

        return $sql;
    } // end getMany2manyQuery

    protected function getQueryHavingByFieldFilter(IStoreField $field, $filterValue, bool $hasExpression): array
    {
        $fieldName = $field->getName();

        $having = array();
        if ($hasExpression) {
            if (is_array($filterValue)) {
                $having['t.'.$fieldName.'&BETWEEN'] = $filterValue;
            } else {
                $having['t.'.$fieldName] = $filterValue;
            }
        } else {
            $having['t.__'.$field->getIndex().'&IS NOT'] = 'NULL';
        }

        return $having;
    } // end getQueryHavingByFieldFilter

    public function createAuditTable(string $auditTableName, string $originalTableName): bool
    {
        try {
            $sql = "SELECT COUNT(*) FROM ".$auditTableName;
            $this->connection->getRow($sql);
        } catch (\Exception $exp) {
            $sql = "SELECT * INTO ".$auditTableName." FROM ".$originalTableName." WHERE 1 = 2";

            $this->connection->query($sql);

            $sql = "ALTER TABLE ".$auditTableName." DROP COLUMN id;";
            $this->connection->query($sql);

            $queries = array(
                "ALTER TABLE ".$auditTableName." ADD __id int identity",
                "ALTER TABLE ".$auditTableName." ADD __action varchar(100) NULL",
                "ALTER TABLE ".$auditTableName." ADD __id_author int NULL",
                "ALTER TABLE ".$auditTableName." ADD __action_date datetime",
                "ALTER TABLE ".$auditTableName." ADD id int"
            );

            foreach ($queries as $query) {
                $this->connection->query($query);
            }
        }

        return true;
    } // end createAuditTable

    protected function getDataForUpdateForeignTableValues(
        string $tableName, string $foreignTableName, array $values, array $foreignValues, ?array $currentRow = null
    ): array
    {
        // XXX: Fixed IDENTITY_INSERT issues
        $data = parent::getDataForUpdateForeignTableValues(
            $tableName,
            $foreignTableName,
            $values,
            $foreignValues,
            $currentRow
        );

        // FIXME: Hardcoded primary key name
        if (!empty($data['values']['id'])) {
            unset($data['values']['id']);
        }

        return $data;
    } // end getDataForUpdateForeignTableValues

    /**
     * @return string|null
     */
    protected function getQueryGroupBy(): ?string
    {
        $groupBy = $this->model->get('groupBy');
        if (!$groupBy) {
            return null;
        }

        // XXX: Column 'X' is invalid in the select list because it is not contained in either an aggregate function
        // or the GROUP BY clause.
        $orderByFieldName = $this->store->getOrderByFieldName();
        if ($orderByFieldName) {
            if (strpos($groupBy, $orderByFieldName) === false) {
                $groupBy .= ', '.$orderByFieldName;
            }
        }

        return ' GROUP BY '.$groupBy;
    } // end getQueryGroupBy

    /**
     * @override
     * @param array $options
     * @return array
     */
    public function loadForeignValues(array $options): array
    {
        $values = parent::loadForeignValues($options);

        foreach ($values as &$value) {
            if (is_array($value)) {
                unset($value['__row_num']);

                $value = array_shift($value);
            }
        }

        return $values;
    }
}
