<?php

require_once FS_RESOURCES_ROOT.'plugins'.DIRECTORY_SEPARATOR.'TestPlugin.php';

class EntityTest extends FestiTestCase
{
    /** @var Entity */
    protected $entity;

    protected function setUp(): void
    {
        $this->entity = new Entity();
    }

    public function testGetPreparedDataObject(): void
    {
        $fields = array(
            'object' => array(
                Entity::OPTION_KEY_TYPE        => Entity::FIELD_TYPE_OBJECT,
                Entity::OPTION_KEY_OBJECT_TYPE => TestPlugin::class,
            ),
        );

        $request = array(
            'object' => new TestPlugin(),
        );

        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertEmpty($errors);
        $this->assertArrayHasKey('object', $data);
        $this->assertInstanceOf(TestPlugin::class, $data['object']);
    }

    public function testGetPreparedDataObjectWithoutType(): void
    {
        $fields = array(
            'object' => array(
                Entity::OPTION_KEY_TYPE => Entity::FIELD_TYPE_OBJECT,
            ),
        );

        $request = array(
            'object' => new TestPlugin(),
        );

        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertEmpty($errors);
        $this->assertArrayHasKey('object', $data);
        $this->assertInstanceOf(TestPlugin::class, $data['object']);
    }

    public function testGetPreparedDataObjectIncorrectType(): void
    {
        $fields = array(
            'object' => array(
                Entity::OPTION_KEY_TYPE        => Entity::FIELD_TYPE_OBJECT,
                Entity::OPTION_KEY_OBJECT_TYPE => Core::class,
            ),
        );

        $request = array(
            'object' => new TestPlugin(),
        );

        $errors = array();
        $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertArrayHasKey('object', $errors);
    }

    public function testGetPreparedDataWithDefault(): void
    {
        $defaultValue = 0;

        $fields = array(
            'page' => array(
                Entity::OPTION_KEY_TYPE    => Entity::FIELD_TYPE_INT,
                Entity::OPTION_KEY_DEFAULT => $defaultValue,
            ),
        );

        $request = array();

        $this->_checkDefaultValue($request, $fields, $defaultValue);

        $defaultValue = 3;

        $fields['page'][Entity::OPTION_KEY_DEFAULT] = $defaultValue;
        $request['page'] = 0;

        $this->_checkDefaultValue($request, $fields, $request['page']);
    }

    private function _checkDefaultValue(array $request, array $fields, int $excepted): void
    {
        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertEmpty($errors);
        $this->assertArrayHasKey('page', $data);
        $this->assertSame($excepted, $data['page']);
    }

    public function testNullValueWithoutDefault(): void
    {
        $fields = array(
            'link' => Entity::FIELD_TYPE_SECURITY_STRING,
        );

        $request = array(
            'link' => null,
        );

        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertEmpty($errors);
        $this->assertArrayHasKey('link', $data);
        $this->assertNull($data['link']);
    }

    public function testRequiredZeroValue(): void
    {
        $fields = array(
            'page' => array(
                Entity::OPTION_KEY_TYPE     => Entity::FIELD_TYPE_INT,
                Entity::OPTION_KEY_REQUIRED => true,
            ),
        );

        $request = array(
            'page' => 0,
        );

        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertEmpty($errors);
        $this->assertArrayHasKey('page', $data);
        $this->assertEquals($request['page'], $data['page']);
    }

    private function _getUnsignedIntRule(): callable
    {
        return static function ($value, ?string &$error) {
            if (!is_numeric($value)) {
                $error = 'Must be integer!';
                return false;
            }

            if ($value < 1) {
                $error = 'Must be greater than 0!';
                return false;
            }

            return $value;
        };
    }

    public function testGetPreparedDataCustomRulePositive(): void
    {
        $unsignedIntRule = $this->_getUnsignedIntRule();

        $fields = array(
            'id' => array(
                Entity::OPTION_KEY_RULE => $unsignedIntRule,
            ),
        );

        $request = array(
            'id' => 5,
        );

        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertEmpty($errors);
        $this->assertArrayHasKey('id', $data);
    }

    public function testGetPreparedDataCustomRuleNegative(): void
    {
        $unsignedIntRule = $this->_getUnsignedIntRule();

        $fields = array(
            'id' => array(
                Entity::OPTION_KEY_RULE => $unsignedIntRule,
            ),
        );

        $request = array(
            'id' => -5,
        );
        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertNotEmpty($errors);
        $this->assertArrayHasKey('id', $data);
        $this->assertNull($data['id']);
        $this->assertArrayHasKey('id', $errors);
        $this->assertEquals($errors['id'], 'Must be greater than 0!');

        $request = array(
            'id' => 'test',
        );
        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertNotEmpty($errors);
        $this->assertArrayHasKey('id', $data);
        $this->assertNull($data['id']);
        $this->assertArrayHasKey('id', $errors);
        $this->assertEquals($errors['id'], 'Must be integer!');
    }

    public function testGetPreparedDataCustomRuleInvalid(): void
    {
        $fields = array(
            'id' => array(
                Entity::OPTION_KEY_RULE => 'not callable rule',
            ),
        );

        $request = array(
            'id' => -5,
        );

        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);

        $this->assertEmpty($errors);
        $this->assertArrayHasKey('id', $data);
    }
    
    public function testRequiredSecurityStringEmpty(): void
    {
        $fields = array(
            'status' => array(
                Entity::OPTION_KEY_TYPE     => Entity::FIELD_TYPE_SECURITY_STRING,
                Entity::OPTION_KEY_REQUIRED => true,
            ),
        );
        
        $request = array(
            'status' => '',
        );
        
        $errors = array();
        $data = $this->entity->getPreparedData($request, $fields, $errors);
        
        $this->assertNotEmpty($errors);
        $this->assertArrayHasKey('status', $errors);
        $this->assertArrayHasKey('status', $data);
        $this->assertNull($data['status']);
        $this->assertFalse($errors['status']);
    }
}
