<?php

namespace tests\Bundle\Store\Action;

use AbstractField;
use DataAccessObject;
use FestiTestCase;
use ForeignKeyField;
use ForeignKeyLoadAction;
use IDataAccessObject;
use Many2manyField;
use Response;
use Store;
use StoreActionException;

class ForeignKeyLoadActionTest extends FestiTestCase
{
    protected const AUTOCOMPLETE_LIMIT = 15;

    public function testLoadForeignFieldValues(): void
    {
        $store = $this->core->createStoreInstance('dgs_foreignkey_autocomplete');

        $fieldName = 'id_param';
        $foreignField = $store->getModel()->getFieldByName($fieldName);
        $this->assertTrue($foreignField instanceof ForeignKeyField);

        $this->assertTrue($foreignField->isCustomLoadValues());

        $this->assertMatchesRegularExpression(
            "#".Store::ACTION_FOREIGN_KEY_LOAD."#Umis",
            $foreignField->getAutocompleteUrl()
        );

        //
        $foreignKeyLoadAction = $store->createActionInstance(Store::ACTION_FOREIGN_KEY_LOAD);

        $this->assertTrue($foreignKeyLoadAction instanceof ForeignKeyLoadAction);

        //
        $values = array(
            array('caption' => 'Test Autocomplete 00'),
            array('caption' => 'Test Autocomplete 01'),
            array('caption' => 'Test Autocomplete 02'),
            array('caption' => 'Test Autocomplete 03'),
            array('caption' => 'Test Autocomplete 04'),
            array('caption' => 'Test Autocomplete 05'),
            array('caption' => 'Test Autocomplete 06'),
            array('caption' => 'Test Autocomplete 07'),
            array('caption' => 'Test Autocomplete 08'),
            array('caption' => 'Test Autocomplete 09'),
            array('caption' => 'Test Autocomplete 10'),
            array('caption' => 'Test Autocomplete 11'),
            array('caption' => 'Test Autocomplete 12'),
            array('caption' => 'Test Autocomplete 13'),
            array('caption' => 'Test Autocomplete 14'),
            array('caption' => 'Test Autocomplete 15'),
            array('caption' => 'Test Autocomplete 16'),
        );

        $this->core->db->massInsert("settings", $values);

        //
        $response = new Response();

        $_REQUEST = array(
            $store->getIdent() => array(
                ForeignKeyLoadAction::REQUEST_PARAM_FIELD_NAME => $fieldName,
            ),
        );

        $foreignKeyLoadAction->onStart($response);

        $this->assertObjectHasAttribute('results', $response);

        $this->assertTrue(is_array($response->results));

        $this->assertEquals(count($response->results), static::AUTOCOMPLETE_LIMIT);

        foreach ($response->results as $i => $result) {
            $this->assertArrayHasKey($i, $values);
            $expectedValue = $values[$i];

            $this->assertTrue(array_key_exists('key', $result), 'Response format mismatch');
            $this->assertTrue(array_key_exists('value', $result), 'Response format mismatch');

            $this->assertEquals($expectedValue['caption'], $result['key']);
        }
    } // end testLoadForeignFieldValues

    /**
     * @return void
     */
    public function testRelatedFields(): void
    {
        $core = $this->getCoreWithConnection();

        $parentValues = array(
            'id'        => 1,
            'id_parent' => null,
            'caption'   => 'param 1',
            'value'     => 12,
        );

        $childValues = array(
            'id'        => 2,
            'id_parent' => 1,
            'caption'   => 'child param 1',
            'value'     => 14,
        );

        if ($core->db->getDatabaseType() === DataAccessObject::TYPE_MSSQL) {
            $core->db->query('SET IDENTITY_INSERT settings ON');
        }

        $core->db->insert('settings', $parentValues);
        $core->db->insert('settings', $childValues);

        if ($core->db->getDatabaseType() === DataAccessObject::TYPE_MSSQL) {
            $core->db->query('SET IDENTITY_INSERT settings OFF');
        }

        $this->_testRelatedForeignKeyFields($parentValues, $childValues);
        $this->_testRelatedSelectFields($parentValues, $childValues);
    }

    /**
     * @param array $parentValues
     * @param array $childValues
     * @return void
     */
    private function _testRelatedForeignKeyFields(array $parentValues, array $childValues): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance('dgs_foreignkey_related');

        $foreignKeyLoadAction = $store->createActionInstance(Store::ACTION_FOREIGN_KEY_LOAD);

        assert($foreignKeyLoadAction instanceof ForeignKeyLoadAction);

        //
        $fieldName = 'id_param';
        $childFieldName = 'id_child_param';

        $_REQUEST = array(
            $store->getIdent() => array(
                ForeignKeyLoadAction::REQUEST_PARAM_AJAX_CHILD        => $childFieldName,
                ForeignKeyLoadAction::REQUEST_PARAM_AJAX_CHILD_VALUES => array($childFieldName => $childValues['id']),
                ForeignKeyLoadAction::REQUEST_PARAM_AJAX_PARENT       => $fieldName,
            ),
        );

        $response = new Response();
        $hasException = false;
        try {
            $foreignKeyLoadAction->onStart($response);
        } catch (StoreActionException $exp) {
            $hasException = true;
        }

        $this->assertTrue($hasException);

        //
        $_REQUEST[ForeignKeyLoadAction::REQUEST_PARAM_VALUE] = $parentValues['id'];
        $foreignKeyLoadAction->onStart($response);

        $this->assertTrue($response->isFlush());
        $this->assertNotEmpty($response->getContent());

        $values = $foreignKeyLoadAction->getValues();

        $this->assertTrue(array_key_exists($childValues['id'], $values));
        $this->assertTrue($values[$childValues['id']] == $childValues['caption']);

        //
        $this->_testWhereOption($foreignKeyLoadAction, $store, $parentValues['id']);

        $this->_testScopeConditions($foreignKeyLoadAction, $store, $parentValues['id']);
    }

    /**
     * @param array $parentValues
     * @param array $childValues
     * @return void
     */
    private function _testRelatedSelectFields(array $parentValues, array $childValues): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance('dgs_select_related');

        $foreignKeyLoadAction = $store->createActionInstance(Store::ACTION_FOREIGN_KEY_LOAD);

        assert($foreignKeyLoadAction instanceof ForeignKeyLoadAction);

        //
        $fieldName = 'id_param';
        $childFieldName = 'id_child_param';

        $_REQUEST = array(
            $store->getIdent() => array(
                ForeignKeyLoadAction::REQUEST_PARAM_AJAX_CHILD        => $childFieldName,
                ForeignKeyLoadAction::REQUEST_PARAM_AJAX_CHILD_VALUES => array(
                    $childFieldName => $childValues['caption'],
                ),
                ForeignKeyLoadAction::REQUEST_PARAM_AJAX_PARENT => $fieldName,
            ),
        );

        $response = new Response();

        $hasException = false;
        try {
            $foreignKeyLoadAction->onStart($response);
        } catch (StoreActionException $exp) {
            $hasException = true;
        }

        $this->assertTrue($hasException);

        //
        $_REQUEST[ForeignKeyLoadAction::REQUEST_PARAM_VALUE] = $parentValues['caption'];
        $foreignKeyLoadAction->onStart($response);

        $this->assertTrue($response->isFlush());
        $this->assertNotEmpty($response->getContent());

        $values = $foreignKeyLoadAction->getValues();

        $this->assertTrue(array_key_exists($parentValues['id'], $values));
        $this->assertTrue($values[$parentValues['id']] == $parentValues['caption']);
    }

    private function _testWhereOption(
        ForeignKeyLoadAction $foreignKeyLoadAction,
        Store $store,
        $idParentParam
    ): void
    {
        $_REQUEST = array(
            $store->getIdent() => array(
                ForeignKeyLoadAction::REQUEST_PARAM_FIELD_NAME => 'id_param',
                ForeignKeyLoadAction::REQUEST_PARAM_VALUE      => $idParentParam,
            ),
        );

        $response = new Response();
        $foreignKeyLoadAction->onStart($response);

        $values = $foreignKeyLoadAction->getValues();

        $this->assertTrue(count($values) == 1);
        $this->assertTrue(array_key_exists($idParentParam, $values));
    } // end _testWhereOption

    private function _testScopeConditions(
        ForeignKeyLoadAction $foreignKeyLoadAction,
        Store $store,
        $idParentParam
    ): void
    {
        $idParamField = &$store->getModel()->getFieldByName('id_param');

        $cases = array(
            array(
                'scope' => 'G',
                'handler' => function () use ($idParentParam) {
                    $GLOBALS['id_param'] = $idParentParam;
                },
                'count' => 1
            ),
            array(
                'scope' => 'S',
                'handler' => function () use ($idParentParam) {
                    $_SESSION['id_param'] = $idParentParam;
                },
                'count' => 1
            ),
            array(
                'scope' => 'C',
                'handler' => function () use ($store, $idParentParam) {
                    $search = &$store->getModel()->getSearch();
                    $search['id_param'] = $idParentParam;
                },
                'count' => 1
            ),
            array(
                'scope' => 'S',
                'handler' => function () use ($store, $idParentParam) {
                   unset($_SESSION['id_param']);
                },
                'count' => 0
            )
        );

        foreach ($cases as $helper) {
            $scope = $helper['scope'];
            $initHandler = $helper['handler'];

            $initHandler();
            $idParamField->set(ForeignKeyField::OPTION_VALUES_WHERE, 'id='.$scope.'%id_param%');
            $_REQUEST = array(
                $store->getIdent() => array(
                    ForeignKeyLoadAction::REQUEST_PARAM_FIELD_NAME => 'id_param',
                ),
            );

            $response = new Response();
            $foreignKeyLoadAction->onStart($response);

            $values = $foreignKeyLoadAction->getValues();

            $this->assertTrue(count($values) == $helper['count']);
            if ($helper['count']) {
                $this->assertTrue(array_key_exists($idParentParam, $values));
            }

        }
    }

    public function testLoadManyToManyAutocompleteField(): void
    {
        $value               = 'loadManyToManyAutocompleteField';
        $fieldName           = 'settings';
        $valuesCount         = 15;
        $autocompleteLimit   = 10;
        $insertedValuesCount = 11;

        $values = $this->_createSettings($value, $valuesCount);

        $store = $this->core->createStoreInstance('many2many_autocomplete');

        $field = &$store->getModel()->getFieldByName($fieldName);
        assert($field instanceof Many2manyField);

        $field->set(AbstractField::OPTION_AUTOCOMPLETE_LIMIT, $autocompleteLimit);
        $field->set('foreignOrderBy', 'caption ASC');

        $insertedValues = array_slice($values, 0, $insertedValuesCount);
        $proposedValues = array_slice($values, $insertedValuesCount);

        $siteContent           = $this->_createSiteContent($store, $value, $insertedValues);
        $siteContentPrimaryKey = (int) $siteContent['id'];

        $fieldValues = $this->_getManyToManyAutocompleteFieldValues($store, $field, $siteContentPrimaryKey);

        $this->_testManyToManyAutocompleteLoadInsertedValues($store, $fieldName, $fieldValues, $insertedValues);
        $this->_testManyToManyAutocompleteLoadProposedValues($store, $fieldName, $fieldValues, $value, $proposedValues);

        $this->_removeSettings($values);
        $this->_removeSiteContent($siteContent);
    }

    /**
     * @param Store $store
     * @param Many2manyField $field
     * @param int $primaryKey
     * @return array
     */
    private function _getManyToManyAutocompleteFieldValues(Store $store, Many2manyField $field, int $primaryKey): array
    {
        $options = $field->getAttributes();
        $options['onlySelectedValues'] = true;

        return $store->getProxy()->loadForeignAssigns($primaryKey, $options);
    }

    /**
     * @param Store $store
     * @param string $fieldName
     * @param array $fieldValues
     * @param array $expectedValues
     * @return void
     */
    private function _testManyToManyAutocompleteLoadInsertedValues(
        Store &$store, string $fieldName, array $fieldValues, array $expectedValues
    ): void
    {
        $action = $store->createActionInstance(Store::ACTION_FOREIGN_KEY_LOAD);

        $_REQUEST = array(
            $store->getIdent() => array(
                Store::ACTION_KEY_IN_REQUEST                   => Store::ACTION_FOREIGN_KEY_LOAD,
                ForeignKeyLoadAction::REQUEST_PARAM_FIELD_NAME => $fieldName,
                ForeignKeyLoadAction::REQUEST_PARAM_VALUE      => join(',', array_keys($fieldValues)),
            ),
        );

        $response = new Response();
        $action->onStart($response);

        $this->assertEquals(count($expectedValues), count($response->results));

        foreach ($response->results as $i => $result) {
            $this->assertArrayHasKey($i, $expectedValues);
            $setting = $expectedValues[$i];

            $this->assertEquals($setting['caption'], $result['key']);
        }
    }

    /**
     * @param Store $store
     * @param string $fieldName
     * @param array $fieldValues
     * @param string $value
     * @param array $expectedValues
     * @return void
     */
    private function _testManyToManyAutocompleteLoadProposedValues(
        Store &$store, string $fieldName, array $fieldValues, string $value, array $expectedValues
    ): void
    {
        $action = $store->createActionInstance(Store::ACTION_FOREIGN_KEY_LOAD);

        $_REQUEST = array(
            $store->getIdent() => array(
                Store::ACTION_KEY_IN_REQUEST                   => Store::ACTION_FOREIGN_KEY_LOAD,
                ForeignKeyLoadAction::REQUEST_PARAM_FIELD_NAME => $fieldName,
                ForeignKeyLoadAction::REQUEST_PARAM_TERM       => $value,
                ForeignKeyLoadAction::REQUEST_PARAM_EXCLUDE    => join(',', array_keys($fieldValues)),
            ),
        );

        $response = new Response();
        $action->onStart($response);

        $this->assertEquals(count($expectedValues), count($response->results));

        foreach ($response->results as $i => $result) {
            $this->assertArrayHasKey($i, $expectedValues);
            $setting = $expectedValues[$i];

            $this->assertEquals($setting['caption'], $result['key']);
        }
    }

    /**
     * @param Store $store
     * @param string $name
     * @param array $settings
     * @return array
     */
    private function _createSiteContent(Store &$store, string $name, array $settings): array
    {
        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $values = array(
            'name' => $name,
        );
        
        FestiTestCase::preparePostForStore(Store::ACTION_INSERT, $values, $store);

        $_POST['m2m_site_contents2settings'] = array_column($settings, 'id');

        $response = new Response();
        $action->onStart($response);

        $search = $values;

        return $store->loadRow($search);
    }

    /**
     * @param string $prefix
     * @param int $count
     * @return array
     */
    private function _createSettings(string $prefix, int $count): array
    {
        $core = $this->getCoreWithConnection();

        $startIndex = 1;

        $values = array();
        for ($index = $startIndex; $index <= $count; $index++) {
            $values[] = array(
                'caption' => $prefix.$index,
            );
        }

        $core->db->massInsert('settings', $values);

        $sql = 'SELECT * FROM settings';

        $search = array(
            'caption&LIKE' => $prefix.'%',
        );

        $orderBy = array(
            'caption ASC',
        );

        return $core->db->select($sql, $search, $orderBy, IDataAccessObject::FETCH_ALL);
    }

    /**
     * @param array $settings
     * @return void
     */
    private function _removeSettings(array $settings): void
    {
        $core = $this->getCoreWithConnection();

        $ids = array_column($settings, 'id');

        $search = array(
            'id_setting&IN' => $ids,
        );

        $core->db->delete('site_contents2settings', $search);

        $search = array(
            'id&IN' => $ids,
        );

        $core->db->delete('settings', $search);
    }

    /**
     * @param array $siteContent
     * @return void
     */
    private function _removeSiteContent(array $siteContent): void
    {
        $core = $this->getCoreWithConnection();

        $id = (int) $siteContent['id'];

        $search = array(
            'id' => $id,
        );

        $core->db->delete('site_contents', $search);
    }
}
