<?php

namespace tests\Bundle\Store\Action;

use Core;
use FestiEvent;
use FestiTestCase;
use FieldException;
use InsertAction;
use Response;
use Store;
use StoreModel;
use SystemException;
use Throwable;

class InsertActionTest extends FestiTestCase
{
    private const CAPTION_VALUE = 'Test';
    
    private function _testInsertRequest(&$store, &$action): Response
    {
        $values = array(
            'caption' => static::CAPTION_VALUE,
            'value'   => 'test',
        );
    
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
    
        $response = new Response();
    
        $action->onStart($response);
    
        $data = $store->load();
    
        $this->assertTrue($data[0]['value'] == $values['value']);

        return $response;
    } // end _testInsertRequest
    
    public function testOnStart(): void
    {
        $store = $this->getDefaultStore();
        
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $this->assertInstanceOf(InsertAction::class, $action);
        
        $this->_testInsertForm($action);
        $this->_testInsertRequest($store, $action);
    
    } // end testGetInstanceWithoutOptions
    
    private function _testInsertForm(&$action): void
    {
        $response = new Response();
        
        $action->onStart($response);
        
        $this->assertMatchesRegularExpression(
            "#f-db-form#Umis", 
            $response->content, 
            "Not found insert form."
        );
    } // end _testInsertForm

    public function testRequiredField(): void
    {
        $store = $this->getDefaultStore();
        
        $store->setOption('storeExceptionMode', true);
        
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $this->assertInstanceOf(InsertAction::class, $action);
        
        $values = array(
            'caption' => '',
            'value'   => ''
        );
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
        
        $response = new Response();
        
        try {
            $action->onStart($response);
            $this->fail('Expect FieldException');
        } catch (FieldException $exp) {
            $this->assertTrue(
                $exp->getMessage() === 'Caption is required field'
            );
        }
        
        $store->setOption('storeExceptionMode', false);
        
        /*
        ob_start();
        
        $response->send();
        $content = ob_get_contents();
        
        ob_end_clean();
        
        $this->assertMatchesRegularExpression(
            "#\"type\":\"alert\"#Umis",
            $content,
            "Must be Alert Type"
        );*/
    }
    
    public function testFilterValues(): void
    {
        $store = $this->getDefaultStore();
        $filters = &$store->getModel()->getFilters();
        
        $filterValue = 'test_filter';
        $filters['value'] = $filterValue;
        
        $values = array(
            'caption' => 'caption',
        );
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
        
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $response = new Response();
        
        $action->onStart($response);
        
        $data = $store->load();
        
        $this->assertNotEmpty($data);
        $this->assertTrue($data[0]['value'] == $filterValue);
    }
    
    public function testRedirectUrl(): void
    {
        $store = $this->getDefaultStore();
    
        $action = $store->createActionInstance(Store::ACTION_INSERT);
    
        $actionName = $store->getAction();
        $actionData = &$store->getModel()->getActionByRef($actionName);
    
        $expectedUrl = '/new/test/';
    
        $actionData[StoreModel::ACTION_ATTRIBUTE_REDIRECT_URL] = $expectedUrl;
    
        $response = $this->_testInsertRequest($store, $action);
    
        $this->assertTrue($expectedUrl == $response->url);
    }

    public function testCurrentUrl(): void
    {
        $this->core->setOption('http_base', '/');

        $store = $this->getDefaultStore();
        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $store->setOption(Store::OPTION_CURRENT_URL, '/new/url/');

        $expectedUrl = '/new/url/';

        $response = $this->_testInsertRequest($store, $action);

        $firstStoreUrl = $response->url;

        $this->assertTrue($expectedUrl == $response->url);

        $this->core->setOption('http_base', '/px/');

        $store = $this->getDefaultStore();
        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $store->setOption(Store::OPTION_CURRENT_URL, '/new/url/');

        $expectedUrl = '/px/new/url/';

        $response = $this->_testInsertRequest($store, $action);

        $secondStoreUrl = $response->url;

        $this->assertNotTrue($firstStoreUrl == $secondStoreUrl);
        $this->assertTrue($expectedUrl == $response->url);

        $this->core->setOption('http_base', '/');
    }

    public function testParentTransactionSupportWhenStoreHasError(): void
    {
        $inTransaction = $this->core->db->inTransaction();
        $this->assertFalse($inTransaction);
        
        $this->core->db->begin();
        
        $inTransaction = $this->core->db->inTransaction();
        $this->assertTrue($inTransaction);
        
        $values = array(
            'caption' => 'asdadasd',
            'value'   => 'test_transaction'
        );
        
        $store = $this->getDefaultStore();
        
        $store->setOption('storeExceptionMode', true);
        
        $store->addEventListener(
            Store::EVENT_UPDATE_VALUES, 
            function ($event) {
                throw new SystemException('Test transaction');
            }
        );
        
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
        
        $response = new Response();
        
        $isException = false;
        try {
            $action->onStart($response);
        } catch (Throwable $exp) {
            $isException = true;
        }
        
        $this->assertTrue($isException);
        
        $inTransaction = $this->core->db->inTransaction();
        $this->assertTrue($inTransaction);
        
        $this->core->db->rollback();
        $store->setOption('storeExceptionMode', false);
    }
    
    public function testCustomUpdateAndSetPrimaryKeyUseReference(): void
    {
        $store = $this->getDefaultStore();
        
        $id = 100;
        $store->addEventListener(
            Store::EVENT_BEFORE_INSERT,
            static function (FestiEvent &$event) use ($id) {
                $event->target['isUpdated'] = true;
                $event->target['id'] = $id;
            }
        );
        
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $request = array(
            'caption' => 'asdadasd',
            'value'   => 'test_custom'
        );
        $info = $action->apply($request);
        $this->assertTrue($info['id'] == $id);
    }

    public function testExceptionInEventPrepareValues(): void
    {
        $store = $this->getDefaultStore();
        
        $errorMessage = "Test Exception";
        $store->addEventListener(
            Store::EVENT_PREPARE_VALUES,
            static function (FestiEvent &$event) use ($errorMessage) {
                throw new SystemException($errorMessage);
            }
        );
        
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $request = array(
            'caption' => 'asdadasd',
            'value'   => 'test_custom'
        );
        
        $info = $action->apply($request);
        
        $this->assertNull($info);
        $this->assertTrue($action->getLastError() == $errorMessage);
    }

    public function testInsertQuotes(): void
    {
        $store = $this->getDefaultStore();
    
        $action = $store->createActionInstance(Store::ACTION_INSERT);
    
        $this->assertInstanceOf(InsertAction::class, $action);
        
        $values = array(
            'caption' => 'Test',
            'value'   => 'test\'s'
        );
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
        
        $response = new Response();
        
        $action->onStart($response);
        
        $data = $store->load();
        
        $this->assertTrue($data[0]['value'] == $values['value']);
    } // end testInsertQuotes
    
    public function testOnStartPrefill(): void
    {
        $store = $this->core->createStoreInstance('settings_prefill');
        
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $this->assertInstanceOf(InsertAction::class, $action);
    
        $this->_testInsertRequest($store, $action);
    
        $action = $store->createActionInstance(Store::ACTION_INSERT);
    
        $this->assertInstanceOf(InsertAction::class, $action);
    
        $values = array(
            'value' => 'try'
        );
    
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
    
        $response = new Response();
    
        $action->onStart($response);
    
        $data = $store->load();
        
        $row = array_pop($data);
    
        $this->assertTrue($row['value'] == $values['value']);
        $this->assertTrue($row['caption'] == static::CAPTION_VALUE);
    } // end testOnStartPrefill
}
