<?php

namespace tests\Bundle\Store\Action;

use FestiTestCase;
use ListAction;
use Response;
use Store;
use StoreModel;

class ListActionTest extends FestiTestCase
{
    const TOTAL_COUNT = 2;
    
    public function testOnStart(): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance(STORE_DEFAULT);

        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $this->assertNotEmpty($response->content);
        
        $this->assertInstanceOf(ListAction::class, $response->storeAction);
    } // end testOnStart

    public function testOnStartApiMode(): void
    {
        $this->_onStartApiMode();
    }

    public function testOnStartApiModeWithoutThemes(): void
    {
        $core = $this->getCoreWithConnection();

        $templatePath = $core->getOption('theme_template_path');

        $this->_onStartApiMode('empty_templates');

        $core->setOption('theme_template_path', $templatePath);
    } // end testOnStartApiModeWithoutThemes
    
    public function testOnCountTotal(): void
    {
        $core  = $this->getCoreWithConnection();
    
        $store = $core->createStoreInstance('settings_total');
        
        $this->_prepareStoreForCountTotal($store);
      
        $action = $store->createActionInstance(Store::ACTION_LIST);
    
        $this->assertInstanceOf(ListAction::class, $action);
    
        $response = new Response();
    
        $action->onStart($response);
        
        $this->assertNotEmpty($response->content);
        
        $total = $store->getTotalCount();
        
        $this->assertEquals($total, static::TOTAL_COUNT, "Total Not Equals");
      
        $this->assertInstanceOf(ListAction::class, $response->storeAction);
    } // end testOnCountTotal
    
    private function _onStartApiMode($themeTemplatePath = false): void
    {
        $core = $this->getCoreWithConnection();

        if ($themeTemplatePath) {
            $core->setOption(
                'theme_template_path',
                FS_THEME_ROOT.$themeTemplatePath.DIRECTORY_SEPARATOR
            );
        }

        $store  = $core->createStoreInstance('settings_api');

        $this->_prepareStoreApiMode($store);

        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $response = $response->getArrayCopy();
        
        $this->assertTrue(array_key_exists('rowsPerPage', $response));
        $this->assertTrue(array_key_exists('totalRows', $response));

        $this->assertNotEmpty($response['rowsPerPage']);
        $this->assertNotEmpty($response['totalRows']);
    } // end testOnStartApiMode

    private function _prepareStoreApiMode(Store $store): void
    {
        $action = $store->createActionInstance(Store::ACTION_BATCH_INSERT);

        $values = array(
            array(
                'caption' => 'Test',
                'value'   => '1'
            ),
            array(
                'caption' => 'Test',
                'value'   => '2'
            ),
        );

        $filePath = $this->_getRequestPath();

        $this->_prepareStream($action, $values, $filePath);

        $response = new Response();

        $action->onUpdate($response);

        $this->assertTrue((bool) unlink($filePath));

        $this->assertEmpty($action->getLastError());
    } // end _prepareStoreForApi

    public function testOnStartFieldExpressionAndOrderBy(): void
    {
        $core = $this->getCoreWithConnection();

        $store  = $core->createStoreInstance('order_expression_field');

        $this->_prepareStore($store, 2);

        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $this->assertNotEmpty($response->content);
    } // end testOnStartFieldExpressionAndOrderBy

    public function testTestFilterWithAmbiguousColumns(): void
    {
        $filters = array(
            'caption' => 'test'
        );
    
        $this->_testFilters($filters);
    } // end testTestFilterWithAmbiguousColumns

    public function testOnFiltersXSSInjection(): void
    {
        $action  = false;
        
        $filters = array(
            'caption' => "<script>alert('XSS')</script>",
        );
    
        $this->_testFilters($filters, $action);
    
        $this->assertInstanceOf(ListAction::class, $action);
        
        $session = $action->getStore()->getSession();
        
        $sessionFiltersKey = Store::FILTERS_KEY_IN_SESSION;
        
        $this->assertArrayHasKey($sessionFiltersKey, $session);
    
        $sessionFilters   = $session[$sessionFiltersKey];
    
        $filtersKey       = 'filters';
        $captionFilterKey = 'filters_caption';
        
        $this->assertArrayHasKey($filtersKey, $sessionFilters);
        $this->assertArrayHasKey(
            $captionFilterKey,
            $sessionFilters[$filtersKey]
        );
  
        $captionFilter = $sessionFilters[$filtersKey][$captionFilterKey];
        
        $this->assertNotEquals($filters['caption'], $captionFilter);
    } // end testOnFiltersXSSInjection
    
    private function _testFilters($filters, &$action = false): void
    {
        $core = $this->getCoreWithConnection();
    
        $store = $core->createStoreInstance('filters');
    
        FestiTestCase::prepareFiltersForStore($filters, $store);
    
        $action = $store->createActionInstance(Store::ACTION_LIST);
    
        $response = new Response();
        
        $action->onStart($response);

        $this->assertNotEmpty($response->getContent());
    } // end _testFilters
    
    private function _prepareStream($action, $values, $filePath): void
    {
        $isCreate = file_put_contents($filePath, json_encode($values));

        $this->assertTrue((bool) $isCreate);

        $stream = &$action->getStore()->getModel()->getStream();
        $stream = $filePath;
    } // end _prepareStream

    private function _getRequestPath(): string
    {
        return FS_RESOURCES_ROOT.'import/request.json';
    } // end _getRequestPath

    public function testStoreActionsMode(): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance(STORE_DEFAULT);

        $this->assertEquals(
            $store->getModel()->getOption(StoreModel::OPTION_ACTIONS_MODE),
            StoreModel::OPTION_ACTIONS_MODE_LIST
        );

        $actions = &$store->getModel()->getActions();
        $options = &$store->getModel()->getOptions();

        $actions[Store::ACTION_EDIT]['mode'] = 'new';
        $actions[Store::ACTION_INFO]['mode'] = 'default';

        $options[StoreModel::OPTION_ACTIONS_MODE] =
            StoreModel::OPTION_ACTIONS_MODE_LIST;

        $this->_prepareStore($store, 2);

        $content = $this->_getListContent($store);

        $this->assertMatchesRegularExpression(
            "#<a id=\"action-item-\w+\" href=\"\/\?".STORE_DEFAULT.
            "%5Baction%5D=edit&".STORE_DEFAULT.
            "%5BID%5D=\d+\"#Umis",
            $content,
            'Not found Action button'
        );
        $this->assertMatchesRegularExpression(
            "#<a id=\"action-item-\w+\" href=\"\/\?".STORE_DEFAULT.
            "%5Baction%5D=info&".STORE_DEFAULT.
            "%5BID%5D=\d+\&popup=true\"#Umis",
            $content,
            'Not found Action button'
        );

        $options[StoreModel::OPTION_ACTIONS_MODE] =
            StoreModel::OPTION_ACTIONS_MODE_BUTTONS;

        $content = $this->_getListContent($store);

        $this->assertMatchesRegularExpression(
            "#<a id=\"action-item-\w+\" href=\"\/\?".STORE_DEFAULT.
            "%5Baction%5D=edit&".STORE_DEFAULT.
            "%5BID%5D=\d+\"[^>]+class=\"btn btn-default#Umis",
            $content,
            'Not found Action button'
        );

        $this->assertMatchesRegularExpression(
            "#<a id=\"action-item-\w+\" href=\"\/\?".STORE_DEFAULT.
            "%5Baction%5D=info&".STORE_DEFAULT.
            "%5BID%5D=\d+\&popup=true\"[^>]+class=\"btn btn-default#Umis",
            $content,
            'Not found Action button'
        );

        $options = $this->_getOptions();

        $store = $this->core->createStoreInstance(STORE_DEFAULT, $options);

        $this->assertEquals(
            $store->getModel()->getOption(StoreModel::OPTION_ACTIONS_MODE),
            StoreModel::OPTION_ACTIONS_MODE_LIST
        );
    } // end testStoreActionsMode

    private function _getListContent($store): string
    {
        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $this->assertNotEmpty($response->content);

        return $response->content;
    } // end _getListContent

    public function testPagination(): void
    {
        $core = $this->getCoreWithConnection();

        $options = $this->_getOptions();
        $options['model']['table']['rowsForPage'] = 2;

        $store = $core->createStoreInstance(STORE_DEFAULT, $options);

        $this->_prepareStore($store, 4);

        $this->_testPaging($store, array(), 1);

        $values = array(
            $store->getIdent() => array(
                Store::PAGE_INDEX_KEY_IN_REQUEST => 2
            )
        );

        $this->_testPaging($store, $values, 2);

        // Test compability with old paging
        $values = array(
            Store::PAGE_INDEX_KEY_IN_REQUEST => 1
        );

        $this->_testPaging($store, $values, 1);
    }

    private function _testPaging(&$store, $values, $expectPage): void
    {
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );

        $store->setCurrentPageIndex(null);
        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $content = $response->getContent();
        $this->assertNotEmpty($content);

        $this->assertMatchesRegularExpression(
            "#<a href=\"\/\?".$store->getIdent()."%5BpageID%5D=\d+\" ".
            "class=\"[^\"]+\">\d+<\/a>#Umis",
            $content,
            "Not found page url"
        );

        $this->assertEquals($expectPage, $store->getCurrentPageIndex());
    }

    private function _prepareStore(Store $store, $count): void
    {
        $action = $store->createActionInstance(Store::ACTION_BATCH_INSERT);

        $values = array();

        foreach (range(1, $count) as $i) {
            $values[] = array(
                'caption' => 'Test',
                'value'   => $i
            );
        }

        $response = new Response();

        $_REQUEST = $values;

        $action->onUpdate($response);

        $this->assertEmpty($action->getLastError());
    } // end _prepareStore

    private function _getOptions(): array
    {
        $options = array();

        $options['model'] = array(
            'table'   => array(
                'name'          => STORE_DEFAULT,
                'primaryKey'    => 'id',
                'exceptionMode' => true,
            ),

            'fields'  => array(
                'caption' => array(
                    'type'      => 'text',
                    'name'      => 'caption',
                    'caption'   => 'Caption',
                    'required'  => 'true',
                ),
                'value' => array(
                    'type'      => 'textarea',
                    'name'      => 'value',
                    'caption'   => 'Value',
                    'required'  => 'true',
                )
            ),

            'actions' => array(
                Store::ACTION_LIST => array(
                    'type'    => Store::ACTION_LIST,
                    'caption' => 'List'
                )
            )
        );

        return $options;
    } // end _getOptions

    public function testInsertFastForm(): void
    {
        $core = $this->getCoreWithConnection();

        $options = $this->_getOptions();
        $options['model']['actions'][Store::ACTION_INSERT] = array(
            'type'    => Store::ACTION_INSERT,
            'fast'    => 'true',
            'caption' => 'Create'
        );

        $store = $core->createStoreInstance(STORE_DEFAULT, $options);

        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $content = $response->getContent();
        $this->assertNotEmpty($content);

        $this->assertMatchesRegularExpression(
            "#<form class=\"form-horizontal f-db-form f-db-form-\w+\"#Umis",
            $content,
            'Not found fast form'
        );
    } // end testInsertFastForm

    public function testDisableInsertFastForm(): void
    {
        $core = $this->getCoreWithConnection();

        $options = $this->_getOptions();
        $options['model']['actions'][Store::ACTION_INSERT] = array(
            'type'    => Store::ACTION_INSERT,
            'fast'    => 'false',
            'caption' => 'Create'
        );

        $store = $core->createStoreInstance(STORE_DEFAULT, $options);

        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $content = $response->getContent();
        $this->assertNotEmpty($content);

        $this->assertDoesNotMatchRegularExpression(
            "#<form class=\"form-horizontal f-db-form f-db-form-\w+\"#Umis",
            $content,
            "Not found fast form"
        );
    } // end testDisableInsertFastForm
    
    private function _prepareStoreForCountTotal(Store &$store): void
    {
        $values = array(
            'caption' => 'TestTotal',
            'value'   => '1'
        );
    
        $this->_insertValues($values, $store);
    
        $values = array(
            'caption' => 'TestTotalTwo',
            'value'   => '2'
        );
    
        $this->_insertValues($values, $store);
    } // end _prepareStoreForCountTotal
    
    private function _insertValues(array $values, Store &$store): void
    {
        $action = $store->createActionInstance(Store::ACTION_INSERT);

        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
    
        $response = new Response();
    
        $action->onUpdate($response);
        
        $this->assertEmpty($action->getLastError());
    } // end _insertValues
}
