<?php

namespace tests\Bundle\Store\Action;

use FestiEvent;
use FestiTestCase;
use Response;
use StageBuilder;
use Store;

class RelationActionTest extends FestiTestCase
{
     /**
     * Bug 3322
     * @link http://localhost.in.ua/issues/3322
     */
    public function testDoParentRelation(): void
    {
        $core = $this->getCoreWithConnection();
        
        $store = $core->createStoreInstance('site_contents');

        $this->_testInsertRequest($store);
        $this->_testInsertRequest($store);
        
        $data = $store->load();

        $idParent = $data[0]['id'];
        $idChild = $data[1]['id'];
        
        $values = array(
            'id_parent' => $idParent
        );
        
        $store->updateByPrimaryKey($idChild, $values);
        
        $allowIDs = array(
            $idParent,
            $idChild
        );
        $this->_addAllowedTableRow($store, $allowIDs);

        $this->_doRequestActionChild($store, $idChild);
        $this->_doRequestActionParent($store);
        
        $this->assertEquals($this->_getParentIDBySession($store), $idParent);
    } // end testDoParentRelation
    
    private function _testInsertRequest(&$store): void
    {
        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $values = array(
           'name' => 'Test'.rand(1, 100),
           'path'   => 'testDir'
        );
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT, 
            $values, 
            $store
        );
       
        $response = new Response();
        
        $action->onStart($response);
        
        $data = $store->load();

        $this->assertTrue($data[0]['name'] == $values['name']);
    } // end _testInsertRequest
    
    private function _getParentIDBySession($store)
    {
        return $store->getSessionValue(Store::PARENT_PRIMARY_KEY_IN_SESSION);
    } // end _getParentIDBySession
    
    private function _addAllowedTableRow(&$store, $allowIDs): void
    {
        $store->addAllowedTableRow($allowIDs);
        
        $childID = array_pop($allowIDs);
        FestiTestCase::preparePrimaryKey($store, $childID);
    } // end _addAllowedTableRow
    
    private function _doRequestActionChild(&$store, $idParent): void
    {
        $_REQUEST = array(
            $store->getIdent() => array(
                'action' => Store::ACTION_CHILD,
                'ID'     => $idParent
            )
        );
        
        $action = $store->createActionInstance(Store::ACTION_CHILD);
        $response = new Response();
        
        $expectUrl = '/test/';
        $store->setOption('current_url', $expectUrl);
        
        $action->onStart($response);
        
        $this->assertTrue($expectUrl == $response->url);
        
        // XXX: Check system store URL
        
        $store->setOption('current_url', '/festi/test/');
        
        $expectUrl = '/festi/site_contents/';
        $action->onStart($response);
        
        $this->assertTrue($expectUrl == $response->url);
    } // end _doRequestActionChild
    
    private function _doRequestActionParent(&$store): void
    {
        $_REQUEST = array(
            $store->getIdent() => array(
                'action' => Store::ACTION_PARENT
            )
        );
        
        $action = $store->createActionInstance(Store::ACTION_PARENT);
        
        $response = new Response();
        
        $action->onStart($response);
    } // end _doRequestActionParent

    public function testCreateFromArrayModel(): void
    {
        $relations = array(
            array(
                'type'          => 'parent',
                'field'         => 'id_parent',
                'foreignTable'  => 'site_contents',
                'foreignField'  => 'id',
            ),
            array(
                'type'          => 'child',
                'field'         => 'id',
                'foreignTable'  => 'site_contents',
                'foreignField'  => 'id_parent',
                'cascade'       => 'true',
                'treeCaption'   => 'name',
            )
        );

        $options = $this->_getOptions($relations);
        $store = $this->core->createStoreInstance(STORE_DEFAULT, $options);

        $this->assertNotNull($store);

        $relation = $store->getModel()->getRelation(Store::ACTION_PARENT);
        $this->assertNotFalse($relation);
        $this->assertTrue(array_key_exists('site_contents', $relation));

        $arrayDiff = array_diff($relations[0], $relation['site_contents']);
        $this->assertEquals(0, count($arrayDiff));

        $relation = $store->getModel()->getRelation(Store::ACTION_CHILD);
        $this->assertNotFalse($relation);
        $this->assertTrue(array_key_exists('site_contents', $relation));
        $arrayDiff = array_diff($relations[1], $relation['site_contents']);
        $this->assertEquals(0, count($arrayDiff));
    }

    private function _getOptions($relations): array
    {
        $options = array();

        $options['model'] = array(
            'table'   => array(
                'name'          => 'site_contents',
                'primaryKey'    => 'id',
                'exceptionMode' => true,
            ),

            'relations'  => $relations,

            'fields'  => array(
                'name' => array(
                    'type'      => 'text',
                    'name'      => 'name',
                    'caption'   => 'Name',
                    'required'  => 'true',
                )
            ),

            'actions' => array(
                Store::ACTION_LIST => array(
                    'type'    => Store::ACTION_LIST,
                    'caption' => 'List'
                ),
                Store::ACTION_PARENT => array(
                    'type'    => Store::ACTION_PARENT,
                    'caption' => 'Parent',
                    'relation' => 'site_contents',
                    'relationType' => 'parent'
                ),
                Store::ACTION_CHILD => array(
                    'type'    => Store::ACTION_CHILD,
                    'caption' => 'Child Contents',
                    'relation' => 'site_contents',
                    'relationType' => 'child'
                ),
            )
        );

        return $options;
    } // end _getOptions

    public function testRelationInStageBuilder(): void
    {
        $core = $this->getCoreWithConnection();
        $stage = new StageBuilder(1, StageBuilder::LAYOUT_ROWS);
        $stage->addBlock('test content');

        $_REQUEST['site_contents'] = [
            Store::ACTION_KEY_IN_REQUEST => Store::RELATION_TYPE_CHILD,
        ];
        $store = $core->createStoreInstance('site_contents');
        $idParent = $this->_getParentIDBySession($store);
        
        $store->addAllowedTableRow($idParent);
        FestiTestCase::preparePrimaryKey($store, $idParent);

        $blockIndex = $stage->addStoreBlock($store);

        $_REQUEST[StageBuilder::REQUEST_KEY_BLOCK_INDEX] =
            $blockIndex;

        $response = new Response();
        $stage->onRequest($response);

        $this->assertEquals($response->getAction(), Response::ACTION_REDIRECT);
        $this->assertNotEmpty($response->url);
    } // end testRelationInStageBuilder

    public function testDirectOpenStoreWithoutParent(): void
    {
        $store = $this->core->createStoreInstance('site_contents_parent_relation');

        $response = new Response();
        $store->onRequest($response);

        $this->assertTrue($response->getAction() == Response::ACTION_REDIRECT);
        $this->assertMatchesRegularExpression("#parent#Umis", $response->url);

    } // end testDirectOpenStoreWithoutParent

    /**
     * @return void
     */
    public function testStoreCaption(): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance('site_contents');

        $rootElement        = $this->_createContent($this->getRandomString());
        $secondLevelElement = $this->_createContent($this->getRandomString(), $rootElement['id']);
        $thirdLevelElement  = $this->_createContent($this->getRandomString(), $secondLevelElement['id']);

        $listActionCaption = __('Contents');

        $firstLevelCaption  = $rootElement['name'].' / '.$listActionCaption;
        $secondLevelCaption = $rootElement['name'].' / '.$secondLevelElement['name'].' / '.$listActionCaption;
        $thirdLevelCaption  = $rootElement['name'].' / '.$secondLevelElement['name'].' / '.
                              $thirdLevelElement['name'].' / '.$listActionCaption;

        $this->_testStoreCaption($store, $listActionCaption);

        $this->_doChildAction($store, $rootElement['id']);
        $this->_testStoreCaption($store, $firstLevelCaption);

        $this->_doChildAction($store, $secondLevelElement['id']);
        $this->_testStoreCaption($store, $secondLevelCaption);
        
        $this->_doChildAction($store, $thirdLevelElement['id']);
        $this->_testStoreCaption($store, $thirdLevelCaption);
        
        $this->_doParentAction($store);
        $this->_testStoreCaption($store, $secondLevelCaption);
        
        $this->_doParentAction($store);
        $this->_testStoreCaption($store, $firstLevelCaption);
        
        $this->_doParentAction($store);
        $this->_testStoreCaption($store, $listActionCaption);

        $this->_removeContentByID($rootElement['id']);

        $this->assertEmpty($store->load());
    }

    /**
     * @param Store $store
     * @param string $expectedCaption
     * @return void
     */
    private function _testStoreCaption(Store $store, string $expectedCaption): void
    {
        $isExecutedTestListener = false;

        $testListener = function (FestiEvent $event) use (&$isExecutedTestListener, $expectedCaption) {
            $isExecutedTestListener = true;

            $this->assertEquals($expectedCaption, $event->target['info']['caption']);
        };

        $store->addEventListener(Store::EVENT_ON_LOAD_LIST_DATA, $testListener);

        $response = new Response();
        $store->onRequest($response);

        $this->assertTrue($isExecutedTestListener, 'Test listener is not executed');

        $store->removeEventListener(Store::EVENT_ON_LOAD_LIST_DATA, $testListener);
    }

    /**
     * @param Store $store
     * @param int $primaryKeyValue
     * @return void
     */
    private function _doChildAction(Store &$store, int $primaryKeyValue): void
    {
        FestiTestCase::preparePostForStore(Store::ACTION_CHILD, array(), $store, $primaryKeyValue);

        $response = new Response();
        $store->onRequest($response);

        FestiTestCase::cleanPostByStore($store);
    }

    /**
     * @param Store $store
     * @param int $primaryKeyValue
     * @return void
     */
    private function _doParentAction(Store &$store): void
    {
        FestiTestCase::preparePostForStore(Store::ACTION_PARENT, array(), $store);

        $response = new Response();
        $store->onRequest($response);

        FestiTestCase::cleanPostByStore($store);
    }

    /**
     * @param string $name
     * @param int|null $idParent
     * @return array
     */
    private function _createContent(string $name, ?int $idParent = null): array
    {
        $values = array(
            'name'      => $name,
            'id_parent' => $idParent,
        );
        $values['id'] = (int) $this->core->db->insert('site_contents', $values);

        return $values;
    }

    /**
     * @param int $id
     * @return void
     */
    private function _removeContentByID(int $id): void
    {
        $search = array(
            'id' => $id,
        );
        $this->core->db->delete('site_contents', $search);
    }
}
