<?php

namespace tests\Bundle\Store\Action;

use FestiEvent;
use FestiTestCase;
use PermissionsException;
use RemoveAction;
use Response;
use Store;

class RemoveActionTest extends FestiTestCase
{
    private $_primaryKeyValue;
    private $_onCalledEvent;

    public function testOnStart(): void
    {
        $store = $this->getDefaultStore();

        $this->_testInsertRequest($store);

        $action = $store->createActionInstance(Store::ACTION_REMOVE);
        $this->assertInstanceOf(RemoveAction::class, $action);
        
        $this->_testRemoveForm($store, $action);
        $this->_testRemoveRequest($store, $action);
    } // end testGetInstanceWithoutOptions
    
    public function testEventBeforeRemove(): void
    {
        $store = $this->getDefaultStore();
    
        $this->_testInsertRequest($store);
    
        $action = $store->createActionInstance(Store::ACTION_REMOVE);
        
        $store->addEventListener(
            Store::EVENT_BEFORE_REMOVE,
            array(&$this, 'onCallBeforeRemove')
        );
        
        $this->_testRemoveRequest($store, $action);
        
        $this->assertTrue($this->_onCalledEvent);
    }
    
    public function onCallBeforeRemove(FestiEvent &$event): void
    {
        $this->_onCalledEvent = true;
        
        $values = $event->target['data'];
        
        $this->assertNotEmpty($values);
    }
    
    private function _testRemoveForm(&$store, &$action): void
    {
        FestiTestCase::preparePrimaryKey($store, $this->_primaryKeyValue);
        
        $response = new Response();
        $action->onStart($response);
        
        $this->assertMatchesRegularExpression(
            "#f-db-form#Umis", 
            $response->content, 
            "Not found insert form."
        );
    } // end _testInsertForm
    
    private function _testInsertRequest(&$store): void
    {
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $values = array(
            'caption' => 'Test',
            'value'   => 'test'
        );
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT, 
            $values, 
            $store
        );
        
        $response = new Response();
        $action->onStart($response);
        
        $data = $store->load();
        
        $this->assertTrue(count($data) == 1);
        $this->assertTrue($data[0]['value'] == $values['value']);
        
        $this->_primaryKeyValue = $data[0]['id'];
        $this->preparePrimaryKey($store, $this->_primaryKeyValue);
    
        $store->addAllowedTableRow($this->_primaryKeyValue);
    } // end _testInsertRequest
    
    private function _testRemoveRequest(&$store, &$action): void
    {
        FestiTestCase::preparePostForStore(
            Store::ACTION_REMOVE, 
            array(), 
            $store, 
            $this->_primaryKeyValue
        );
        
        $response = new Response();
        
        $action->onStart($response);
        
        $data = $store->load();
        
        $this->assertTrue(count($data) == 0);
    } // end _testRemoveRequest
    
    public function testRemoveUndefinedField(): void
    {
        $store = $this->getDefaultStore();
        
        $action = $store->createActionInstance(Store::ACTION_REMOVE);
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_REMOVE, 
            array(), 
            $store, 
            999999999
        );

        $response = new Response();
        $isException = false;

        try {
            $action->onStart($response);
            $this->fail('Expected PermissionsException');
        } catch (PermissionsException $exp) {
            $isException = true;
        }
        
        $this->assertTrue($isException);
        
        ob_start();
        
        $response->send();
        $content = ob_get_contents();
        
        ob_end_clean();
        
        $this->assertMatchesRegularExpression(
            "#\"type\":\"alert\"#Umis",
            $content,
            "Must be Alert Type"
        );
    } // end testRemoveUndefinedField

    public function testFailAction(): void
    {
        $store = $this->getDefaultStore($this->_getOptions());

        $action = $store->createActionInstance(Store::ACTION_REMOVE);
        $this->assertInstanceOf(RemoveAction::class, $action);

        $this->_primaryKeyValue = 1;

        $this->preparePrimaryKey($store, $this->_primaryKeyValue);
        $store->addAllowedTableRow($this->_primaryKeyValue);

        FestiTestCase::preparePostForStore(
            Store::ACTION_REMOVE,
            array(),
            $store,
            $this->_primaryKeyValue
        );

        $response = new Response();

        $action->onStart($response);
        
        $this->assertNotEmpty($response->messages);
    }

    private function _getOptions(): array
    {
        $options = array();

        $options['model'] = array(
            'table'   => array(
                'name'          => 'unknown_table',
                'primaryKey'    => 'id',
                'exceptionMode' => false,
            ),

            'fields'  => array(
                'caption' => array(
                    'type'      => 'text',
                    'name'      => 'caption',
                    'caption'   => 'Caption',
                    'required'  => 'true',
                ),
            ),

            'actions' => array(
                Store::ACTION_REMOVE => array(
                    'type'    => Store::ACTION_REMOVE,
                    'caption' => 'Remove'
                )
            )
        );

        return $options;
    } // end _getOptions
}
