<?php

namespace tests\Bundle\Store;

use FestiTestCase;
use InsertAction;
use ListAction;
use Response;
use Store;
use StoreAggregationEvent;
use StoreException;

class AggregationsTest extends FestiTestCase
{
    private $_currentType;

    private const AGGREGATION_FILED_NAME = 'amount';

    public function testOnStart(): void
    {
        $store = $this->getDefaultStore();

        $this->_prepareStore($store);
        $expectedTotals = $this->_calculateAggregations($store);

        $store->addEventListener(
            Store::EVENT_ON_AGGREGATIONS,
            array(&$this, 'onTestAggregationData')
        );

        $aggregations = &$store->getModel()->getAggregations();

        $types = array('sum', 'avg', 'custom');

        foreach ($types as $type) {

            $this->_currentType = $type;

            $aggregations = array(
                array(
                    'field' => static::AGGREGATION_FILED_NAME,
                    'type'  => $type,
                ),
            );

            $totalValue = $this->_getTotalFromList($store);

            $message = 'Type: '.$type.' '.$expectedTotals[ $type ];
            $this->assertEquals($expectedTotals[$type], $totalValue, $message);
        }

        // Exception
        $aggregations = array(
            array(
                'field' => static::AGGREGATION_FILED_NAME,
                'type'  => 'unknown',
            ),
        );

        $hasException = false;
        try {
            $this->_getTotalFromList($store);
        } catch (StoreException $exp) {
            $hasException = true;
        }

        $this->assertTrue($hasException);

        //
        $aggregations = array(
            array('field' => static::AGGREGATION_FILED_NAME, 'type' => 'sum')
        );

        $valuesField = &$store->getModel()
                              ->getFieldByName(static::AGGREGATION_FILED_NAME);

        $valuesField->set('where', 'id < 0');

        $totalValue = $this->_getTotalFromList($store);
        $this->assertTrue($totalValue == 0);
    }

    private function _prepareStore(&$store): void
    {
        $this->_clearStore($store);

        for ($i = 0; $i < 5; ++$i) {
            $values = array(
                'caption' => 'Test_'.chr(rand(97, 122)),
                static::AGGREGATION_FILED_NAME   => rand(3, 20)
            );

            $this->_insertTestValues($store, $values);
        }
    }

    private function _calculateAggregations(&$store): array
    {
        $data = $store->load();

        $expectedTotals = array(
            'sum' => 0,
            'avg' => 0,
            'custom' => 0,
        );

        foreach ($data as $row) {
            $amount = $row[static::AGGREGATION_FILED_NAME];

            $expectedTotals['sum']    += $amount;
            $expectedTotals['avg']    += $amount;
            $expectedTotals['custom'] += $amount * $amount;
        }

        $expectedTotals['avg']    = $expectedTotals['avg'] / count($data);
        $expectedTotals['custom'] = $expectedTotals['custom'] / count($data);

        return $expectedTotals;
    }

    private function _getTotalFromList(&$store)
    {
        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $content = $response->getContent();

        $this->assertNotEmpty($content);
        $this->assertInstanceOf(ListAction::class, $response->storeAction);


        $regExp = '<td class="e-db-table-total-item-textarea">'.
            '\s*([\d\.,]+)\s*<\/td>';

        if (preg_match(
            "#".$regExp."#Umis",
            $content,
            $matches
        )) {
            return $matches[1];
        }

        return 0;
    }

    public function onTestAggregationData(StoreAggregationEvent &$event): bool
    {
        if ($this->_currentType != 'custom') {
            return true;
        }

        $values = &$event->getValues();
        $data = &$event->getData();

        $values[static::AGGREGATION_FILED_NAME] = 0;
        foreach ($data as $row) {
            $amount = $row[static::AGGREGATION_FILED_NAME];
            $values[static::AGGREGATION_FILED_NAME] += $amount * $amount;
        }

        if ($data) {
            $values[static::AGGREGATION_FILED_NAME] /= count($data);
        }

        return true;
    }

    private function _insertTestValues(&$store, $values): void
    {
        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );

        $response = new Response();

        $action->onStart($response);
    } // end _insertTestValues

    private function _clearStore($store): void
    {
        $data = $store->load();
        foreach ($data as $item) {
            $store->removeByPrimaryKey($item['id']);
        }
    }

    public function testUnsupportableProxy(): void
    {
        $options = $this->_getOptions();
        $store = $this->core->createStoreInstance(STORE_DEFAULT, $options);

        $this->assertNotNull($store);
    }

    private function _getOptions(): array
    {
        $options = array();

        $options['model'] = array(
            'table'   => array(
                'name'          => 'site_contents',
                'primaryKey'    => 'id',
                'exceptionMode' => true
            ),

            'aggregations'  => array(
                array(
                    'field'     => 'name',
                    'type'      => 'sum',
                )
            ),

            'fields'  => array(
                'name' => array(
                    'type'      => 'text',
                    'name'      => 'name',
                    'caption'   => 'Name',
                    'required'  => 'true',
                )
            ),

            'actions' => array(
                Store::ACTION_LIST => array(
                    'type'    => Store::ACTION_LIST,
                    'caption' => 'List'
                )
            )
        );

        return $options;
    }
}
