<?php

namespace tests\Bundle\Store\Field;

use Core;
use DataAccessObject;
use Exception;
use FestiEvent;
use FestiTestCase;
use FieldException;
use InsertAction;
use ListAction;
use Response;
use Store;

class AbstractFieldTest extends FestiTestCase
{
    public function testInsertRegexpFieldFailed(): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance('settings');

        $store->setOption('storeExceptionMode', true);

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        $values = array(
            'caption' => '3454',
            'value'   => 'test',
        );

        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );

        $response = new Response();

        try {
            $action->onStart($response);
            $this->fail("Expect FieldException");
        } catch (FieldException $exp) {
            $this->assertEquals(
                'Value is not valid in "Caption"',
                $exp->getMessage()
            );
        }
    }

    public function testInsertRegexpFieldSuccess(): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance('settings');

        $store->setOption('storeExceptionMode', true);

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        $values = array(
            'caption' => 'Test_test',
            'value'   => 'test',
        );

        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );

        $response = new Response();

        try {
            $action->onStart($response);
        } catch (Exception $exp) {
            $this->fail(__('Not Expected Exception %s', $exp->getMessage()));
        }

        $data        = $store->load();
        $storeValues = array_pop($data);

        $this->assertEquals($values['caption'], $storeValues['caption']);

        $store->setOption("storeExceptionMode", false);
    }

    public function testHookFetchFilter(): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance('settings');

        $field = $store->getModel()->getFieldByName('caption');

        $field->set('filter', 'text');

        $store->addEventListener(
            Store::EVENT_ON_FETCH_FIELD_FILTER,
            array($this, 'onFilterFetchValue')
        );

        $content = $this->_getListContent($store);

        $regexp = '#<input\s+type="text"\s+name="settings\[filter\]'.
            '\[settings_caption\]"\s+value="caption test filter value"\s+'.
            'class=".*filter.*" \/>#Umis';

        $this->assertMatchesRegularExpression($regexp, $content);

        $field->set('filter', 'select');

        $store->removeEventListener(
            Store::EVENT_ON_FETCH_FIELD_FILTER,
            array($this, 'onFilterFetchValue')
        );

        $store->addEventListener(
            Store::EVENT_ON_FETCH_FIELD_FILTER,
            array($this, 'onFilterFetchValues')
        );

        $content = $this->_getListContent($store);

        $regexp = '#<option value="one" >'.
            'caption test filter value 1<\/option>#Umis';

        $this->assertMatchesRegularExpression($regexp, $content);

        $regexp = '#<option value="two" >'.
            'caption test filter value 2<\/option>#Umis';

        $this->assertMatchesRegularExpression($regexp, $content);
    }

    private function _getListContent($store): string
    {
        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $content = $response->content;

        $this->assertNotEmpty($content);
        $this->assertInstanceOf(ListAction::class, $response->storeAction);

        return $content;
    }

    public function onFilterFetchValue(FestiEvent &$event): bool
    {
        $field = $event->target['field'];
        $event->target['value'] = $field.' test filter value';

        return true;
    }

    public function onFilterFetchValues(FestiEvent &$event): bool
    {
        $field = $event->target['field'];
        $event->target['values'] = array(
            'one' => $field.' test filter value 1',
            'two' => $field.' test filter value 2',
        );

        return true;
    }
    
    public function testDisplayFormatValue(): void
    {
        $core = Core::getInstance();
        
        $store = $core->createStoreInstance(STORE_DEFAULT);
        $fields = &$store->getModel()->getFields();
    
        $fields['caption']->set('format', '%01.2f');
        $action = $store->createActionInstance(Store::ACTION_INFO);
        
        $response = new Response();
        $action->onStart($response);
        
        $this->assertNotEmpty($response->getContent());
        $this->assertMatchesRegularExpression('#0.00#Umis', $response->getContent());
    }

    public function testCustomQueryWhere(): void
    {
        $core = $this->getCoreWithConnection();

        $expectedCaption = 'Test';
        $expectedID = 15;
        $values = array(
            array(
                'caption' => $expectedCaption,
                'value' => 2,
                'id' => 3
            ),
            array(
                'caption' => $expectedCaption,
                'value' => 2,
                'id' => $expectedID
            )
        );

        if ($core->db->getDatabaseType() === DataAccessObject::TYPE_MSSQL) {
            $core->db->query("SET IDENTITY_INSERT ".STORE_DEFAULT." ON");
        }

        $core->db->massInsert(STORE_DEFAULT, $values);

        if ($core->db->getDatabaseType() === DataAccessObject::TYPE_MSSQL) {
            $core->db->query("SET IDENTITY_INSERT ".STORE_DEFAULT." OFF");
        }

        $store = $core->createStoreInstance('settings');

        $field = $store->getModel()->getFieldByName('caption');
        $field->set('where', 'id = 15');

        $store->getModel()->set('additionalWhere', 'caption = '.$core->db->quote($expectedCaption));

        $action = $store->createActionInstance(Store::ACTION_LIST);
        assert($action instanceof ListAction);

        $response = new Response();
        $action->onStart($response);
        $values = $action->getValues();

        $this->assertTrue(count($values) == 1);

        list($row) = $values;
        $this->assertTrue($row['id'] == $expectedID);
        $this->assertTrue($row['caption'] == $expectedCaption);
    }

    public function testAutocompleteOff(): void
    {
        $fieldName = 'caption';

        $core = $this->getCoreWithConnection();
        $store = $core->createStoreInstance('settings');

        $field = &$store->getModel()->getFieldByName($fieldName);
        $field->set('autocomplete', 'off');

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $response = new Response();
        $action->onStart($response);

        $content = $response->getContent();

        $this->assertMatchesRegularExpression("/<input.*id=\"{$fieldName}\".*(autocomplete=\"off\").*>/Umis", $content);
        $this->assertDoesNotMatchRegularExpression("/jQuery\(\"#{$fieldName}\"\).autocomplete/Umis", $content);
    }

    public function testAutocompleteUrlFillID(): void
    {
        $store = Core::getInstance()->createStoreInstance('dgs_foreignkey_autocomplete');
        $field = &$store->getModel()->getFieldByName('id_param');
        $field->set('autocomplete', '/test/%id%/');

        $primaryKey = rand(1, 10);
        $store->setRequestParam(Store::PRIMARY_KEY_IN_REQUEST, $primaryKey);

        $autocompleteUrl = $field->getAutocompleteUrl();

        $this->assertEquals("/test/{$primaryKey}/", $autocompleteUrl);
    }
    
    public function testIsCustomField()
    {
        $core = $this->getCoreWithConnection();
        
        $store = $core->createStoreInstance('settings_custom_field');
        
        $action = $store->createActionInstance(Store::ACTION_LIST);
        
        $this->assertInstanceOf('ListAction', $action);
        
        $response = new Response();
        
        $action->onStart($response);
        
        $content = $response->content;
        
        $this->assertNotEmpty($content);
    }
    
    public function testIsCustomFieldError()
    {
        $core = $this->getCoreWithConnection();
        
        $store = $core->createStoreInstance('settings_custom_field');
        
        $fields = &$store->getModel()->getFields();
        
        $fields['custom_amount']->set('isCustom', 'yes');
        
        $action = $store->createActionInstance(Store::ACTION_LIST);
        
        $this->assertInstanceOf('ListAction', $action);
        
        $response = new Response();
        
        try {
            $action->onStart($response);
            $this->fail('Expected Exception');
        } catch (Exception $exp) {
            $regexp = '/custom_amount/';
            $this->assertMatchesRegularExpression($regexp, $exp->getMessage());
        }
    }
}
