<?php

namespace tests\Bundle\Store\Field;

use DataAccessObject;
use FestiEvent;
use FestiTestCase;
use InsertAction;
use ListAction;
use Response;
use Store;
use EditAction;

class CheckboxFieldTest extends FestiTestCase
{
    private $_primaryKeyValue;
    
    public function testOnList(): void
    {
        $core = $this->getCoreWithConnection();
        
        $store = $core->createStoreInstance('checkbox_field');
        
        $this->_testInsertBoolValueTrue($store);
        $this->_testInsertBoolValueFalse($store);
 
        $this->_testList($store);
    }

    public function testOnEdit(): void
    {
        $core = $this->getCoreWithConnection();
    
        $store = $core->createStoreInstance('checkbox_field');
    
        $this->_testInsertBoolValueTrue($store);
    
        $this->_testList($store);
    
        $values = array(
            'caption' => 'Test',
            'value'   => 'test2',
            'is_used' => '0',
        );
    
        FestiTestCase::preparePostForStore(
            Store::ACTION_EDIT,
            $values,
            $store,
            $this->_primaryKeyValue
        );
        
        $action = $store->createActionInstance(Store::ACTION_EDIT);
        
        $this->assertInstanceOf(EditAction::class, $action);

        $store->addEventListener(
            Store::EVENT_PREPARE_VALUES,
            function (FestiEvent $event): void {
                $values = $event->getTargetValueByKey('values');

                $this->assertEquals('0', $values['is_used']);
            }
        );
        
        $response = new Response();
        
        $action->onStart($response);
    
        $data = $store->load();
        
        $this->assertNotEmpty($data);
    
        $dbValues = end($data);
    
        $this->assertArrayHasKey('is_used', $dbValues);

        $this->assertEquals($this->_getExpectedValueByDatabase(false), $dbValues['is_used']);
    }

    private function _testList(Store $store): void
    {
        $action = $store->createActionInstance(Store::ACTION_LIST);
    
        $this->assertInstanceOf(ListAction::class, $action);
    
        $response = new Response();
    
        $action->onStart($response);
    
        $this->assertNotEmpty($response->content);
        $this->assertInstanceOf(ListAction::class, $response->storeAction);
    }
    
    private function _testInsertBoolValueTrue(Store $store): void
    {
        $values = array(
            'caption'  => 'Caption Test',
            'value'    => 'setting',
            'id_param' => '1',
            'is_used'  => '1',
        );
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
        
        $response = new Response();
        
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $this->assertInstanceOf(InsertAction::class, $action);

        $listener = function (FestiEvent $event): void {
            $values = $event->getTargetValueByKey('values');

            $this->assertEquals('1', $values['is_used']);
        };

        $store->addEventListener(Store::EVENT_PREPARE_VALUES, $listener);
        
        $action->onStart($response);

        $store->removeEventListener(Store::EVENT_PREPARE_VALUES, $listener);
    
        $data = $store->load();
        
        $this->assertNotEmpty($data);
        
        $dbValues = array_shift($data);
        
        $this->assertArrayHasKey('is_used', $dbValues);

        $this->assertEquals($this->_getExpectedValueByDatabase(true), $dbValues['is_used']);
        
        $this->_primaryKeyValue = $dbValues['id'];
    }
    
    private function _testInsertBoolValueFalse(Store $store): void
    {
        $values = array(
            'caption'  => 'Caption Test 2',
            'value'    => 'setting 2',
            'id_param' => '2',
            'is_used'  => '0',
        );
    
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
    
        $response = new Response();
    
        $action = $store->createActionInstance(Store::ACTION_INSERT);
    
        $this->assertInstanceOf(InsertAction::class, $action);

        $listener = function (FestiEvent $event): void {
            $values = $event->getTargetValueByKey('values');

            $this->assertEquals('0', $values['is_used']);
        };

        $store->addEventListener(Store::EVENT_PREPARE_VALUES, $listener);

        $action->onStart($response);

        $store->removeEventListener(Store::EVENT_PREPARE_VALUES, $listener);
    
        $data = $store->load();

        $this->assertNotEmpty($data);
        
        $dbValues = end($data);
    
        $this->assertArrayHasKey('is_used', $dbValues);

        $this->assertEquals($this->_getExpectedValueByDatabase(false), $dbValues['is_used']);
    }

    /**
     * @return mixed
     */
    private function _getExpectedValueByDatabase(bool $value)
    {
        switch ($this->core->db->getDatabaseType()) {
            case DataAccessObject::TYPE_PGSQL:
                return $value;
        }

        return $value ? '1' : '0';
    }
}
