<?php

namespace tests\Bundle\Store\Field;

use DateTime;
use DateTimeZone;
use EditAction;
use Exception;
use FestiTestCase;
use FieldException;
use InsertAction;
use ListAction;
use Response;
use Store;

class DateTimeFieldTest extends FestiTestCase
{
    public function testInsertRequiredFieldFailed(): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance('site_contents_datetime');

        $store->setOption('storeExceptionMode', true);

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        $values = array(
            'name' => 'Test',
        );

        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );

        $response = new Response();

        try {
            $action->onStart($response);
            $this->fail("Expect FieldException");
        } catch (FieldException $exp) {
            $this->assertEquals('Cdate is required field', $exp->getMessage());
        }
    }

    public function testInsertRequiredFieldSuccess(): void
    {
        $core = $this->getCoreWithConnection();

        $store = $core->createStoreInstance('site_contents_datetime');

        $store->setOption('storeExceptionMode', true);

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        $values = array(
            'name'  => 'Test',
            'cdate' => date('Y-m-d H:i:s'),
        );

        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );

        $response = new Response();

        try {
            $action->onStart($response);
        } catch (Exception $exp) {
            $this->fail(__("Not Expected Exception %s", $exp->getMessage()));
        }

        $data        = $store->load();
        $storeValues = array_pop($data);

        $this->assertEquals(strtotime($values['cdate']), strtotime($storeValues['cdate']));

        $store->setOption('storeExceptionMode', false);
    }
    
    public function testOnDifferentTimezone(): void
    {
        $core = $this->getCoreWithConnection();
        
        $store = $core->createStoreInstance('site_contents_datetime');
    
        $values = array(
            'name'  => 'Test',
            'cdate' => date('Y-m-d H:i:s'),
        );
    
        $sourceTimezone = 'Europe/London';
        $userTimezone = 'Europe/Kiev';
    
        $id = $this->_testOnInsertDifferentTimezone($store, $values, $sourceTimezone, $userTimezone);
        
        $action = $store->createActionInstance(Store::ACTION_LIST);
    
        $this->assertInstanceOf(ListAction::class, $action);
    
        $response = new Response();
    
        $action->onStart($response);
    
        $regexp = '/fieldName="cdate" primaryValue="'.$id.'"\>(?P<value>\n.*)\<\/td\>/';
    
        preg_match($regexp, $response->getContent(), $matches);
    
        $cdate = trim($matches['value']);
        $this->assertNotEmpty($cdate);
        
        $data = $action->getValues();
        $storeValues = array_pop($data);
        $this->assertNotEmpty($storeValues);
        
        $this->_testOnDifferentTimezone($values, $storeValues, $sourceTimezone, $userTimezone, $cdate);
    }
    
    public function testOnDifferentTimezoneEditForm(): void
    {
        $core = $this->getCoreWithConnection();
    
        $store = $core->createStoreInstance('site_contents_datetime');
    
        $values = array(
            'name'  => 'Test',
            'cdate' => date('Y-m-d H:i:s'),
        );
    
        $sourceTimezone = 'Europe/London';
        $userTimezone = 'Europe/Kiev';
    
        $id = $this->_testOnInsertDifferentTimezone($store, $values, $sourceTimezone, $userTimezone);
        
        $action = $store->createActionInstance(Store::ACTION_EDIT);
        $this->assertInstanceOf(EditAction::class, $action);
        
        FestiTestCase::preparePrimaryKey($store, $id);
    
        $response = new Response();
        
        $action->onStart($response);

        $regexp = '/id="cdate"\s+name="cdate"\s+value="(?P<value>.*)"/';
    
        preg_match($regexp, $response->getContent(), $matches);
    
        $cdate = trim($matches['value']);
        $this->assertNotEmpty($cdate);
    
        $storeValues = $action->getValues();
        
        $this->assertNotEmpty($storeValues);
        
        $this->_testOnDifferentTimezone($values, $storeValues, $sourceTimezone, $userTimezone, $cdate);
    }
    
    private function _testOnDifferentTimezone(
        array $values, array $storeValues, string $sourceTimezone, string $userTimezone, string $cdate
    ): void
    {
        $this->assertEquals(strtotime($values['cdate']), strtotime($cdate));
        $this->assertNotEquals(strtotime($values['cdate']), strtotime($storeValues['cdate']));
    
        $dateTime = new DateTime($storeValues['cdate'], new DateTimeZone($sourceTimezone));
        $dateTime->setTimezone(new DateTimeZone($userTimezone));
    
        $this->assertEquals(strtotime($values['cdate']), strtotime($dateTime->format('Y-m-d H:i:s')));
    }
    
    private function _testOnInsertDifferentTimezone(
        Store &$store,
        array $values,
        string $sourceTimezone,
        string $userTimezone
    ): int
    {
        $store->getModel()->set('sourceTimezone', $sourceTimezone);
        $store->getModel()->set('userTimezone', $userTimezone);
    
        $action = $store->createActionInstance(Store::ACTION_INSERT);
    
        $this->assertInstanceOf(InsertAction::class, $action);
    
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );
    
        $response = new Response();
    
        $action->onStart($response);
    
        return $action->getPrimaryKeyValue();
    }
}
