<?php

namespace tests\Bundle\Store\Field;

use Core;
use EditAction;
use FestiTestCase;
use FieldException;
use InfoAction;
use InsertAction;
use ListAction;
use RemoveAction;
use Response;
use Store;

class FileFieldTest extends FestiTestCase
{
    private const VALIDATE_VALUE_ATTRIBUTE = 'Test::onValidateFileField';

    public function testPearIsUploadedFile(): void
    {
        $path  = dirname(__FILE__).DIRECTORY_SEPARATOR.'file.phpt';

        $cmd = 'pear run-tests '.$path;

        $output = array();
        $this->core->exec($cmd, $output);

        $result = join(' ', $output);

        $this->assertMatchesRegularExpression(
            "#1 PASSED TESTS#Umis",
            $result,
            'Problem with file field.'
        );
    } // end testPearIsUploadedFile

    public function testUploadFileInput(): void
    {
        $options = $this->_getOptions();

        $_FILES['name'] = array(
            'name'     => 'phpunit.png',
            'tmp_name' => FS_RESOURCES_ROOT.'phpunit.png',
            'type'     => 'image/png'
        );

        $values = array(
            'name' => $_FILES['name']['name'],
            'path'   => $_FILES['name']['tmp_name']
        );

        list($data, $response) = $this->_insertRow($options, $values);

        $this->assertEquals($data['id'], '1');
        $this->assertEquals($data['name'], '1_phpunit.png;0;image/png');
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);
    } // end testUploadFileInput

    public function testInfoFileInput(): void
    {
        $options = $this->_getOptions();

        $idRow = $this->_testUploadNewFile();

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_INFO);

        $this->assertInstanceOf(InfoAction::class, $action);

        FestiTestCase::preparePostForStore(
            Store::ACTION_INFO,
            array(),
            $store,
            $idRow
        );

        $response = new Response();

        $action->onStart($response);

        $file = $idRow.'_phpunit.png';

        $this->assertMatchesRegularExpression(
            "#>{$file}</a>#Umis",
            $response->getContent(),
            "Not found Info form."
        );
    }

    public function testEditFileInput(): void
    {
        $idRow = $this->_testUploadNewFile();

        $this->_testUpdateWithNewFile($idRow);

        $this->_testUpdateWithoutNewFile($idRow);

        $this->_testRemoveFile($idRow);

    } // end testUploadFileInput

    private function _insertRow($options, $values): array
    {
        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        $this->_onPrepareFileField($store);

        FestiTestCase::preparePostForStore($action, $values, $store);

        $response = new Response();

        $action->onStart($response);

        $data = $store->load();

        return array(array_pop($data), $response, $store);
    }

    private function _testUploadNewFile()
    {
        $options = $this->_getOptions();

        $_FILES['name'] = array(
            'name' => 'phpunit.png',
            'tmp_name' => FS_RESOURCES_ROOT . 'phpunit.png',
            'type' => 'image/png'
        );

        $values = array(
            'name' => $_FILES['name']['name'],
            'path' => $_FILES['name']['tmp_name']
        );

        list($data, $response) = $this->_insertRow($options, $values);

        $idRow = $data['id'];

        $this->assertEquals($data['name'], $idRow.'_phpunit.png;0;image/png');
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);

        return $idRow;
    }

    private function _testUpdateWithNewFile($idRow)
    {
        $options = $this->_getOptions();

        $fileName = '__ID_____NAME_____FIELD__.__EXT__';
        $options['model']['fields']['value']['fileName'] = $fileName;

        $options['model']['fields']['id_parent'] = array(
            'type'      => 'text',
            'name'      => 'id_parent',
            'caption'   => 'Parent ID',
        );

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $this->_onPrepareFileField($store);

        $action = $store->createActionInstance(Store::ACTION_EDIT);

        $this->assertInstanceOf(EditAction::class, $action);

        $_FILES['name'] = array(
            'name' => 'phpunit.png',
            'tmp_name' => FS_RESOURCES_ROOT . 'phpunit.png',
            'type' => 'image/png'
        );

        $values = array(
            'name' => $_FILES['name']['name'],
            'path' => $_FILES['name']['tmp_name'],
            'id_parent' => time()
        );

        FestiTestCase::preparePostForStore($action, $values, $store, $idRow);

        $response = new Response();

        $action->onStart($response);

        $data = $store->load();

        $data = array_pop($data);

        $this->assertEquals($data['name'], $idRow.'_phpunit_name.png;0;image/png');
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);

        return $store;
    }

    private function _testUpdateWithoutNewFile( $idRow)
    {
        $options = $this->_getOptions();

        $fileName = '__ID_____NAME_____FIELD__.__EXT__';
        $options['model']['fields']['value']['fileName'] = $fileName;

        $options['model']['fields']['id_parent'] = array(
            'type'      => 'text',
            'name'      => 'id_parent',
            'caption'   => 'Parent ID',
        );

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $this->_onPrepareFileField($store);

        $action = $store->createActionInstance(Store::ACTION_EDIT);

        $this->assertInstanceOf(EditAction::class, $action);

        $_FILES = array();
        $values = array(
            '_name' => $idRow.'_phpunit_name.png;0;image/png',
            'id_parent' => time()
        );

        FestiTestCase::preparePostForStore($action, $values, $store, $idRow);

        $response = new Response();

        $action->onStart($response);

        $data = $store->load();

        $data = array_pop($data);

        $this->assertEquals($data['name'], $idRow.'_phpunit_name.png;0;image/png');
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);

        return $store;
    }

    private function _testRemoveFile($idRow)
    {
        $options = $this->_getOptions();

        $fileName = '__ID_____NAME_____FIELD__.__EXT__';
        $options['model']['fields']['value']['fileName'] = $fileName;
        $options['model']['fields']['value']['required'] = false;

        $options['model']['fields']['id_parent'] = array(
            'type'      => 'text',
            'name'      => 'id_parent',
            'caption'   => 'Parent ID',
        );

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_EDIT);

        $this->assertInstanceOf(EditAction::class, $action);

        $_FILES = array();
        $values = array(
            '_name' => '',
            'id_parent' => time()
        );

        FestiTestCase::preparePostForStore($action, $values, $store, $idRow);

        $response = new Response();

        $action->onStart($response);

        $data = $store->load();

        $data = array_pop($data);

        $this->assertTrue(empty($data['name']));
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);
    }

    private function _onPrepareFileField(&$store): void
    {
        $fileField = &$store->getModel()->getFieldByName('name');

        $mockMethods = array('isUploaded', 'copy', 'getUploadFilePath');

        $fileFieldMock = $this->getMockBuilder('ImageField')
                              ->setConstructorArgs(array(&$store))
                              ->setMethods($mockMethods)
                              ->getMock();

        $attributes = $fileField->getAttributes();

        foreach ($attributes as $key => $value) {
            $fileFieldMock->set($key, $value);
        }

        $fileFieldMock->expects($this->any())
                      ->method('isUploaded')
                      ->will($this->returnValue(true));

        $fileFieldMock->expects($this->any())
            ->method('getUploadFilePath')
            ->will($this->returnValue($store->getDefaultUploadFilePath()));

        $fileFieldMock->expects($this->any())
                      ->method('copy')
                      ->will($this->returnValue(true));
        
        $fileFieldMock->setName($fileField->getName());

        $fileField = $fileFieldMock;
    } // _onPrepareFileField

    public function testOnFieldException(): void
    {
        $options = $this->_getOptions();

        $value = static::VALIDATE_VALUE_ATTRIBUTE;

        $options['model']['fields']['value']['onValidateValue'] = $value;

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $values = array();
        FestiTestCase::preparePostForStore($action, $values, $store);

        $_FILES['name'] = array(
            'name'     => 'phpunit.csv',
            'tmp_name' => FS_RESOURCES_ROOT.'phpunit.png',
            'type'     => 'image/png'
        );

        $response = new Response();

        try {
            $action->onStart($response);
            $this->fail("Expect FieldException");
        } catch (FieldException $exp) {
            $this->assertTrue($exp->getMessage() == "Invalid File Extension");
        }
    } // end testOnFieldException

    public function testOnRequiredFieldException(): void
    {
        $options = $this->_getOptions();

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $values = array();
        FestiTestCase::preparePostForStore($action, $values, $store);

        $_FILES = array();

        $response = new Response();

        try {
            $action->onStart($response);
            $this->fail("Expect FieldException");
        } catch (FieldException $exp) {
            $this->assertEquals('Value is required field', $exp->getMessage());
        }
    } // end testOnRequiredFieldException

    private function _getOptions(): array
    {
        $options = array();

        $options['model'] = array(
            'table'   => array(
                'name'          => 'site_contents',
                'primaryKey'    => 'id',
                'exceptionMode' => true
            ),

            'fields'  => array(
                'value' => array(
                    'type'      => 'image',
                    'name'      => 'name',
                    'caption'   => 'Value',
                    'fileName'  => '__ID_____NAME__.__EXT__',
                    'required'  => 'true',
                )
            ),

            'actions' => array(
                Store::ACTION_INSERT => array(
                    'type'    => Store::ACTION_INSERT,
                    'caption' => 'Add'
                ),
                Store::ACTION_EDIT => array(
                    'type'    => Store::ACTION_EDIT,
                    'caption' => 'Edit'
                ),
                Store::ACTION_INFO => array(
                    'type'    => Store::ACTION_INFO,
                    'caption' => 'Info'
                ),
                Store::ACTION_REMOVE => array(
                    'type'    => Store::ACTION_REMOVE,
                    'caption' => 'Remove'
                ),
                Store::ACTION_LIST => array(
                    'type'    => Store::ACTION_LIST,
                    'caption' => 'List'
                )
            )
        );
        
        return $options;
    }

    public function testCustomFileLink(): void
    {
        $idRow = $this->_testUploadNewFile();

        $options = $this->_getOptions();

        $options['model']['fields']['value']['link'] = '/test_download/%id%/';

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_EDIT);

        $this->assertInstanceOf(EditAction::class, $action);

        $_FILES = array();

        FestiTestCase::preparePrimaryKey($store, $idRow);

        $response = new Response();

        $action->onStart($response);

        $content = $response->content;

        $this->assertNotEmpty($content);

        $this->assertMatchesRegularExpression(
            "#<a href=\"/test_download/".$idRow."/\">".$idRow."_phpunit.png</a>#Umis",
            $content,
            "Incorrect file url."
        );
    } // end testCustomFileLink

    public function testDisplayFileLink(): void
    {
        $idRow = $this->_testUploadNewFile();

        $options = $this->_getOptions();

        $content = $this->_getStoreListContent($options);

        $this->assertMatchesRegularExpression(
            "#<a href=\"/storages/site_contents/".$idRow."_phpunit.png\">".
            $idRow."_phpunit.png</a>#Umis",
            $content,
            "Incorrect file url."
        );
        $options['model']['fields']['value']['thumb'] = '100x100';

        $content = $this->_getStoreListContent($options);

        $this->assertMatchesRegularExpression(
            "#<img class=\"img-rounded\" src=\"/storages/site_contents/thumbs/".
            $idRow."_phpunit.png\" />#Umis",
            $content,
            "Incorrect image url."
        );
    } // end testDisplayFileLink

    private function _getStoreListContent($options)
    {
        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $content = $response->content;

        $this->assertNotEmpty($content);

        return $content;
    }

    public function testDisplayFileInRemoveForm(): void
    {
        $idRow = $this->_testUploadNewFile();

        $options = $this->_getOptions();
        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $store->addAllowedTableRow($idRow);

        $action = $store->createActionInstance(Store::ACTION_REMOVE);

        $this->assertInstanceOf(RemoveAction::class, $action);

        $_POST = array();
        $_REQUEST = $_POST;
        $_FILES = array();

        FestiTestCase::preparePrimaryKey($store, $idRow);

        $response = new Response();

        $action->onStart($response);

        $content = $response->content;

        $this->assertNotEmpty($content);

        $caption =  $options['model']['fields']['value']['caption'];

        $this->assertMatchesRegularExpression(
            "#<span [^>]+>".$caption.":<\/span> ".
            $idRow."_phpunit.png#Umis",
            $content,
            "Incorrect file name."
        );
    } // end testUploadFileInput
}
