<?php

namespace tests\Bundle\Store\Field;

use Core;
use DataAccessObject;
use FestiTestCase;
use ForeignKeyField;
use InsertAction;
use ListAction;
use Response;
use Store;
use core\store\field\IStoreForeignField;
use SystemException;
use TestForeignKeyPlugin;

class ForeignKeyFieldTest extends FestiTestCase
{
    private const REGEXP_SELECT_SEARCH_RESULT   = "#<b>Section:</b> Test 11#Umis";
    private const REGEXP_TEXT_SEARCH_RESULT     = "#<b>Section:</b> Test 11#Umis";
    private const REGEXP_EXACT_SEARCH_RESULT    = "#<b>Section:</b> 1#Umis";
    private const REGEXP_MULTIPLY_SEARCH_RESULT = "#<b>Section:</b> Test 10, Test 11#Umis";

    public function testTopFilterSearchResult(): void
    {
        $this->_prepareData();
        
        $this->_testSelectFilter();
        $this->_testMultipleFilter();
        $this->_testTextFilter();
        $this->_testExactFilter();
    }

    public function testStoreProxyForeignKeyValuesListener(): void
    {
        $dgsModel = $this->_getModel('text');
        $dgsModel['model']['table']['plugin'] = 'TestForeignKey';

        $store = Core::getInstance()->createStoreInstance('filters', $dgsModel);

        $action = $store->createActionInstance(Store::ACTION_INSERT);
        assert($action instanceof InsertAction);

        $response = new Response();
        $action->onStart($response);

        $field = $action->getStore()->getModel()->getFieldByName('id_parent');
        assert($field instanceof ForeignKeyField);
        $this->assertInstanceOf(IStoreForeignField::class, $field);

        $this->assertEquals(count(TestForeignKeyPlugin::$values), count($field->getValuesList()));
    } // end testStoreProxyForeignKeyValuesListener

    private function _testSelectFilter(): void
    {
        $filters = array(
            'id_parent' => '11',
        );
    
        $regexp  = static::REGEXP_SELECT_SEARCH_RESULT;
    
        $this->_testFilter('select', $filters, $regexp);
    } // _testSelectFilter
    
    private function _testTextFilter(): void
    {
        $filters = array(
            'id_parent' => 'Test 11',
        );
    
        $regexp  = static::REGEXP_TEXT_SEARCH_RESULT;
    
        $this->_testFilter('text', $filters, $regexp);
    }
    
    private function _testExactFilter(): void
    {
        $filters = array(
            'id_parent' => '1',
        );
    
        $regexp  = static::REGEXP_EXACT_SEARCH_RESULT;
    
        $this->_testFilter('exact', $filters, $regexp);
    }
    
    private function _testMultipleFilter(): void
    {
        $filters = array(
            'id_parent' => array(
                10,
                11
            ),
        );
    
        $regexp  = static::REGEXP_MULTIPLY_SEARCH_RESULT;
        
        $this->_testFilter('multiple', $filters, $regexp);
    } // _testMultipleFilter
    
    private function _testFilter($filterType, $filters, $regexp): void
    {
        $dgsModel = $this->_getModel($filterType);
    
        $store = Core::getInstance()->createStoreInstance('filters', $dgsModel);
        
        FestiTestCase::prepareFiltersForStore($filters, $store);
    
        $action   = $store->createActionInstance(Store::ACTION_LIST);
    
        $this->assertInstanceOf(ListAction::class, $action);
    
        $response = new Response();
    
        $action->onStart($response);
        
        $this->assertMatchesRegularExpression(
            $regexp,
            $response->content,
            "Not found search result."
        );
    } // _testFilter
    
    private function _getModel(string $filterType): array
    {
        $options = array();
    
        $foreignValueField = $this->_getValueFieldByFilterType($filterType);
        
        $options['model'] = array(
            'table'   => array(
                'name'                  => 'settings',
                'primaryKey'            => 'id',
                'filter'                => 'top'
            ),
            'fields'  => array(
                array(
                    'type'              => 'text',
                    'name'              => 'caption',
                    'caption'           => 'Caption',
                    'filter'            => 'text',
                ),
                array(
                    'type'              => 'foreignKey',
                    'name'              => 'id_parent',
                    'caption'           => 'Section',
                    'alias'             => 'parent_settings',
                    'foreignTable'      => 'settings',
                    'foreignKeyField'   => 'id',
                    'foreignValueField' => $foreignValueField,
                    'filter'            => $filterType,
                ),
            ),
            'actions' => array(
                Store::ACTION_LIST      => array(
                    'type'    => Store::ACTION_LIST,
                    'caption' => 'List'
                ),
                Store::ACTION_INSERT      => array(
                    'type'    => Store::ACTION_INSERT,
                    'caption' => 'Insert'
                ),
            ),
        );
        
        return $options;
    } // end _getModel
    
    private function _getValueFieldByFilterType(string $filterType): string
    {
        $fieldValues = array(
            'select'   => 'caption',
            'multiple' => 'caption',
            'text'     => "CONCAT(settings.caption, ' ')",
            'exact'    => "caption"
        );
        
        if (!array_key_exists($filterType, $fieldValues)) {
            throw new SystemException('Unsupported filter type');
        }
        
        return $fieldValues[$filterType];
    }
    
    private function _prepareData(): void
    {
        $values = array(
            array(
                'id'        => 10, 
                'caption'   => 'Test 10', 
                'id_parent' => null
            ),
            array(
                'id'        => 11, 
                'caption'   => 'Test 11', 
                'id_parent' => 10
            ),
            array(
                'id'        => 12,
                'caption'   => 'Test 12', 
                'id_parent' => 11
            ),
            array(
                'id'        => 13,
                'caption'   => 'Test 13',
                'id_parent' => 11
            ),
        );
    
        $connection = $this->getDbConnection();

        if ($connection->getDatabaseType() === DataAccessObject::TYPE_MSSQL) {
            $connection->query('SET IDENTITY_INSERT settings ON');
        }
        
        $connection->massInsert('settings', $values);

        if ($connection->getDatabaseType() === DataAccessObject::TYPE_MSSQL) {
            $connection->query('SET IDENTITY_INSERT settings OFF');
        }
    } // end _prepareData
}
