<?php

namespace tests\Bundle\Store\Field;

use AbstractField;
use DatabaseException;
use EditAction;
use FestiTestCase;
use FieldException;
use InsertAction;
use ListAction;
use Many2manyField;
use Response;
use Store;

class Many2manyFieldTest extends FestiTestCase
{
    const DEFAULT_VALUE = 1;
    
    private $_selectIDs;

    public function testOnSave(): void
    {
        $this->_testMany2ManyFilterSelectInsertForm();

        $this->_testSetSettings();

        $this->_testMany2ManyDefaultDisplay();
 
        $this->_testMany2ManyMultipleSelectDisplay();
    } // end testOnSave

    private function _testMany2ManyDefaultDisplay(): void
    {
        $store = $this->core->createStoreInstance('many2many');
        
        $postParams = array_fill_keys($this->_selectIDs, static::DEFAULT_VALUE);
        
        $id = $this->_testInsertRequest($store, $postParams);

        $this->_testMany2ManyWithInfoAction($store, $id);
        
        $this->_testRemoveRequest($store, $id);
    } // end _testMany2ManyDefaultDisplay
    
    private function _testMany2ManyMultipleSelectDisplay(): void
    { 
        $store = $this->core->createStoreInstance('many2many_multiple');
        
        $id = $this->_testInsertRequest($store, $this->_selectIDs);
        
        $this->_testRemoveRequest($store, $id);
    } // end _testMany2ManyMultipleSelectDisplay

    private function _testMany2ManyFilterSelectInsertForm(): void
    {
        $store = $this->core->createStoreInstance('many2many_multiple');

        $field = &$store->getModel()->getFieldByName(1);
        $field->set('filter', 'select');

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        $_POST = array();
        $_REQUEST = $_POST;

        $response = new Response();
        $action->onStart($response);

        $this->assertMatchesRegularExpression(
            "#f-db-form#Umis",
            $response->content,
            "Not found insert form."
        );
    } // end _testMany2ManyFilterSelectInsertForm

    private function _getIDs($settings): array
    {
        $ids = array();
        foreach ($settings as $setting) {
            $ids[] = $setting['id'];
        }
        return $ids;
    } // end _getIDs
    
    private function _testSetSettings(): void
    {
        $store = $this->core->createStoreInstance('settings');
        $settings = array();
        
        $values = array(
            0 => array(
                'caption' => 'Test_Setting',
                'value'   => 'Values_Setting'
            ),
            1 => array(
                'caption' => 'Test_Setting_Two',
                'value'   => 'Values_Setting2'
            )
        );
        
        $settings[] = $this->_doInsertSettings($store, $values[0]);
        
        $settings[] = $this->_doInsertSettings($store, $values[1]);
        
        $data = $store->load();
        
        $this->assertTrue($data[0]['value'] == $values[0]['value']);
        $this->assertTrue($data[1]['value'] == $values[1]['value']);
        
        $this->_selectIDs = $this->_getIDs($data);
    } // end _testSetSettings
    
    private function _doInsertSettings(&$store, $values): void
    {
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT, 
            $values, 
            $store
        );
        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $response = new Response();
        
        $action->onStart($response);
    } // _doInsertSettings
    
    private function _testRemoveRequest(&$store, $primaryKey): void
    {
        $store->addAllowedTableRow($primaryKey);
        $action = $store->createActionInstance(Store::ACTION_REMOVE);
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_REMOVE, 
            array(), 
            $store, 
            $primaryKey
        );
        
        $response = new Response();
        
        $action->onStart($response);
        
        $data = $store->load();
        
        $this->assertEquals(0, count($data));
    } // end _testRemoveRequest
    
    private function _testInsertRequest(&$store, $postParams = array())
    {
        $values = array(
            'name' => 'Test'
        );
        
        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT, 
            $values, 
            $store
        );
        $postKey = 'm2m_site_contents2settings';
        
        $_POST[$postKey] = $postParams;

        $action = $store->createActionInstance(Store::ACTION_INSERT);
        
        $response = new Response();
        $action->onStart($response);

        $data = $store->load();

        $this->assertTrue(count($data) == 1);
        if (!empty($postParams)) {
            $this->assertMatchesRegularExpression(
                "#Test_Setting::Test_Setting_Two#Umis",
                $data[0]['__1'],
                "Not found Many2many fields"
            );
        }

        return $data[0]['id'];
    } // end _testInsertRequest

    private function _testMany2ManyWithInfoAction($store, $id): void
    {
        $action = $store->createActionInstance(Store::ACTION_INFO);
        FestiTestCase::preparePostForStore(
            Store::ACTION_INFO,
            array(),
            $store,
            $id
        );

        $response = new Response();
        $action->onStart($response);
        
        $content = $response->content;
        
        $this->assertNotEmpty($content);
        $this->assertMatchesRegularExpression(
            "#<div class=\"e-dgs-field-input e-dgs-field-input-many2many ".
            "[^>]+><ul>(<li>\w+<\/li>){2}<\/ul><\/div>#Umis",
            $content,
            "Not found Many2many fields values"
        );
    } // end _testMany2ManyWithInfoAction


    public function testOnDisplayEmptyCell(): void
    {
        $store = $this->core->createStoreInstance('many2many_multiple');

        $id = $this->_testInsertRequest($store);

        $action = $store->createActionInstance(Store::ACTION_LIST);

        $this->assertInstanceOf(ListAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $content = $response->content;

        $this->assertNotEmpty($content);
        $this->assertMatchesRegularExpression(
            "#<td[^>]+primaryValue=\"".$id."\"[^>]*>\s*".
            "<div class=\"[^\"]+e-dgs-field-input-__[0-9]+-value\"><\/div>#Umis",
            $content,
            "Not found Many2many fields values"
        );
    } // end testOnDisplayEmptyCell

    /**
     * Test bug:
     * DatabaseException: SQLSTATE[HY000]: General error: 1366 Incorrect integer value: '' for column
     * 'id_setting' at row 1
     * @throws \SystemException
     */
    public function testInsertEmptyValuesWithAutocompleteMode(): void
    {
        $store = $this->core->createStoreInstance('many2many_autocomplete');

        $values = array(
            'name' => __METHOD__
        );

        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );

        $field = $store->getModel()->getFieldByOption("type", "many2many");

        $postKey = $field->getKeyInRequest();
        $_POST[$postKey] = "";

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $response = new Response();
        $action->onStart($response);

        $data = $store->loadRow();

        $this->assertTrue($data['name'] == __METHOD__);
    } // end testInsertEmptyValuesWithAutocompleteMode

    public function testInsertEmptyValuesWithMultipleMode(): void
    {
        $this->_testSetSettings();

        $store = $this->core->createStoreInstance('many2many_multiple');

        $store->setOption("storeExceptionMode", true);

        $linkTable = 'site_contents2settings';
        $field = &$store->getModel()->getFieldByOption('linkTable', $linkTable);
        $field->set('required', true);

        $this->_testInsertEmptyValuesForm($store, $linkTable);
        $this->_testInsertEmptyValuesRequest($store);
        $store->setOption("storeExceptionMode", false);
    } // end testInsertEmptyValuesWithMultipleMode

    private function _testInsertEmptyValuesForm($store, $linkTable): void
    {
        $_POST = array();
        $_REQUEST = $_POST;

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        $response = new Response();

        $action->onStart($response);

        $content = $response->content;

        $this->assertNotEmpty($content);
        $this->assertMatchesRegularExpression(
            "#<select [^>]* name=\"m2m_".$linkTable.
            "\[\]\" multiple=\"multiple\" [^>]* required>#Umis",
            $content,
            "Not found Many2many fields values"
        );
    } // end _testInsertEmptyValuesForm

    private function _testInsertEmptyValuesRequest($store): void
    {
        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            array(),
            $store
        );

        $response = new Response();

        try {
            $action->onStart($response);
            $this->fail('FieldException expected');
        } catch (FieldException $exception) {
            $this->assertEquals(
                'Name is required field',
                $exception->getMessage()
            );
        }
    } // end _testInsertEmptyValuesRequest

    public function testWithOptions(): void
    {
        $expectedCaption = 'Test_Setting';
        $values = array(
            'caption' => $expectedCaption,
            'value'   => 'Values_Setting'
        );

        $idSetting = $this->core->db->insert(STORE_DEFAULT, $values);

        $values = array(
            'caption' => 'Test_Setting 2',
            'value'   => 'Values_Setting'
        );

        $idSecondSetting = $this->core->db->insert(STORE_DEFAULT, $values);

        //
        $store = $this->core->createStoreInstance('many2many_options');

        $field = &$store->getModel()->getFieldByName(1);

        assert($field instanceof Many2manyField);

        $values = array(
            'name' => __METHOD__
        );

        FestiTestCase::preparePostForStore(
            Store::ACTION_INSERT,
            $values,
            $store
        );

        $postKey = $field->getKeyInRequest();
        $_POST[$postKey] = array($idSetting => array(2, 4));

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $response = new Response();
        $action->onStart($response);

        $data = $store->loadRow();

        $idContent = $data['id'];

        //
        $this->assertTrue($data[$field->getElementName()] == $expectedCaption);

        $sql = "SELECT * FROM site_contents2settings";
        $data = $this->core->db->getRow($sql);

        $this->assertTrue($data['value'] == array_sum($_POST[$postKey][$idSetting]));

        //
        $this->_testComplexForeignFieldValue($store, $field, $idContent, $idSetting, $idSecondSetting);
        $this->_testComplexForeignFieldValueHasID($store, $field, $idContent, $idSetting, $idSecondSetting);
        $this->_testDisplayRemovedForeignValues($store, $field, $idContent, $idSetting);
    } // end testWithOptions

    private function _testComplexForeignFieldValue(
        Store $store, Many2manyField $field, $idContent, $idSetting, $idSecondSetting
    ): void
    {
        $_POST = array();

        $field->set(Many2manyField::OPTION_FOREIGN_VALUES_FIELD, "CONCAT(caption, ' - ', caption)");

        $request = array(
            Store::PRIMARY_KEY_IN_REQUEST => $idContent
        );

        $store->setRequest($request);

        $action = $store->createActionInstance(Store::ACTION_EDIT);
        assert($action instanceof EditAction);

        //
        $field->set(Many2manyField::OPTION_VALUES_WHERE, 'id IN ('.$idSetting.', '.$idSecondSetting.')');
        $field->set(Many2manyField::OPTION_VALUES_ORDER_BY, 'id DESC');

        $values = $action->load($idContent)->getValues();

        $this->assertTrue($values[$field->getElementName()] == 'Test_Setting - Test_Setting');
        $this->assertTrue($values['id'] == $idContent);

        $response = new Response();
        $action->onStart($response);
        $this->assertNotEmpty($response->getContent());
        $fieldValues = $field->getValues();

        $this->assertTrue(count($fieldValues) == 2);

        //
        $field->set(AbstractField::OPTION_AUTOCOMPLETE, true);

        $response = new Response();
        $action->onStart($response);
        $fieldValues = $field->getValues();

        $this->assertTrue(count($fieldValues) == 1);

        $row = reset($fieldValues);
        $this->assertTrue($row['value'] == 'Test_Setting - Test_Setting');
    }
    
    private function _testComplexForeignFieldValueHasID(
        Store $store, Many2manyField $field, $idContent, $idSetting, $idSecondSetting
    )
    {
        FestiTestCase::cleanPostByStoreName($store->getName());
        
        $field->set(Many2manyField::OPTION_FOREIGN_VALUES_FIELD, "CONCAT(id, ' - ', caption)");
        
        $request = array(
            Store::PRIMARY_KEY_IN_REQUEST => $idContent
        );
    
        $store->setRequest($request);
    
        $action = $store->createActionInstance(Store::ACTION_EDIT);
        assert($action instanceof EditAction);
        
        $field->set(Many2manyField::OPTION_VALUES_WHERE, 'id IN ('.$idSetting.', '.$idSecondSetting.')');
        $field->set(Many2manyField::OPTION_VALUES_ORDER_BY, 'id DESC');
    
        try {
            $action->load($idContent)->getValues();
            $this->fail('Expected DatabaseException');
        } catch (DatabaseException $exp) {
            $this->assertMatchesRegularExpression('/(?i)ambiguous/', $exp->getMessage());
        }
        
        $field->set(Many2manyField::OPTION_FOREIGN_VALUES_FIELD, "CONCAT(settings.id, ' - ', settings.caption)");
      
        $values = $action->load($idContent)->getValues();
        
        $this->assertTrue($values[$field->getElementName()] == '1 - Test_Setting');
        $this->assertTrue($values['id'] == $idContent);
        
        $response = new Response();
        $action->onStart($response);
        $this->assertNotEmpty($response->getContent());
        $fieldValues = $field->getValues();
        
        $this->assertTrue(count($fieldValues) == 1);
        
        $field->set(AbstractField::OPTION_AUTOCOMPLETE, true);
        
        $response = new Response();
        $action->onStart($response);
        $fieldValues = $field->getValues();
        
        $this->assertTrue(count($fieldValues) == 1);
        
        $row = reset($fieldValues);
        $this->assertTrue($row['value'] == '1 - Test_Setting');
    }
    
    private function _testDisplayRemovedForeignValues(
        Store $store,
        Many2manyField $field,
        $idContent,
        $idSetting
    ): void
    {
        $_POST = array();

        $action = $store->createActionInstance(Store::ACTION_EDIT);
        assert($action instanceof EditAction);

        $request = array(
            Store::PRIMARY_KEY_IN_REQUEST => $idContent
        );

        $store->setRequest($request);

        $field->set(AbstractField::OPTION_AUTOCOMPLETE, false);

        $this->core->db->delete(STORE_DEFAULT, array('id' => $idSetting));
        $values = $action->load($idContent)->getValues();

        $this->assertTrue(is_null($values[$field->getElementName()]));

        $response = new Response();
        $action->onStart($response);
        $fieldValues = $field->getValues();

        $this->assertTrue(count($fieldValues) == 1);
    }
}
