<?php

namespace tests\Bundle\Store\Field;

use Core;
use EditAction;
use FestiTestCase;
use FieldException;
use InsertAction;
use Response;
use Store;

class MultiFileFieldTest extends FestiTestCase
{
    const VALIDATE_VALUE_ATTRIBUTE = 'Test::onValidateFileField';

    private function _getFileValues(): array
    {
        $_FILES['name'] = array(
            'name'     => array(
                'phpunit.png',
                'phpunit2.png',
            ),
            'tmp_name' => array(
                FS_RESOURCES_ROOT.'phpunit.png',
                FS_RESOURCES_ROOT.'phpunit.png',
            ),
            'type'     => array(
                'image/png',
                'image/png',
            ),
        );

        return array(
            'name' => $_FILES['name']['name'],
            'path' => $_FILES['name']['tmp_name'],
        );
    }

    public function testUploadFileInput(): void
    {
        $options = $this->_getOptions();

        $values = $this->_getFileValues();
        list($data, $response) = $this->_insertRow($options, $values);

        $result = json_encode(
            array(
                '1_phpunit.png;0;image/png',
                '1_phpunit2.png;0;image/png',
            )
        );

        $this->assertEquals($data['id'], '1');
        $this->assertEquals($data['name'], $result);
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);
    } // end testUploadFileInput

    public function testEditFileInput(): void
    {
        $idRow = $this->_testUploadNewFile();

        $this->_testUpdateWithNewFile($idRow);

        $this->_testUpdateWithoutNewFile($idRow);

        $this->_testRemoveFile($idRow);

    } // end testUploadFileInput

    private function _insertRow($options, $values): array
    {
        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $this->assertInstanceOf(InsertAction::class, $action);

        $this->_onPrepareFileField($store);

        FestiTestCase::preparePostForStore($action, $values, $store);

        $response = new Response();

        $action->onStart($response);

        $data = $store->load();

        return array(array_pop($data), $response, $store);
    }

    private function _testUploadNewFile()
    {
        $options = $this->_getOptions();

        $_FILES['name'] = array(
            'name'     => array(
                'phpunitNew.png',
                'phpunitNew2.png',
            ),
            'tmp_name' => array(
                FS_RESOURCES_ROOT.'phpunit.png',
                FS_RESOURCES_ROOT.'phpunit.png',
            ),
            'type'     => array(
                'image/png',
                'image/png'
            )
        );

        $values = array(
            'name' => $_FILES['name']['name'],
            'path' => $_FILES['name']['tmp_name']
        );

        list($data, $response) = $this->_insertRow($options, $values);

        $idRow = $data['id'];

        $result = json_encode(
            array(
                $idRow.'_phpunitNew.png;0;image/png',
                $idRow.'_phpunitNew2.png;0;image/png',
            )
        );

        $this->assertEquals($data['name'], $result);
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);

        return $idRow;
    }

    private function _testUpdateWithNewFile($idRow)
    {
        $options = $this->_getOptions();

        $fileName = '__ID_____NAME_____FIELD__.__EXT__';
        $options['model']['fields']['value']['fileName'] = $fileName;

        $options['model']['fields']['id_parent'] = array(
            'type'      => 'text',
            'name'      => 'id_parent',
            'caption'   => 'Parent ID',
        );

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $this->_onPrepareFileField($store);

        $action = $store->createActionInstance(Store::ACTION_EDIT);

        $this->assertInstanceOf(EditAction::class, $action);

        $_FILES['name'] = array(
            'name'     => array(
                'phpunitNew.png',
                'phpunitNew2.png',
            ),
            'tmp_name' => array(
                FS_RESOURCES_ROOT.'phpunit.png',
                FS_RESOURCES_ROOT.'phpunit.png',
            ),
            'type'     => array(
                'image/png',
                'image/png'
            )
        );

        $values = array(
            'name' => $_FILES['name']['name'],
            'path' => $_FILES['name']['tmp_name'],
            'id_parent' => time()
        );

        FestiTestCase::preparePostForStore($action, $values, $store, $idRow);

        $response = new Response();

        $action->onStart($response);

        $data = $store->load();

        $data = array_pop($data);


        $result = array(
            $idRow.'_phpunitNew_name.png;0;image/png',
            $idRow.'_phpunitNew2_name.png;0;image/png',
        );
        $json   = json_encode($result);

        $this->assertEquals($data['name'], $json);
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);

        return $store;
    }

    private function _testUpdateWithoutNewFile( $idRow)
    {
        $options = $this->_getOptions();

        $fileName = '__ID_____NAME_____FIELD__.__EXT__';
        $options['model']['fields']['value']['fileName'] = $fileName;

        $options['model']['fields']['id_parent'] = array(
            'type'      => 'text',
            'name'      => 'id_parent',
            'caption'   => 'Parent ID',
        );

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $this->_onPrepareFileField($store);

        $action = $store->createActionInstance(Store::ACTION_EDIT);

        $this->assertInstanceOf(EditAction::class, $action);

        $_FILES = array();
        $values = array(
            '_name' => $idRow.'_phpunit_name.png;0;image/png',
            'id_parent' => time()
        );

        FestiTestCase::preparePostForStore($action, $values, $store, $idRow);

        $response = new Response();

        $action->onStart($response);

        $data = $store->load();

        $data = array_pop($data);

        $result = array(
            $idRow.'_phpunitNew_name.png;0;image/png',
            $idRow.'_phpunitNew2_name.png;0;image/png',
        );
        $json   = json_encode($result);

        $this->assertEquals($data['name'], $json);
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);

        return $store;
    }

    private function _testRemoveFile($idRow)
    {
        $options = $this->_getOptions();

        $fileName = '__ID_____NAME_____FIELD__.__EXT__';
        $options['model']['fields']['value']['fileName'] = $fileName;
        $options['model']['fields']['value']['required'] = false;

        $options['model']['fields']['id_parent'] = array(
            'type'      => 'text',
            'name'      => 'id_parent',
            'caption'   => 'Parent ID',
        );

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_EDIT);

        $this->assertInstanceOf(EditAction::class, $action);

        $_FILES = array();
        $values = array(
            '_name' => '',
            'id_parent' => time()
        );

        FestiTestCase::preparePostForStore($action, $values, $store, $idRow);

        $response = new Response();

        $action->onStart($response);

        $data = $store->load();

        $data = array_pop($data);

        $this->assertTrue(empty($data['name']));
        $this->assertNotEmpty($_POST);
        $this->assertNotEmpty($response->notifications);
    }

    private function _onPrepareFileField(&$store)
    {
        $fileField = &$store->getModel()->getFieldByName('name');

        $fileFieldMock = $this->getMockBuilder('MultiFileField')
            ->setConstructorArgs(array(&$store))
            ->setMethods(array('isUploaded', 'copy', 'getUploadFilePath'))
            ->getMock();

        $attributes = $fileField->getAttributes();

        foreach ($attributes as $key => $value) {
            $fileFieldMock->set($key, $value);
        }

        $fileFieldMock->expects($this->any())
            ->method('isUploaded')
            ->will($this->returnValue(true));

        $fileFieldMock->expects($this->any())
            ->method('getUploadFilePath')
            ->will($this->returnValue($store->getDefaultUploadFilePath()));

        $fileFieldMock->expects($this->any())
            ->method('copy')
            ->will($this->returnValue(true));

        $fileFieldMock->setName($fileField->getName());

        $fileField = $fileFieldMock;

        return true;
    } // _onPrepareFileField

    public function testOnFieldException(): void
    {
        $options = $this->_getOptions();

        $value = static::VALIDATE_VALUE_ATTRIBUTE;

        $options['model']['fields']['value']['onValidateValue'] = $value;

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $values = array();
        FestiTestCase::preparePostForStore($action, $values, $store);

        $_FILES['name'] = array(
            'name'     => array(
                'phpunit.csv',
                'phpunit2.csv',
            ),
            'tmp_name' => array(
                FS_RESOURCES_ROOT.'phpunit.png',
                FS_RESOURCES_ROOT.'phpunit.png',
            ),
            'type'     => array(
                'image/png',
                'image/png'
            )
        );

        $response = new Response();

        try {
            $action->onStart($response);
            $this->fail("Expect FieldException");
        } catch (FieldException $exp) {
            $this->assertTrue($exp->getMessage() == "Invalid File Extension");
        }
    } // end testOnFieldException

    public function testOnRequiredFieldException(): void
    {
        $options = $this->_getOptions();

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $values = array();
        FestiTestCase::preparePostForStore($action, $values, $store);

        $_FILES = array();

        $response = new Response();

        try {
            $action->onStart($response);
            $this->fail("Expect FieldException");
        } catch (FieldException $exp) {
            $this->assertTrue($exp->getMessage() == "Value is required field");
        }
    } // end testOnRequiredFieldException

    private function _getOptions(): array
    {
        $options = array();

        $options['model'] = array(
            'table'   => array(
                'name'          => 'site_contents',
                'primaryKey'    => 'id',
                'exceptionMode' => true
            ),

            'fields'  => array(
                'value' => array(
                    'type'      => 'multiFile',
                    'name'      => 'name',
                    'caption'   => 'Value',
                    'fileName'  => '__ID_____NAME__.__EXT__',
                    'required'  => 'true',
                )
            ),

            'actions' => array(
                Store::ACTION_INSERT => array(
                    'type'    => Store::ACTION_INSERT,
                    'caption' => 'Add',
                ),
                Store::ACTION_EDIT => array(
                    'type'    => Store::ACTION_EDIT,
                    'caption' => 'Edit',
                ),
                Store::ACTION_REMOVE => array(
                    'type'    => Store::ACTION_REMOVE,
                    'caption' => 'Remove',
                ),
            )
        );

        return $options;
    }

    public function testCustomFileLink(): void
    {
        $idRow = $this->_testUploadNewFile();

        $options = $this->_getOptions();

        $options['model']['fields']['value']['link'] = '/test_download/%id%/';

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $action = $store->createActionInstance(Store::ACTION_EDIT);

        $this->assertInstanceOf(EditAction::class, $action);

        $_FILES = array();

        FestiTestCase::preparePrimaryKey($store, $idRow);

        $response = new Response();

        $action->onStart($response);

        $content = $response->content;

        $this->assertNotEmpty($content);

        $pattern = "#<a href=\"\/test_download/".$idRow."\/\">\s*".
                   $idRow."_phpunitNew.png\s*</a>#Umis";

        $this->assertMatchesRegularExpression(
            $pattern,
            $content,
            "Incorrect file url."
        );
    } // end testCustomFileLink

    public function testRemoveMessage(): void
    {
        $options = $this->_getOptions();

        $values = $this->_getFileValues();
        list($data, , $store) = $this->_insertRow($options, $values);

        $action = $store->createActionInstance(Store::ACTION_REMOVE);
        $primaryKeyValue = $data['id'];

        FestiTestCase::preparePrimaryKey($store, $primaryKeyValue);

        $store->addAllowedTableRow($primaryKeyValue);

        $response = new Response();
        $action->onStart($response);

        $needle = "content : \"<span class='txt-color-orangeDark'>".
                  "Value:</span> 1_phpunit.png, 1_phpunit2.png<br />\",";

        $this->assertNotFalse(strpos($response->getContent(), $needle));
    }
}
