<?php

namespace tests\Bundle\Store\Field;

use Core;
use FestiEvent;
use FestiTestCase;
use Response;
use SelectField;
use Store;
use StoreActionEvent;
use SystemPlugin;

class SelectFieldTest extends FestiTestCase
{
    public const EXPECTED_TOTAL_ROWS = 2;
    public const FORBIDDEN_SECTION   = 'section_2';
    
    public function testListViewWithNullValue(): void
    {
        $options = array();

        $options['model'] = $this->_getDefaultModel();

        $testValueCaption = "Test NULL value";
        $options['model']['fields']['value']['options'] =
            "NULL:".$testValueCaption."\n0:Zero";

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        //
        $action = $store->createActionInstance(Store::ACTION_INSERT);


        $values = array(
            'name' => 'NULL'
        );

        FestiTestCase::preparePostForStore($action, $values, $store);

        $response = new Response();
        $action->onStart($response);

        //
        $action = $store->createActionInstance(Store::ACTION_LIST);

        $response = new Response();
        $action->onStart($response);


        $this->assertMatchesRegularExpression(
            "#".$testValueCaption."#Umis",
            $response->getContent()
        );
    }

    public function testAllowEmpty(): void
    {
        $options = array();

        $options['model'] = $this->_getDefaultModel();

        $options['model']['fields']['value']['options'] =
            "option1:Option 1\noption2:Option 2";
        $options['model']['fields']['value']['allowEmpty'] =
            "true";
        $options['model']['fields']['value']['filter'] =
            "select";

        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );

        $field = $store->getModel()->getFieldByName('name');
        $this->assertCount(2, $field->valuesList);

        $this->_testInsertForm($store);
        $this->_testFilter($store);
    }
    
    public function testSelectFieldFilterConditions(): void
    {
        $options = array();
        
        $options['model'] = $this->_getDefaultModel();
    
        $options['model']['fields']['value']['options'] =
            "option1:Option 1\noption2:Option 2";
        $options['model']['fields']['value']['filter'] =
            "select";
        
        $sessionData = array(
          'DB_FILTERS' => array(
              'site_contents' => array(
                  'site_contents_name' => 'option1'
              )
          )  
        );
        
        Core::getInstance()->setSessionData($sessionData);
        
        $store = Core::getInstance()->createStoreInstance(
            'site_contents',
            $options
        );
        
        $result = $store->getFieldFiltersConditions();
    
        FestiTestCase::cleanFiltersForStore($store);
        
        $this->assertEquals($result['site_contents_name']['value'], 'option1');
        
        $this->assertEquals(
            $result['site_contents_name']['display_value'],
            'Option 1'
        );
    }

    private function _testInsertForm(&$store): void
    {
        $action = $store->createActionInstance(Store::ACTION_INSERT);

        $response = new Response();
        $action->onStart($response);

        $content = $response->getContent();

        $this->assertMatchesRegularExpression(
            "#<option value=\"\">\.\.\.<\/option>#Umis",
            $content
        );

        $this->assertMatchesRegularExpression(
            "#<option value=\"option1\"\s*>Option 1<\/option>#Umis",
            $content
        );

        $this->assertMatchesRegularExpression(
            "#<option value=\"option2\"\s*>Option 2<\/option>#Umis",
            $content
        );
    }
    
    public function testOnSelectFieldOptionPermisions(): void
    {
        $core = Core::getInstance();
        $systemPlugin = new SystemPlugin();
        $systemPlugin->addEventListener(
            FestiTestCase::EVENT_ON_HAS_USER_PERMISSION_TO_SECTION,
            static function (FestiEvent &$event): void {
                $section = $event->getTargetValueByKey('sectionName');
                if ($section == static::FORBIDDEN_SECTION) {
                    $hasPermission = &$event->getTargetValueByKey('has_permission');
                    $hasPermission = false;
                }
            }
        );
        $core->setSystemPlugin($systemPlugin);
        
        $store = $core->createStoreInstance('select_permission_field');
        $store->addEventListener(
            Store::EVENT_ON_BEFORE_LOAD_VALUES,
            function (StoreActionEvent &$event): void {
                $store = &$event->getTargetValueByKey('store');
    
                $fields = &$store->getModel()->getFields();
                $field = $fields['name'];
    
                $this->assertTrue($field instanceof SelectField);
                
                $values = $field->getValuesList();
    
                $search = &$store->getModel()->getSearch();
                $search['name&IN'] = array_keys($values);
            }
        );
        
        $values = array(
            array(
                'name' => 'active'
            ),
            array(
                'name' => 'pending'
            ),
            array(
                'name' => 'error'
            )
        );
    
        $core->db->massInsert($store->getName(), $values);
        
        $rows = $store->load();
        
        $this->assertEquals(count($values), count($rows));
    
        $action = $store->createActionInstance(Store::ACTION_LIST);
    
        $response = new Response();
        $action->onStart($response);
    
        $this->assertNotEmpty($response->content);
        
        $total = $store->getTotalCount();
    
        $this->assertEquals(static::EXPECTED_TOTAL_ROWS, $total, 'Total Not Equals');
    }
    
    private function _testFilter(&$store): void
    {
        $action = $store->createActionInstance(Store::ACTION_LIST);

        $response = new Response();
        $action->onStart($response);

        $content = $response->getContent();
        
        $this->assertMatchesRegularExpression(
            "#<select.*class=\".*db-filed-filter-select.*\".*>\s+".
            "<option value=\"\">\.\.\.<\/option>\s+".
            "<option value=\"option1\"\s*>Option 1<\/option>\s+".
            "<option value=\"option2\"\s*>Option 2<\/option>#Umis",
            $content
        );
    }

    private function _getDefaultModel(): array
    {
        return array(
            'table' => array(
                'name'       => 'site_contents',
                'primaryKey' => 'id',
                'exceptionMode' => true
            ),

            'fields' => array(
                'value' => array(
                    'type' => 'select',
                    'name' => 'name',
                    'caption' => 'Value'
                )
            ),
            'actions' => array(
                Store::ACTION_LIST => array(
                    'type' => Store::ACTION_LIST,
                    'caption' => 'List'
                ),
                Store::ACTION_INSERT => array(
                    'type' => Store::ACTION_INSERT,
                    'caption' => 'Add'
                )
            )
        );
    }
}
