<?php

class TestUtils
{
    public static function getDbConnection($dbConfig)
    {
        $db = new PDO(
            $dbConfig['DB_DSN'],
            $dbConfig['DB_USER'],
            $dbConfig['DB_PASSWD']
        );

        $GLOBALS['config']['db']['dsn'] = $dbConfig['DB_DSN'];
        $GLOBALS['config']['db']['user'] = $dbConfig['DB_USER'];
        $GLOBALS['config']['db']['pass'] = $dbConfig['DB_PASSWD'];

        $db->setAttribute(PDO::ATTR_DEFAULT_FETCH_MODE, PDO::FETCH_ASSOC);
        $db->setAttribute(PDO::ATTR_ORACLE_NULLS, PDO::NULL_EMPTY_STRING);
        $db->setAttribute(PDO::ATTR_CASE, PDO::CASE_NATURAL);
        $db->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);

        /*
        $res = $db->query('SET NAMES utf8');

        if (!$res) {
            throw new SystemException('Database connection error');
        }
        */

        $db = DataAccessObject::factory($db);

        return $db;
    }

    public static function cleanDatabase(): void
    {
        $db = Core::getInstance()->db;

        if (!$db) {
            throw new SystemException("Undefined database connection.");
        }

        $db->getDriver()->cleanDatabase($db);
    } // end cleanDatabase

    /**
     * Setup a database dump. You can use single file or folder with dump files:
     * /dump/init.sql
     * /dump/updates1.sql
     * /dump/updates2.sql
     * /dump/updatesN.sql
     *
     * @param $dumpPath
     * @throws SystemException
     * @deprecated use FestiTool::migrateDatabase
     */
    public static function installDatabase($dumpPath)
    {
        $db = Core::getInstance()->db;

        if (!$db) {
            throw new SystemException("Undefined database connection.");
        }

        $files = array();

        if (is_dir($dumpPath)) {
            $filesIterator = new RecursiveIteratorIterator(
                new RecursiveDirectoryIterator(
                    $dumpPath,
                    RecursiveDirectoryIterator::SKIP_DOTS
                ),
                RecursiveIteratorIterator::CHILD_FIRST
            );

            foreach ($filesIterator as $fileInfo) {
                if ($fileInfo->isFile()) {
                    $path = $fileInfo->getRealPath();

                    $regExp = "#updates([0-9]+)\.(sql|php)#Umis";
                    if (preg_match($regExp, $path, $matches)) {
                        $files[$matches[1]] = $path;
                    } else if (preg_match("#init.sql#Umis", $path, $matches)) {
                        $files[0] = $path;
                    }
                }
            }

            ksort($files, SORT_NUMERIC);
        } else if (is_file($dumpPath)) {
            $files[] = $dumpPath;
        } else {
            throw new SystemException("Not found dump path: ".$dumpPath);
        }

        foreach ($files as $file) {

            $fileInfo = pathinfo($file);
            if ($fileInfo['extension'] == "php") {
                require_once $file;

                $migrationClassName =  ucfirst($fileInfo['filename']);
                $migration = new $migrationClassName();
                $migration->apply();

                continue;
            }
            
            if (defined('MYSQL_PATH')) {
                $initDatabasesCmd = MYSQL_PATH.' -u '.
                    $GLOBALS['config']['db']['user'].
                    ' -p'.$GLOBALS['config']['db']['pass'].' '.
                    $GLOBALS['config']['db']['name'].
                    ' < '.$file;
                
                    `$initDatabasesCmd`;
            } else {
                $content = file_get_contents($file);
                
                $rows = array_filter(explode(";", $content));
                
                foreach ($rows as $sql) {
                    $sql = trim($sql);
                    if (!empty($sql)) {
                        $db->query($sql);
                    }
                }
            }
        }

        return true;
    } // end installDatabase

    /**
     * Running all tests from plugins.
     *
     * @throws SystemException
     */
    public static function execPlugins() : void
    {
        $core = Core::getInstance();
        $path = $core->getOption(Core::OPTION_PLUGINS_PATH);

        static::execTests($path);
    } // end  execPlugins

    /**
     * Running all tests from theme.
     *
     * @throws SystemException
     */
    public static function execThemeTests() : void
    {
        $core = Core::getInstance();
        $path = $core->getOption(Core::OPTION_THEME_PATH);

        static::execTests($path.'/../');
    } // end execThemeTests

    /**
     * Running all tests from the path.
     *
     * @param string $path
     * @throws SystemException
     */
    public static function execTests(string $path) : void
    {
        $handle = opendir($path);
        if (!$handle) {
            throw new SystemException("Permission error: ".$path);
        }
        
        $paths = array();
        while (false !== ($entry = readdir($handle))) {
            
            if ($entry == "." || $entry == "..") {
                continue;
            }
            
            $testPath = $path.$entry.DIRECTORY_SEPARATOR.'tests';

            if (is_dir($testPath)) {
                $paths[] = $testPath;
            }
        }

        closedir($handle);

        foreach ($paths as $path) {
            static::run($path);
        }
    } // end execTests

    /**
     * Running all tests from file.
     *
     * @param $path Path to test file
     * @return mixed
     */
    public static function run($path)
    {
        if (class_exists('PHPUnit_TextUI_TestRunner')) {
            $runner = new PHPUnit_TextUI_TestRunner();
        } else {
            $runner = new PHPUnit\TextUI\TestRunner();
        }

        $test = $runner->getTest($path, '', 'Test.php');
        $result = $runner->dorun($test, [], false);

        return $result;
    } // end run
    

}