<?php

abstract class AbstractObjectDriver implements IObjectDriver
{
    protected $reservedWords = array(
        'NOW()',
        'NOT NULL',
        'NULL',
        'CURRENT_DATE()',
        'CURRENT_TIME()',
        'CURRENT_DATE',
        'CURRENT_TIME',
        'NOW'
    );

    /**
     * Returns general error code.
     *
     * @param $code
     * @return mixed
     */
    public function getErrorCode($code): int
    {
        return (int) $code;
    } // end getErrorCode

    public function createSelectQuery(
        array $columns,
        string $from,
        ?array $joins = null,
        ?array $where = null,
        ?array $orderBy = null,
        ?int $limit = null,
        ?int $offset = null,
        ?array $groupBy = null,
        ?array $having = null
    ): string
    {
        $sql = "SELECT ";

        $queryColumns = array();
        foreach ($columns as $columnAlias => $columnName) {
            if ($columnName == $columnAlias) {
                $columnAlias = null;
            }
            $columnName = $this->quoteColumnName($columnName);
            $queryColumns[] = $columnAlias ? $columnName.' AS '.$columnAlias : $columnName;
        }

        if (!$queryColumns) {
            $queryColumns[] = '*';
        }

        $sql .= join(", ", $queryColumns)." FROM ".$from;

        if ($joins) {
            $sql .= " ".join(" ", $joins);
        }

        if ($where) {
            $sql .= " WHERE ".join(" AND ", $where);
        }

        if ($groupBy) {
            $sql .= " GROUP BY ".join(", ", $groupBy);
        }

        if ($having) {
            $sql .= " HAVING ".join(" AND ", $having);
        }

        if ($orderBy) {
            $sql .= " ORDER BY ".join(", ", $orderBy);
        }

        if ($limit) {
            $sql .= " LIMIT ".$limit;
            if ($offset) {
                $sql .= " OFFSET ".$offset;
            }
        }

        return $sql;
    } // end createSelectQuery

    public function getSplitOnPages(IDataAccessObject $object, string $query, int $col, int $page): array
    {
        throw new DatabaseException("Undefined method");
    }

    public function createInsertQuery(
        IDataAccessObject $object, string $table, array $values, bool $isUpdateDuplicate = false
    ): string
    {
        $values = $this->_getInsertValues($object, $values);
        $columns = array_keys($values);

        foreach ($columns as &$column) {
            $column = $this->quoteColumnName($column);
        }
        unset($column);

        $sql = "INSERT INTO ".
            $this->quoteTableName($table)." ".
            "(".join(", ", $columns).") ".
            "VALUES (".join(", ", $values).")";

        if ($isUpdateDuplicate) {
            $sql = $this->_getUpdateDuplicateSQL($object, $sql, $values);
        }

        return $sql;
    }
    
    protected function isExpression($column)
    {
        return strpos($column, '(') !== false;
    } // end isExpression
    
    protected function quoteExpressionColumnName(string $key): string
    {
        $expression = strstr($key, '(', true);
        $key = trim(strstr($key, '('));
        
        $key = substr($key, 0, -1);
        $key = substr($key, 1);
        
        $columns = explode(',', $key);
        
        foreach ($columns as &$column) {
            $column = trim($column);
            
            if (!preg_match('/^[a-zA-Z0-9._\(\)]*$/Umis', $column)) {
                continue;
            }
            $column = $this->quoteColumnName($column);
        }
        
        return $expression.'('.implode(', ', $columns).')';
    } // end quoteExpressionColumnName
    
    private function _getInsertValues(IDataAccessObject $object, array $values): array
    {
        foreach ($values as &$item) {
            if ( is_null($item) ) {
                $item = 'NULL';
                continue;
            }

            if (!in_array($item, $this->reservedWords)) {
                $item = $object->quote($item);
            }
        }
        unset($item);

        return $values;
    } // end _getInsertValues

    private function _getUpdateDuplicateSQL(IDataAccessObject $object, string $sql, array $values): string
    {
        $type = $object->getDatabaseType();

        switch ($type) {
            case DataAccessObject::TYPE_MYSQL:
                $sql .= " ON duplicate KEY UPDATE ";
                break;

            case DataAccessObject::TYPE_PGSQL:
                $msg = sprintf(
                    "Method Insert Not Support Third Param For %s DB Type.",
                    DataAccessObject::TYPE_PGSQL
                );
                throw new DatabaseException($msg);

            default:
                throw new DatabaseException("Undefined database type");
        }

        $rows = array();

        foreach ($values as $field => $value) {
            $rows[] = $field." = ".$value;
        }

        $sql .= join(", ", $rows);

        return $sql;
    } // end _getUpdateDuplicateSQL

    public function cleanDatabase(IDataAccessObject $object): bool
    {
        $tables = $object->getTables();

        $object->setForeignKeyChecks(false);

        foreach ($tables as $table) {
            $object->deleteTable($table);
        }

        $object->setForeignKeyChecks(true);

        return true;
    } // end cleanDatabase

}