<?php

namespace core\orm;

use core\orm\relation\IRelation;
use core\orm\relation\ManyToManyRelation;
use core\orm\relation\OneToManyRelation;

class EntityManager
{
    /**
     * @var \IDataAccessObject
     */
    private \IDataAccessObject $_connection;

    /**
     * @var IEntityParse
     */
    private IEntityParse $_parse;

    private function __construct(\IDataAccessObject $connection, ?IEntityParse $parser = null)
    {
        $this->_connection = $connection;

        $this->_parse = $parser ?? new ReflectionEntityParser();
    } // end __construct

    /**
     * @throws EntityException
     */
    public function sync(object &$entity): bool
    {
        $table = $this->_parse->getModel($entity);

        $this->_connection->begin();

        try {
            $this->_syncEntity($table, $entity);

            $this->_syncRelations($table, $entity);

            $this->_connection->commit();

        } catch (\Exception $exp) {
            $this->_connection->rollback();
            throw $exp;
        }

        return true;
    } // end sync

    /**
     * @throws EntityException
     */
    private function _syncEntity(Table $table, &$entity): void
    {
        if ($this->_isNewEntity($table, $entity)) {
            $this->_create($table, $entity);
        } else {
            $this->_update($table, $entity);
        }
    }

    /**
     * @throws EntityException
     */
    private function _update(Table $table, &$entity): void
    {
        if ($entity instanceof IEntityListener) {
            $entity->onUpdate();
        }

        $values = $this->getValues($table, $entity);

        $search = array();
        $primaryKeys = $table->getPrimaryKey();
        foreach ($primaryKeys as $primaryKey) {
            $getter = $primaryKey->getGetter();
            $search[$primaryKey->getName()] = $entity->{$getter}();
        }

        if (!$search) {
            throw new EntityException("Undefined update condition for entity ".get_class($entity));
        }

        $this->_connection->update($table->getName(), $values, $search);
    } // end _update

    protected function getValues(Table $table, &$entity): array
    {
        $columns = $table->getColumns();

        $values = array();
        foreach ($columns as $column) {
            $key = $column->getName();
            $getter = $column->getGetter();

            $values[$key] = $entity->{$getter}();

            if (!$column->isNullable() && !$values[$key]) {
                throw new EntityFieldException($key);
            }

        }

        return $values;
    }

    private function _create(Table $table, &$entity): void
    {
        if ($entity instanceof IEntityListener) {
            $entity->onCreate();
        }

        $primaryKeyValue = $this->_insert($table, $entity);
        $primaryKeys = $table->getPrimaryKey();
        foreach ($primaryKeys as $primaryKey) {
            $setter = $primaryKey->getSetter();
            $entity->{$setter}($primaryKeyValue);
        }
    } // end _create

    /**
     * @throws EntityException
     */
    private function _syncRelations(Table $table, &$entity): void
    {
        $columns = $table->getColumnsWithRelation();

        foreach ($columns as $column) {
            $getter = $column->getGetter();
            $rows = $entity->{$getter}();

            $relation = $column->getRelation();

            if ($relation instanceof OneToManyRelation) {
                $this->_updateOneToMany($relation, $entity, $rows);
            } else if ($relation instanceof ManyToManyRelation) {
                $this->_updateManyToMany($relation, $entity, $rows);
            } else {
                throw new EntityException("Undefined relation ".get_class($relation));
            }
        }
    } // end _syncRelations

    /**
     * @throws EntityException
     */
    private function _updateOneToMany(IRelation $relation, &$entity, &$rows): void
    {
        $relationTable = $relation->getTable();

        $sourceValue = $this->_getSourceRelationValue($relation, $entity);

        $foreignKey = $relation->getForeignKeyColumnName();

        $search = array(
            $foreignKey => $sourceValue
        );

        $this->_connection->delete($relationTable->getName(), $search);

        if (is_array($rows)) {
            foreach ($rows as &$row) {
                $this->_setForeignKeyValue($sourceValue, $relation, $row);
                $this->_syncEntity($relationTable, $row);
            }
        }
    } // end _updateOneToMany

    private function _updateManyToMany(ManyToManyRelation $relation, &$entity, &$rows): void
    {
        $sourceValue = $this->_getSourceRelationValue($relation, $entity);

        $linkedTable = $relation->getLinkTable();
        $foreignColumnName = $relation->getForeignKeyColumnName();
        $linkColumnName = $relation->getLinkKey();

        $search = array(
            $foreignColumnName => $sourceValue,
            $linkColumnName.'&NOT IN' => $rows
        );

        $this->_connection->delete($linkedTable, $search);

        foreach ($rows as $row) {
            $values = array(
                $foreignColumnName => $sourceValue,
                $linkColumnName => $row
            );

            $this->_connection->insert($linkedTable, $values, true);
        }
    } // end _updateManyToMany

    private function _setForeignKeyValue($sourceValue, IRelation $relation, &$entity): void
    {
        $foreignKey = $relation->getForeignKeyColumnName();

        $methodName = 'set'.static::getColumnPostfixMethodName($foreignKey);

        $entity->{$methodName}($sourceValue);
    } // end _setForeignKeyValue

    private function _getSourceRelationValue(IRelation $relation, $entity)
    {
        $sourceKey = $relation->getSourceKeyColumnName();

        $methodName = 'get'.static::getColumnPostfixMethodName($sourceKey);
        return $entity->{$methodName}();
    } // end _getSourceRelationValue

    public static function getColumnPostfixMethodName(string $columnName):string
    {
        if ($columnName == "id") {
            return 'ID';
        }

        $parts = explode("_", $columnName);

        if ($parts[0] == "id") {
            $newParts = array_slice($parts, 1);
            $newParts[] = 'ID';
            $parts = $newParts;
        }

        return join('', array_map('ucfirst', $parts));
    } // end getColumnPostfixMethodName

    private function _insert(Table $table, $entity): int
    {
        $values = $this->getValues($table, $entity);

        return $this->_connection->insert($table->getName(), $values);
    } // end _insert

    /**
     * @throws EntityException
     */
    private function _isNewEntity(Table $table, $entity): bool
    {
        $columns = $table->getPrimaryKey();

        foreach ($columns as $column) {

            $methodName = $column->getGetter();
            $getter = array($entity, $methodName);
            if (!is_callable($getter)) {
                throw new EntityException(
                    "Undefined method ".$methodName." into entity ".get_class($entity)
                );
            }

            try {
                $value = $entity->{$methodName}();
                if ($value) {
                    return false;
                }
            } catch (\Exception $exp) {
                return true;
            }
        }

        return true;
    } // end _isNewEntity

    public static function createInstance(\IDataAccessObject $connection, ?IEntityParse $parser = null): EntityManager
    {
        return new self($connection, $parser);
    } // end createInstance

}