<?php
namespace core\orm;

use core\orm\relation\IRelation;
use core\orm\relation\ManyToManyRelation;
use core\orm\relation\OneToManyRelation;

class ReflectionEntityParser implements IEntityParse
{
    private static $_cache = array();

    public function getModel(object $entity): Table
    {
        $className = get_class($entity);

        return $this->_getSchema($className);
    } // end getModel

    private function _getSchema(string $className): Table
    {
        if (array_key_exists($className, static::$_cache)) {
            return static::$_cache[$className];
        }

        $reflectionClass = new \ReflectionClass($className);

        $tableName = $this->getTableName($reflectionClass);

        $table = new Table();
        $table->setName($tableName);

        $this->_prepareColumns($reflectionClass, $table);

        static::$_cache[$className] = $table;

        return static::$_cache[$className];
    } // end _getSchema


    /**
     * @param \ReflectionClass $reflectClass
     * @param Table $table
     * @return void
     * @throws EntityException
     */
    private function _prepareColumns(\ReflectionClass $reflectClass, Table &$table): void
    {
        $methods = $reflectClass->getMethods();

        foreach ($methods as $method) {
            $column = $this->getColumnByMethod($method, $reflectClass);
            if ($column) {
                $table->addColumn($column);
            }

            $columnWithRelation = $this->getColumnWithRelationByMethod($method, $reflectClass);
            if ($columnWithRelation) {
                $table->addColumn($columnWithRelation);
            }
        }
    } // end _prepareColumns

    protected function getColumnWithRelationByMethod(\ReflectionMethod $method, \ReflectionClass $reflectClass): ?Column
    {
        $comment = $method->getDocComment();

        $mapper = array(
            'oneToMany' => OneToManyRelation::class,
            'manyToMany' => ManyToManyRelation::class
        );

        $instance = null;
        foreach ($mapper as $type => $class) {
            $relation = $this->_getAnnotationValue($type, $comment);

            if ($relation) {
                $instance = new $class($relation);
                break;
            }
        }

        if ($instance) {
            return $this->getColumnWithRelation($instance, $method, $reflectClass);
        }

        return null;
    } // end _getColumnWithRelationByMethod

    protected function getColumnWithRelation(
        IRelation $instance, \ReflectionMethod $method, \ReflectionClass $reflectClass
    ): Column
    {
        $entityClass = $instance->getEntityClass();

        if (!class_exists($entityClass)) {
            throw new EntityException(
                "Undefined entity class ".$entityClass." for method ".$method->getName()
            );
        }

        $table = $this->_getSchema($entityClass);
        $instance->setTable($table);

        $columnName = $this->getColumnNameByMethod($method);

        $column = new Column($columnName);
        $column->setRelation($instance);

        $this->prepareColumnCallbacks($column, $method, $reflectClass);

        return $column;
    } // end getColumnWithRelation

    protected function getColumnByMethod(\ReflectionMethod $method, \ReflectionClass $reflectClass): ?Column
    {
        $comment = $method->getDocComment();

        $attribute = $this->_getAnnotationValue('column', $comment);

        if (is_null($attribute)) {
            return null;
        }

        if (!$attribute || !array_key_exists('name', $attribute)) {
            $attribute['name'] = $this->getColumnNameByMethod($method);
        }

        $column = new Column($attribute['name']);

        $this->_prepareColumn($column, $method, $reflectClass);

        return $column;
    }

    /**
     * @throws EntityException
     */
    private function _prepareColumn(Column &$column, \ReflectionMethod $method, \ReflectionClass $reflectClass): void
    {
        $comment = $method->getDocComment();

        $primaryKey = $this->_getAnnotationValue('primaryKey', $comment);
        if (!is_null($primaryKey)) {
            $column->setPrimaryKey(true);
        }

        $isNullable = $this->_getAnnotationValue('null', $comment);
        if (!is_null($isNullable)) {
            $column->setNullable(true);
        }

        $this->prepareColumnCallbacks($column, $method, $reflectClass);

    } // end _prepareColumn

    protected function prepareColumnCallbacks(
        Column &$column, \ReflectionMethod $method, \ReflectionClass $reflectClass
    ): void
    {
        $methodName = $method->getName();
        $methodPostfix = substr($methodName, 3, strlen($methodName) - 3);
        $methodPrefix = substr($methodName, 0, 3);

        $callbackPrefixes = array('get', 'set');

        if (!in_array($methodPrefix, $callbackPrefixes)) {
            throw new EntityException("Column annotation has to be use on methods with get or set prefix");
        }

        foreach ($callbackPrefixes as $methodPrefix) {

            $callbackMethod = $methodPrefix.$methodPostfix;

            if (!method_exists($reflectClass->getName(), $callbackMethod)) {
                throw new EntityException(
                    "Undefined method ".$callbackMethod." into entity ".$reflectClass->getName()
                );
            }

            $columnMethod = 'set'.ucfirst($methodPrefix).'ter';
            $column->{$columnMethod}($callbackMethod);
        }
    } // end _prepareColumnCallbacks

    protected function getColumnNameByMethod(\ReflectionMethod $method): string
    {
        $methodName = $method->getName();
        $methodName = preg_replace("#^(set|get)#Ums", "", $methodName);
        if ($methodName == "ID") {
            $columnName = "id";
        } else {
            $columnName = $this->convertToSnakeCase($methodName);
        }

        return $columnName;
    } // end _getColumnNameByMethod

    protected function convertToSnakeCase(string $str): string
    {
        return strtolower(preg_replace('/(?<!^)[A-Z]/', '_$0', $str));
    }


    private function _getAnnotationValue($name, $comment): ?array
    {
        //$regExp = "#@".preg_quote($name)." (?<name>[a-zA-Z_=\ ]+)[\r\n]#Umis";

        $annotation = "@".$name;
        $regExp = "#".preg_quote($annotation)."\ {0,1}(?<expression>\(([^\)]*)\)|[a-zA-Z]+|[\r\n])#Umis";

        if (preg_match($regExp, $comment, $matches)) {
            return $this->_getAnnotationAttributes($matches['expression']);
        }

        return null;
    } // end _getAnnotationValue

    private function _getAnnotationAttributes(string $expression): array
    {
        $result = array();

        $regExp = "#(?<keys>[a-zA-Z_]+):\s{0,1}['|\"](?<values>[a-zA-Z0-9_\\\\]+)['|\"]#Umis";
        if (preg_match_all($regExp, $expression, $matches)) {
            $result = array_combine($matches['keys'], $matches['values']);
        }

        return $result;
    } // end _getAnnotationAttributes

    protected function getTableName(\ReflectionClass $reflectClass): string
    {
        $comment = $reflectClass->getDocComment();
        $attributes = $this->_getAnnotationValue('table', $comment);

        if (!$attributes || !array_key_exists('name', $attributes)) {
            throw new EntityException("Undefined table name into ".$reflectClass->getName());
        }

        return $attributes['name'];
    } // end getTableName

}