<?php

class DefaultTheme extends Template implements ITheme
{
    public $core;

    public function __construct(array $info)
    {
        $this->core = Core::getInstance();
        $path = Core::getInstance()->getOption(Core::OPTION_THEME_TEMPLATE_PATH);
        parent::__construct($path);

        $this->info = $info;
    } // end __construct

    /**
     * Fire Theme Event.
     *
     * @param string $themeEventName
     * @param array $target
     * @return array
     * @throws SystemException
     */
    public function doThemeEvent(string $themeEventName, array &$target = array()): array
    {
        $core = Core::getInstance();

        $target[ThemeEvent::TARGET_THEME_INSTANCE_KEY_NAME] = &$this;

        $event = new ThemeEvent($themeEventName, $target);

        $results = $core->dispatchEvent($event);

        // TODO: Move to Core
        if (empty($results)) {
            $results = $core->getSystemPlugin()->dispatchEvent($event);
        }

        if ($results) {
            $results = array_filter($results);
        }

        return $results ?? array();
    } // end doThemeEvent

    /**
     * Returns content in main theme context.
     *
     * @param string $content
     * @param string $template
     * @return string
     * @throws SystemException
     */
    public function fetchMain(string $content, $template = 'main.phtml'): string
    {
        $this->content = $content;

        return $this->fetch($template);
    } // end fetchMain

    /**
     * @param string $eventName
     * @param array $target
     * @throws SystemException
     */
    private function _fireEvent(string $eventName, array &$target): void
    {
        $core = Core::getInstance();

        $event = new FestiEvent($eventName, $target);

        // TODO: Move to Core
        $core->dispatchEvent($event);
        $core->getSystemPlugin()->dispatchEvent($event);
    }

    /**
     * @return string
     * @throws SystemException
     */
    public function fetchLogo(): string
    {
        $core = Core::getInstance();

        $fields = array(
            'basehttp' => array(
                'type'     => static::FIELD_TYPE_SECURITY_STRING,
                'required' => true,
                'error'    => __('basehttp is required'),
            ),
        );

        $data = $this->getPreparedData($this->info, $fields, $errors);

        if ($errors) {
            throw new SystemException(array_shift($errors));
        }

        $target = array(
            'link'    => $data['basehttp'],
            'logo'    => $data['basehttp'].'static/images/logo.png',
            'content' => null,
        );

        $core->fireEvent(static::EVENT_THEME_ON_PREPARE_LOGO, $target);

        if (isset($target['content'])) {
            return $target['content'];
        }

        $this->assign($target);

        return $this->fetch('logo.phtml');
    }

    /**
     * @param $area
     * @param $template
     * @return string
     * @throws SystemException
     */
    public function fetchMenu($area = false, $template = false): string
    {
        $core = Core::getInstance();

        $res = $this->doThemeEvent(static::EVENT_THEME_FETCH_MENU);
        if (!empty($res)) {
            return join('', $res);
        }

        $systemPlugin = $core->getSystemPlugin();

        // TODO: Move getStructureMenu to ISystemPlugin
        $structureMenu = array();
        $callback = array($systemPlugin, 'getStructureMenu');
        if (is_callable($callback)) {
            $params = array($area);
            $structureMenu = call_user_func_array($callback, $params);
        }

        if (!$area) {
            $area = 'default';
        }

        $currentItem = isset($_SERVER['REQUEST_URI']) ? $_SERVER['REQUEST_URI'] : null;

        $vars = array(
            'currentItem' => $currentItem,
            'name'        => $area.'-menu',
            'items'       => array_values($structureMenu),
            'activeItem'  => $systemPlugin->getActiveMenu(),
        );

        $target = &$vars;

        $this->_fireEvent(ISystemPlugin::EVENT_ON_PREPARE_MENU_ITEMS, $target);

        if (!$template) {
            $template = 'menu.php';
        }

        $this->assign($vars);

        return $this->fetch($template);
    }

    /**
     * @return string
     * @throws SystemException
     */
    public function fetchHead(): string
    {
        $core = Core::getInstance();

        $this->jsVersion     = $this->getSettingOrDefault('js_version', '1');
        $this->engineBaseUrl = $core->getOption('engine_http_base');
        $this->engineCssUrl  = $core->getOption('engine_css_url');
        $this->siteCaption   = $this->getSettingOrDefault('site_caption', '');
        $this->host          = $this->getSettingOrDefault('host');

        return $this->fetch('head.phtml');
    }

    /**
     * @param string $key
     * @param string|null $default
     * @return string|null
     * @throws SystemException
     */
    protected function getSettingOrDefault(
        string $key, ?string $default = null
    ): ?string
    {
        $systemPlugin = Core::getInstance()->getSystemPlugin();

        if ($systemPlugin->hasSetting($key)) {
            return $systemPlugin->getSetting($key);
        }

        return $default;
    }

    protected function getSetting(string $key): ?string
    {
        $systemPlugin = Core::getInstance()->getSystemPlugin();

        if ($systemPlugin->hasSetting($key)) {
            return $systemPlugin->getSetting($key);
        }

        return null;
    }

    protected function hasSetting(string $key): bool
    {
        $systemPlugin = Core::getInstance()->getSystemPlugin();

        return $systemPlugin->hasSetting($key);
    }

    /**
     * @return string
     * @throws SystemException
     */
    public function fetchNotifications(): string
    {
        $core = Core::getInstance();

        $this->notifications = $core->popParam('notifications');

        return $this->fetch('notifications.phtml');
    }

    /**
     * @return string
     * @throws SystemException
     */
    public function fetchThemeTitle(): string
    {
        $core = Core::getInstance();

        $data = array(
            'title' => $core->user->get('auth_login'),
        );

        $this->doThemeEvent(static::EVENT_THEME_TITLE, $data);

        return $data['title'];
    } // end fetchThemeTitle

    /**
     * @return string
     * @throws SystemException
     */
    public function fetchHeadBottom(): string
    {
        $list = $this->doThemeEvent(static::EVENT_THEME_HEAD_BOTTOM);

        return join('', $list);
    }

    /**
     * @return string
     * @throws SystemException
     */
    public function fetchGoogleAnalytics(): ?string
    {
        $systemPlugin = Core::getInstance()->getSystemPlugin();

        if (!$systemPlugin->hasSetting('google_analytics')) {
            return '';
        }

        $googleAnalytics = $this->doThemeEvent(
            static::EVENT_THEME_FETCH_GOOGLE_ANALYTICS
        );

        if (!empty($googleAnalytics)) {
            return join('', $googleAnalytics);
        } else {
            return $systemPlugin->getSetting('google_analytics');
        }
    }

    /**
     * @return string
     * @throws SystemException
     */
    public function fetchUserActivities(): string
    {
        $userActivities = $this->doThemeEvent(
            static::EVENT_THEME_HEADER_USER_ACTIVITY
        );

        return join('', $userActivities);
    }

    /**
     * @return string
     * @throws SystemException
     */
    public function fetchProjectsList(): string
    {
        $list = $this->doThemeEvent(static::EVENT_THEME_PROJECTS_LIST);

        return join('', $list);
    }

    public function fetchFooter(): string
    {
        $list = $this->doThemeEvent(static::EVENT_THEME_FETCH_FOOTER);

        return join('', $list);
    }

    /**
     * Override theme options.
     *
     * @param array $options
     */
    public function setOptions(array $options): void
    {
        $this->info = $options;
    } // end setInfo

}
