<?php

use core\ITemplate;
use ui\IUIEngine;
use ui\UIEngine;

/**
 * Class DisplayPlugin
 *
 * @property IUIEngine $ui
 */
class DisplayPlugin extends ObjectPlugin implements ITemplate
{
    /**
     * Reference to the theme instance.
     *
     * @var ?ITheme
     */
    static private ?ITheme $_theme = null;

    public function __construct()
    {
        parent::__construct();

        $this->ui = $this->createUIEngine();
    } // end __construct

    protected function createUIEngine(): IUIEngine
    {
        return new UIEngine($this);
    }
    
    /**
     * @return string|null
     */
    public function getPluginTemplatePath(): ?string
    {
        $pluginPath = $this->getOption('plugin_path');
        if (!$pluginPath) {
            return null;
        }

        $postfix = $this->core->getOption('plugin_template_postfix');

        $path = $pluginPath."templates".$postfix.DIRECTORY_SEPARATOR;
        if (is_dir($path)) {
            return $path;
        }

        return $pluginPath."templates".DIRECTORY_SEPARATOR;
    }
    
    /**
     * @param $key
     * @param null $value
     */
    public function assign($key, $value = null)
    {
        if (is_array($key)) {
            foreach ($key as $name => $item) {
                $this->$name = $item;
            }
        } else {
            $this->$key = $value;
        }
    } // end assign

    /**
     * @param string $file
     * @param array|null $localVars
     * @param string|null $path
     * @return string
     * @throws SystemException
     */
    public function fetchFrom(string $file, ?array $localVars = null, string $path = null): string
    {
        $templatePath = $path.$file;

        if (!file_exists($templatePath)) {
            throw new SystemException('Template file not found:'.$templatePath);
        }

        $vars = $this->getArrayCopy();
        
        extract($vars);
        if ($localVars) {
            extract($localVars);
        }

        ob_start();

        include($templatePath);

        $result = ob_get_contents();
        ob_end_clean();

        return $result;
    } // end fetchFrom

    /**
     * @param string $file
     * @param array|null $localVars
     * @param string|null $templatePath
     * @return string
     * @throws SystemException
     */
    public function fetch(string $file, ?array $localVars = null, ?string $templatePath = null): string
    {
        $pluginPath = $this->getPluginTemplatePath();

        $path = $this->core->getOption(Core::OPTION_TEMPLATE_PATH);

        if (file_exists($pluginPath.$file)) {
            $path = $pluginPath;
        } else if (!file_exists($path.$file)) {
            $path = $this->core->getOption(Core::OPTION_THEME_TEMPLATE_PATH);
        }

        $result = $this->fetchFrom($file, $localVars, $path);

        if (!is_null($templatePath)) {
            $this->content = $result;
            $result = $this->fetch($templatePath);
        }

        return $result;
    } // end fetch

    protected function onDisplay()
    {
    }

    public function onBreadcrumb()
    {
    }
    
    /**
     * @param array $seo
     */
    protected function onMeta(&$seo)
    {
        if (isset($this->core->seo)) {
            $seo = $this->core->seo;
        }

        $metaFields = array(
            'title'       => Entity::FIELD_TYPE_STRING_NULL,
            'description' => Entity::FIELD_TYPE_STRING_NULL,
            'keywords'    => Entity::FIELD_TYPE_STRING_NULL,
            'image_src'   => Entity::FIELD_TYPE_STRING_NULL
        );

        $seo = $this->getExtendData($seo, $metaFields);
    }

    /**
     * Returns content in global theme context.
     *
     * @param $content
     * @param string $template
     * @return string
     * @throws SystemException
     */
    public function fetchMain($content, $template = 'main.phtml'): string
    {
        $this->content = $content;

        $this->info = $this->getInfo();

        $this->onDisplay();
        $this->onBreadcrumb();

        $seo = array();
        $this->onMeta($seo);

        $seoFields = array(
            'title'       => Entity::FIELD_TYPE_STRING_NULL,
            'description' => Entity::FIELD_TYPE_STRING_NULL,
            'keywords'    => Entity::FIELD_TYPE_STRING_NULL
        );

        $this->info['seo'] = $this->getExtendData($seo, $seoFields);

        $theme = $this->_getThemeInstance($this->info);

        return $theme->fetchMain($content, $template);
    } // end fetchMain

    public function display($content, $template = 'main.phtml'): void
    {
        echo $this->fetchMain($content, $template);
    } // end display

    /**
     * Returns theme class instance.
     *
     * @param array $options
     * @return ITheme
     * @throws SystemException
     */
    private function _getThemeInstance(array $options): ITheme
    {
        if (self::$_theme !== null) {
            self::$_theme->setOptions($options);
            return self::$_theme;
        }

        $themePath = $this->core->getOption(Core::OPTION_THEME_PATH);
        $themeClassPath = $themePath.'Theme.php';

        if (file_exists($themeClassPath)) {
            $themeClassName = require($themeClassPath);
            if (is_numeric($themeClassName)) {
                $themeClassName = 'Theme';
            }
        } else {
            $themeClassName = "DefaultTheme";
        }

        if (!class_exists($themeClassName)) {
            throw new SystemException("Undefined theme class: ".$themeClassName);
        }

        self::$_theme = new $themeClassName($options);

        assert(self::$_theme instanceof ITheme);

        return self::$_theme;
    } // end _getThemeInstance

    public function getInfo(): array
    {
        $info = array(
            'basehttp'       => $this->core->getOption('http_base'),
            'engine_url'     => $this->core->getOption('engine_url'),
            'theme_url'      => $this->core->getOption('theme_url'),
            'base_http_icon' => $this->core->getOption('http_base_icon'),
            'charset'        => $this->core->getOption('charset'),
        );

        $info += $this->core->getProperties();

        return $info;
    }
    
    /**
     * @param string $relativePath
     * @throws SystemException
     */
    public function includeStatic(string $relativePath)
    {
        $url = $this->getStaticUrl($relativePath);
        $pathUrl = parse_url($url, PHP_URL_PATH);
        $ext = pathinfo($pathUrl, PATHINFO_EXTENSION);
        
        if ($ext == "js") {
            $this->core->includeJs($url, false);
        } else if ($ext == "css") {
            $this->core->includeCss($url, false);
        } else {
            throw new SystemException("Undefined resource type: ".$url);
        }
    } // end includeStatic
    
    /**
     * @param string $relativePath
     * @return string
     */
    public function getStaticUrl(string $relativePath): string
    {
        return $this->getOption('plugin_http')."static/".$relativePath;
    } // end getStaticUrl
    
    /**
     * @param string $url
     * @return string
     */
    public function getUrl(string $url): string
    {
        $method = array(&$this->core, 'getUrl');
        return call_user_func_array($method, func_get_args());
    } // end getUrl

    /**
     * @param string $themeEventName
     * @return array
     */
    public function doThemeEvent(string $themeEventName): array
    {
        $results = $this->core->fireEvent($themeEventName);

        return $results ?? array();
    }
}
